;;;
;;; banyan/connector
;;;
;;;  Copyright(C) 2003 by Shiro Kawai (shiro@acm.org)
;;;
;;;  Permission to use, copy, modify, distribute this software and
;;;  accompanying documentation for any purpose is hereby granted,
;;;  provided that existing copyright notices are retained in all
;;;  copies and that this notice is included verbatim in all
;;;  distributions.
;;;  This software is provided as is, without express or implied
;;;  warranty.  In no circumstances the author(s) shall be liable
;;;  for any damages arising out of the use of this software.
;;;
;;;  $Id: connector.scm,v 1.3 2003/03/02 08:38:41 shirok Exp $
;;;

(select-module banyan)

;; PROTOTYPE IMPLEMENTATION

;;============================================================
;; Connector
;;

;; Connector is a special node that holds the pointer 
;;
;; The transform hierarchy is "reverse order" in Banyan, i.e. 
;; child transforms keep a pointer to the parent transform.
;; If children and parent belong different part, the children's
;; part would depent on the parent's part---however, this is
;; conterintuitive in the most cases, where subparts are developed
;; individually and the parent parts synthesize them.
;;
;; Connector node lets the parent part include the subpart with
;; having pointer from parent to child.
;;
;; A connector is realized as two nodes, one in the parent part and
;; the other in the child part.   Parent-side connector keeps an
;; attribute to the child-side connector.  The run-time reverse
;; pointer is created when the relation is established.

(define-class <child-connector> (<transform>)
  ((parent-connector :accessor parent-connector-of :initform #f)
   )
  :c-class <%transform>)

(define-class <parent-connector> (<instance>)
  ((child    :accessor child-of :initform #f :init-keyword :child
             :attribute-type <child-connector>
             :validator (lambda (o v)
                          (cond ((not v)
                                 (and-let* ((old (ref o 'child)))
                                   (set! (parent-connector-of old) #f)))
                                ((is-a? v <child-connector>)
                                 (set! (parent-connector-of v) o))
                                (else
                                 (error "parent connector should be connected to child connector, but got" v)))
                          v))
   ))

;; Kludge to avoid initialization time of child slot
(define-method slot-unbound ((class <node-meta>)
                             (object <parent-connector>) slot)
  (if (eq? slot 'child)
      #f
      (next-method)))

;;----------------------------------------------------------
;; Child connector updatable-node protocol
;;

(define-method node-depends ((self <child-connector>))
  (or (and-let* ((p (parent-connector-of self)))
        (list (transform-of p)))
      '()))

(define-method node-update ((self <child-connector>))
  (and-let* ((p (parent-connector-of self)))
    (matrix4f-copy! (gmatrix-of self)
                    (gmatrix-of (transform-of p)))
    ))

;;----------------------------------------------------------
;; Parent connector protocol
;;

(define-method initialize ((self <parent-connector>) initargs)
  (next-method self (list* :object (make <connector-shape>) initargs))
  )

;;----------------------------------------------------------
;; Convenience function
;;

(define (make-connector child-instance . args)
  (unless (is-a? child-instance <instance>)
    (error "connector must take instance as a child, but got" child-instance))
  (let* ((child-conn (with-part (part-of child-instance)
                       (make <child-connector>)))
         (parent-conn (make <parent-connector> :child child-conn)))
    (set! (parent-of (transform-of child-instance)) child-conn)
    (add-instance (current-scene) parent-conn)
    (when (get-keyword :select args #t)
      (reset-selections (current-scene))
      (add-selection (current-scene) parent-conn))
    parent-conn))

(provide "banyan/connector")

