;;;
;;; banyan/export/vrml - export to VRML
;;;
;;;  Copyright(C) 2003 by Shiro Kawai (shiro@acm.org)
;;;
;;;  Permission to use, copy, modify, distribute this software and
;;;  accompanying documentation for any purpose is hereby granted,
;;;  provided that existing copyright notices are retained in all
;;;  copies and that this notice is included verbatim in all
;;;  distributions.
;;;  This software is provided as is, without express or implied
;;;  warranty.  In no circumstances the author(s) shall be liable
;;;  for any damages arising out of the use of this software.
;;;
;;;  $Id: vrml.scm,v 1.10 2003/02/21 06:53:06 shirok Exp $
;;;

(define-module banyan.export.vrml
  (extend banyan)
  (use srfi-1)
  (use srfi-2)
  (use gauche.sequence)
  (use gauche.uvector)
  (use x3d.node)
  (use x3d.vrml)
  (use util.list)
  (use gl)
  (export export-to-vrml))

(select-module banyan.export.vrml)

(define-method export-to-vrml ((scene <scene>) port)
  (let ((trees (make-instance-tree (instances-of scene))))
    (x3d-write-scene 
     (Group :children
            (append (list (NavigationInfo :head-light #f :type '("EXAMINE")))
                    (banyan->vrml default-light-group)
                    (append-map banyan->vrml trees)))
     :port port))
  )

;; A special "children-of", that considers connector node.
(define-method x-children-of ((ins <instance>))
  (children-of ins))

(define-method x-children-of ((conn <parent-connector>))
  (or (and-let* ((xform (transform-of conn))
                 (child-conn (child-of conn)))
        (filter-map (lambda (node&slot)
                      (and (is-a? (car node&slot) <instance>)
                           (eq? (cdr node&slot) 'transform)
                           (car node&slot)))
                    (append-map node-collect-reference
                                (children-of child-conn))))
      '()))

(define-method x-children-of ((obj <node>)) '()) ;;fallback

;; From given instances, construct a tree according to the transformation
;; hierarchy.
(define (make-instance-tree instances)
  (forest->trees instances
                 (lambda (proc tree)
                   (for-each proc (x-children-of tree)))
                 eq?))

;; utility to coerce a 4-element vector (incl: vector4f, point4f) to
;; 3-element vector.
(define (x->vec3 x) (map (cut ref x <>) '(0 1 2)))

;; Fallback method
(define-method banyan->vrml ((dummy <top>))
  '())

;; Instance conversion --------------------------------------------

(define-method banyan->vrml ((ins <instance>))
  (let ((xform  (transform-of ins))
        (shader (shader-of ins))
        (object (object-of ins)))
    (if (and xform object)
        (banyan-transform->vrml xform
                                (list* shader object (x-children-of ins)))
        '())))

;; Transform conversion --------------------------------------------

(define (banyan-transform->vrml xform children)
  (list
   (Transform
    :translation (x->vec3 (translation-of xform))
    :rotation (quatf->rotation (rotation-of xform))
    :scale    (x->vec3 (scale-of xform))
    :children (append-map banyan->vrml children))))

;; Shader conversion  ----------------------------------------------

(define-method banyan->vrml ((shader <shader>))
  ;; fallback
  '())

(define-method banyan->vrml ((shader <gl-shader>))
  (let* ((m (material-of shader)))
    (list
     (Material
      :ambient-intensity (rgb->luminance (ambient-of m))
      :diffuse-color (x->vec3 (diffuse-of m))
      :emissive-color (x->vec3 (emission-of m))
      :shininess (shininess-of m)
      :specular-color (x->vec3 (specular-of m))
      )
     )))

;; Light & Lightgroup conversion -----------------------------------

(define-method banyan->vrml ((lg <light-group>))
  (apply append!
         (map (lambda (c-light)
                (or (and-let* ((c-light)
                               (light (ref c-light 'node))
                               (xform (transform-of light))
                               )
                      (banyan-transform->vrml xform (list light)))
                    '()))
              (lights-of lg))))

(define-method banyan->vrml ((light <light>))
  '())

(define-method banyan->vrml ((light <gl-light>))
  (list (DirectionalLight
         :ambient-intensity (rgb->luminance (ambient-of light))
         :color (diffuse-of light))))

;; Shape conversion ------------------------------------------------
;;  Need to separate face set and line set.
(define-method banyan->vrml ((shape <polygon-shape>))
  (unless (ref shape 'normals)
    (shape-set-vertex-normals! shape))
  (receive (lines faces)
      (partition (lambda (facets)
                   (member (car facets)
                           '(GL_LINES GL_LINE_STRIP GL_LINE_LOOP)))
                 (ref shape 'faces))
    (cond
     ((null? lines)
      (list (banyan-shape->vrml-indexed-faces (ref shape 'vertices)
                                              (ref shape 'normals)
                                              faces)))
     ((null? faces)
      (list (banyan-shape->vrml-indexed-lines (ref shape 'vertices)
                                              lines)))
     (else
      ;; TODO: writeme
      (error "A shape with mixed Lines/Polygons is not supported"))
     )))

(define (banyan-shape->vrml-indexed-faces vertices normals faces)
  (let ((face-index (banyan-shape-serialize-index faces)))
    (IndexedFaceSet
     :coord (Coordinate :point (map (lambda (v) (map (cut ref v <>) '(0 1 2)))
                                    vertices))
     :normal (Normal :point (map (lambda (v) (map (cut ref v <>) '(0 1 2)))
                                 normals))
     :coord-index face-index
     :normal-index face-index
     )))

(define (banyan-shape->vrml-indexed-lines vertices lines)
  (IndexedLineSet
   :coord (Coordinate :point (map (lambda (v) (map (cut ref v <>) '(0 1 2)))
                                  vertices))
   :coord-index (banyan-shape-serialize-index lines)))

(define (banyan-shape-serialize-index isets)
  (with-builder (<list> add! get)
    (dolist (iset isets)
      (let ((type    (car iset))
            (indices (coerce-to <list> (cadr iset))))
        (cond
         ((eq? type GL_POLYGON) (for-each add! indices) (add! -1))
         ((eq? type GL_TRIANGLES)
          (for-each (lambda (tri) (for-each add! tri) (add! -1))
                    (slices indices 3)))
         ((eq? type GL_TRIANGLE_STRIP)
          (letrec ((even (lambda (v0 v1 v2 . rest)
                           (add! v0) (add! v1) (add! v2) (add! -1)
                           (unless (null? rest) (apply odd v2 v1 rest))))
                   (odd  (lambda (v2 v1 v3 . rest)
                           (add! v2) (add! v1) (add! v3) (add! -1)
                           (unless (null? rest) (apply even v2 v3 rest)))))
            (apply even indices)))
         ((eq? type GL_TRIANGLE_FAN)
          (letrec ((v0 (car indices))
                   (loop (lambda (v1 v2 . rest)
                           (add! v0) (add! v1) (add! v2) (add! -1)
                           (unless (null? rest) (apply loop v2 rest)))))
            (apply loop (cdr indices))))
         ((eq? type GL_QUADS)
          (for-each (lambda (quad) (for-each add! quad) (add! -1))
                    (slices indices 4)))
         ((eq? type GL_QUAD_STRIP)
          (letrec ((loop (lambda (v0 v1 v2 v3 . rest)
                           (add! v0) (add! v1) (add! v3) (add! v2) (add! -1)
                           (unless (null? rest) (apply loop v2 v3 rest)))))
            (apply loop indices)))
         ((eq? type GL_LINES)
          (for-each (lambda (line) (for-each add! line) (add! -1))
                    (slices indices 2)))
         ((eq? type GL_LINE_STRIP)
          (for-each add! indices) (add! -1))
         ((eq? type GL_LINE_LOOP)
          (let1 v0 (car indices)
            (for-each add! indices) (add! v0) (add! -1)))
         )))
    (get)))

(provide "banyan/export/vrml")
