;;;
;;; banyan/gui/attribed - attribute editor
;;;
;;;  Copyright(C) 2002-2003 by Shiro Kawai (shiro@acm.org)
;;;
;;;  Permission to use, copy, modify, distribute this software and
;;;  accompanying documentation for any purpose is hereby granted,
;;;  provided that existing copyright notices are retained in all
;;;  copies and that this notice is included verbatim in all
;;;  distributions.
;;;  This software is provided as is, without express or implied
;;;  warranty.  In no circumstances the author(s) shall be liable
;;;  for any damages arising out of the use of this software.
;;;
;;;  $Id: attribed.scm,v 1.17 2003/01/21 11:18:36 shirok Exp $
;;;

(define-module banyan.gui.attribed
  (extend banyan.gui)
  (use srfi-1)
  (use srfi-2)
  (use srfi-13)
  (export <attribute-editor>
          attribute-editor-attach
          attribute-editor-detach))
(select-module banyan.gui.attribed)

;; PROTOTYPE IMPLEMENTATION

;;==================================================================
;; Attribute editor main window
;;

(define-glim-application-frame <attribute-editor> ()
  ((node         :accessor node-of :init-value #f)     ;target node
   (node-history-prev :init-value '())
   (node-history-next :init-value '())
   (fields       :accessor fields-of :init-value '())
   (field-table-num-items :accessor field-table-num-items-of :init-value 0)
   (field-table  :allocation :virtual :accessor field-table-of
                 :slot-ref  (lambda (f)
                              (widget-of (glim-get-frame-pane f 'field-table)))
                 :slot-set! (lambda (f v) #f))
   (field-pool   :allocation :class :accessor field-pool-of
                 :initform '())
   )
  (:panes
   (navigation  (make <attribute-editor-navigation-pane> :parent .frame.))
   (node-name   (glim-gtk-widget->pane (gtk-entry-new)))
   (label       (glim-gtk-widget->pane (gtk-label-new "")))
   (field-table (glim-gtk-widget->pane (gtk-table-new 20 2 #f)))
   )
  (:layouts
   (default
     (glim-vertically ()
       navigation
       (----)
       (glim-labelling (:label "Node: " :label-alignment :left)
         node-name)
;       label
       (:expand
        (glim-scrolling (:scroll-bar :both :width 250 :height 400)
          (glim-vertically () field-table)))
       (----)
       (dialog-buttons
        (dialog-action-gadget "Close" 'ok))
       )
     ))
  )

(define-method initialize ((self <attribute-editor>) initargs)
  (next-method self
               (list* :name 'attribute-editor
                      :pretty-name "Attribute Editor"
                      initargs))
  (add-hook! selection-watchers attribute-editor-selection-changed)
  ;; node-name stuff
  (let ((entry (widget-of (glim-get-frame-pane self 'node-name))))
    (gtk-entry-set-editable entry #f)
    (g-signal-connect entry "activate"
                      (lambda _
                        (and-let* ((node (node-of self))
                                   (name (gtk-entry-get-text entry))
                                   (uname (get-unique-name (part-of node) name)))
                          (set! (name-of node) uname)
                          (unless (equal? name uname)
                            (gtk-entry-set-text entry uname)))))
                                                          
    ))

(define-method dialog-action ((self <attribute-editor>) action)
  (glim-frame-exit self))

(define (get-attribute-field class)
  (let* ((pool (class-slot-ref <attribute-editor> 'field-pool))
         (fobj (find (lambda (x) (eq? (class-of x) class)) pool)))
    (if fobj
        (begin
          (set! (class-slot-ref <attribute-editor> 'field-pool)
                (delete fobj pool eq?))
          fobj)
        (make class))))

;;------------------------------------------------------------------
;; attach and detach
;;

;; attribute-editor-attach node &optional no-history?
(define (attribute-editor-attach node . opts)
  (let ((self (instance-of <attribute-editor>))
        (use-history? (get-optional opts #t)))
    (when use-history?
      (and-let* ((prev (node-of self)))
        (push! (ref self 'node-history-prev) prev))
      (set! (ref self 'node-history-next) '()))
    (attribute-editor-detach)
    (set! (node-of self) node)
    (when node
      (let1 entry (widget-of (glim-get-frame-pane self 'node-name))
        (gtk-entry-set-text entry (name-of node))
        (gtk-entry-set-editable entry #t))
      (for-each (lambda (slot)
                  (for-each (lambda (field) (push! (fields-of self) field))
                            (add-attribute-field 
                             (slot-definition-option slot :attribute-type)
                             self node slot)))
                (node-attribute-slots node)))
    (update-navigation-state self)))

(define (attribute-editor-detach)
  (let ((self (instance-of <attribute-editor>)))
    (set! (node-of self) #f)
    (let1 entry (widget-of (glim-get-frame-pane self 'node-name))
      (gtk-entry-set-text entry "")
      (gtk-entry-set-editable entry #f))
    (for-each (cut attribute-field-detach <> (field-table-of self))
              (fields-of self))
    (set! (field-pool-of self) (append (field-pool-of self) (fields-of self)))
    (set! (field-table-num-items-of self) 0)
    (set! (fields-of self) '())
    ))

(define (attribute-editor-selection-changed scene selected)
  (if  (null? selected)
       (attribute-editor-attach #f)
       ;; TODO: what if more than one obj is selected?
       (attribute-editor-attach (car selected))))

;;------------------------------------------------------------------
;; navigation
;;

(define-class <attribute-editor-navigation-pane> (<glim-gadget>)
  ((parent :init-keyword :parent)
   (prev)
   (next)
   ))

(define-method initialize ((self <attribute-editor-navigation-pane>) initargs)
  (let1 hbox (gtk-hbox-new #f 2)
    (define (stock-button stock cmd)
      (let1 button (gtk-button-new-from-stock stock)
        (gtk-box-pack-start hbox button #f #f 0)
        (gtk-button-set-relief button GTK_RELIEF_NONE)
        (gtk-widget-show button)
        (g-signal-connect button "clicked" (lambda _ (cmd)))
        button))
    (set! (widget-of self) hbox)
    (set! (ref self 'prev) (stock-button GTK_STOCK_GO_BACK
                                         attribute-editor-prev-node))
    (set! (ref self 'next) (stock-button GTK_STOCK_GO_FORWARD
                                         attribute-editor-next-node)))
  (next-method)
  (update-navigation-state self))

(define (attribute-editor-clear-history)
  (let ((self (instance-of <attribute-editor>)))
    (set! (ref self 'node-history-prev) '())
    (set! (ref self 'node-history-next) '())))

(define-method update-navigation-state ((self <attribute-editor-navigation-pane>))
  (let1 frame (ref self 'parent)
    (gtk-widget-set-sensitive (ref self 'prev)
                              (not (null? (ref frame 'node-history-prev))))
    (gtk-widget-set-sensitive (ref self 'next)
                              (not (null? (ref frame 'node-history-next)))))
  )

(define-method update-navigation-state ((self <attribute-editor>))
  (update-navigation-state (glim-get-frame-pane self 'navigation)))

(define (attribute-editor-prev-node)
  (let* ((self (instance-of <attribute-editor>))
         (hist (ref self 'node-history-prev)))
    (unless (null? hist)
      (and-let* ((current (node-of self)))
        (push! (ref self 'node-history-next) current))
      (attribute-editor-attach (pop! (ref self 'node-history-prev)) #f)
      (update-navigation-state self))
    ))

(define (attribute-editor-next-node)
  (let* ((self (instance-of <attribute-editor>))
         (hist (ref self 'node-history-next)))
    (unless (null? hist)
      (and-let* ((current (node-of self)))
        (push! (ref self 'node-history-prev) current))
      (attribute-editor-attach (pop! (ref self 'node-history-next)) #f)
      (update-navigation-state self))
    ))

;;==================================================================
;; attribute-field
;;

;; Attribute field object manages a set of widgets to edit
;; an attribute of specific type.

;; Base ------------------------------------------------------------
(define-class <attribute-field> ()
  ((label-widget :accessor label-widget-of)
   (value-widget :accessor value-widget-of :init-value #f)
   (row-number   :accessor row-number-of)
   (value-changed :accessor value-changed-of :init-value #f)
   (observer-alist :accessor observer-alist-of :init-value '())
   ;; - alist of node & observer procedure
   ))

(define-method initialize ((self <attribute-field>) initargs)
  (next-method)
  (let ((label (gtk-label-new ""))
        )
    (gtk-widget-set-size-request label 80 -1)
    (gtk-misc-set-alignment label 0.0 0.5)
    (gtk-widget-show label)
    (set! (ref self 'label-widget) label)
    ))

(define-method attribute-field-pack ((self <attribute-field>) parent)
  (let ((table (field-table-of parent))
        (num   (field-table-num-items-of parent)))
    (gtk-table-attach-defaults table
                               (label-widget-of self)
                               0 1 num (+ num 1))
    (when (value-widget-of self)
      (gtk-table-attach-defaults table
                                 (value-widget-of self)
                                 1 2 num (+ num 1)))
    (set! (row-number-of self) num)
    (set! (field-table-num-items-of parent) (+ num 1))
    ))

(define-method attribute-field-detach ((self <attribute-field>) parent)
  (gtk-container-remove parent (label-widget-of self))
  (when (value-widget-of self)
    (gtk-container-remove parent (value-widget-of self)))
  (set! (value-changed-of self) #f)
  (for-each (lambda (p)
              (node-observer-delete! (car p) (cdr p)))
            (observer-alist-of self))
  (set! (observer-alist-of self) '())
  )

(define-method attribute-field-value-changed ((self <attribute-field>))
  (and-let* ((callback (value-changed-of self)))
    (callback self)))

(define-method label-text-of ((self <attribute-field>))
  (gtk-label-get-text (label-widget-of self)))

(define-method (setter label-text-of) ((self <attribute-field>) text)
  (gtk-label-set-text (label-widget-of self) text))

(define-method value-text-of ((self <attribute-field>)) "")

(define-method (setter value-text-of) ((self <attribute-field>) text)
  #f)

;; Label -----------------------------------------------------------
(define-class <attribute-field-label> (<attribute-field>)
  ((value-label)))

(define-method initialize ((self <attribute-field-label>) initargs)
  (next-method)
  (let1 w (gtk-label-new "")
    (set! (value-widget-of self) w)
    (gtk-misc-set-padding w 4 4))
  (gtk-widget-show (value-widget-of self))
  )

(define-method value-text-of ((self <attribute-field-label>))
  (gtk-label-get-text (value-widget-of self)))

(define-method (setter value-text-of) ((self <attribute-field-label>) text)
  (gtk-label-set-text (value-widget-of self) text))

;; Entry -----------------------------------------------------------
(define-class <attribute-field-entry> (<attribute-field>)
  ((prev-value :init-value #f)))

(define-method initialize ((self <attribute-field-entry>) initargs)
  (next-method)
  (let1 entry (gtk-entry-new)
    (set! (value-widget-of self) entry)
    (g-signal-connect entry "enter-notify-event"
                      (lambda _
                        (set! (ref self 'prev-value)
                              (gtk-entry-get-text entry))))
    (g-signal-connect entry "leave-notify-event"
                      (lambda _
                        (unless (equal? (ref self 'prev-value)
                                        (gtk-entry-get-text entry))
                          (attribute-field-value-changed self))))
    (gtk-widget-show entry))
  )

(define-method value-text-of ((self <attribute-field-entry>))
  (gtk-entry-get-text (value-widget-of self)))

(define-method (setter value-text-of) ((self <attribute-field-entry>) text)
  (gtk-entry-set-text (value-widget-of self) text))

;; Button ----------------------------------------------------------
(define-class <attribute-field-button> (<attribute-field>)
  ())

(define-method initialize ((self <attribute-field-button>) initargs)
  (next-method)
  (let1 button (gtk-button-new)
    (set! (value-widget-of self) button)
    (g-signal-connect button "clicked"
                      (lambda _ (attribute-field-value-changed self)))
    (gtk-widget-show button)))

(define-method value-text-of ((self <attribute-field-button>))
  (ref (value-widget-of self) 'label-text))

(define-method (setter value-text-of) ((self <attribute-field-button>) text)
  (gtk-button-set-label (value-widget-of self) text))

;; Node ----------------------------------------------------------

;; combination of two buttons
(define-class <attribute-field-node> (<attribute-field>)
  ((node-button)
   (node       :accessor node-of)
   (node-class :accessor node-class-of)
   (connection-changed :init-value #f :accessor connection-changed-of)
   ))

(define-method initialize ((self <attribute-field-node>) initargs)
  (next-method)
  (let* ((hbox (gtk-hbox-new #f 0))
         (node-button (gtk-button-new))
         (selector-arrow (gtk-arrow-new GTK_ARROW_RIGHT GTK_SHADOW_NONE))
         (selector-button (gtk-button-new)))

    (define (selector-callback selected)
      (unless (eq? selected (ref self 'node))
        (gtk-button-set-label node-button
                              (if selected (name-of selected) "(nil)"))
        (set! (ref self 'node) selected)
        (cond ((connection-changed-of self) => (cut <> selected)))))

    (set! (value-widget-of self) hbox)
    (set! (ref self 'node-button) node-button)
    (gtk-box-pack-start hbox node-button #t #t 0)
    (gtk-box-pack-start hbox selector-button #f #f 0)
    (gtk-container-add selector-button selector-arrow)
    (g-signal-connect node-button "clicked"
                      (lambda _ (attribute-field-value-changed self)))
    (g-signal-connect selector-button "clicked"
                      (lambda _
                        (node-selector-activate! selector-callback
                                                 :class (node-class-of self)
                                                 )))
    (gtk-widget-show node-button)
    (gtk-widget-show selector-arrow)
    (gtk-widget-show selector-button)
    (gtk-widget-show hbox)))

(define-method value-text-of ((self <attribute-field-node>))
  (ref (ref self 'node-button) 'label-text))

(define-method (setter value-text-of) ((self <attribute-field-node>) text)
  (gtk-button-set-label (ref self 'node-button) text))

;; Bool -----------------------------------------------------------
(define-class <attribute-field-boolean> (<attribute-field>)
  ())

(define-method initialize ((self <attribute-field-boolean>) initargs)
  (next-method)
  (set! (value-widget-of self) (gtk-check-button-new))
  (g-signal-connect (value-widget-of self) "toggled"
                    (lambda _ (attribute-field-value-changed self)))
  (gtk-widget-show (value-widget-of self))
  )

(define-method value-boolean-of ((self <attribute-field-boolean>))
  (gtk-toggle-button-get-active (value-widget-of self)))

(define-method (setter value-boolean-of) ((self <attribute-field-boolean>) value)
  (gtk-toggle-button-set-active (value-widget-of self) value))

;; Real -----------------------------------------------------------
;;  Value widget is a spin button.
(define-class <attribute-field-real> (<attribute-field>)
  ())

(define-method initialize ((self <attribute-field-real>) initargs)
  (next-method)
  (let* ((adj  (gtk-adjustment-new 0.0 -1000000.0 10000000.0 0.01 0.1 0.1))
         (spin (gtk-spin-button-new adj 0.001 3)))
    (set! (value-widget-of self) spin)
    (g-signal-connect spin "value-changed"
                      (lambda _ (attribute-field-value-changed self)))
    (gtk-widget-show (value-widget-of self))
    ))

(define-method value-real-of ((self <attribute-field-real>))
  (gtk-spin-button-get-value (value-widget-of self)))

(define-method (setter value-real-of) ((self <attribute-field-real>) value)
  (gtk-spin-button-set-value (value-widget-of self) value))

;; Clamp ----------------------------------------------------------
;;  Value widget is a spin button.  Value clamped in [0,1]
(define-class <attribute-field-clamp> (<attribute-field>)
  ())

(define-method initialize ((self <attribute-field-clamp>) initargs)
  (next-method)
  (let* ((adj  (gtk-adjustment-new 0.0 0.0 1.0 0.001 0.01 0.01))
         (spin (gtk-spin-button-new adj 0.001 3)))
    (set! (value-widget-of self) spin)
    (g-signal-connect spin "value-changed"
                      (lambda _ (attribute-field-value-changed self)))
    (gtk-widget-show (value-widget-of self))
    ))

(define-method value-real-of ((self <attribute-field-clamp>))
  (gtk-spin-button-get-value (value-widget-of self)))

(define-method (setter value-real-of) ((self <attribute-field-clamp>) value)
  (gtk-spin-button-set-value (value-widget-of self) value))

;; Color ----------------------------------------------------------
(define-class <attribute-field-color> (<attribute-field>)
  ((csel-dialogue :allocation :class :init-value #f)
   (csel-ok-handler-id :allocation :class :init-value #f)
   (csel-cancel-handler-id :allocation :class :init-value #f)
   ))

(define-method initialize ((self <attribute-field-color>) initargs)
  (next-method)
  (let1 button (gtk-button-new)
    (set! (value-widget-of self) button)
    (g-signal-connect button "clicked"
                      (lambda (w)
                        (color-selection-dialog-popup self)))
    (gtk-button-set-label button " ")
    (gtk-widget-show button)
    ))

(define-method color-selection-dialog-popup ((self <attribute-field-color>))
  (unless (slot-ref self 'csel-dialogue)
    (set! (slot-ref self 'csel-dialogue)
          (gtk-color-selection-dialog-new "Color selection")))
  (let* ((sel  (slot-ref self 'csel-dialogue))
         (selw (ref sel 'colorsel))
         (ok   (ref sel 'ok-button))
         (cancel (ref sel 'cancel-button))
         (color (bn-color->gdk-color (value-color-of self))))
    (gtk-color-selection-set-current-color selw color)
    (gtk-color-selection-set-previous-color selw color)
    (cond ((ref self 'csel-ok-handler-id)
           => (cut g-signal-handler-disconnect ok <>)))
    (cond ((ref self 'csel-cancel-handler-id)
           => (cut g-signal-handler-disconnect cancel <>)))
    (set! (ref self 'csel-ok-handler-id)
          (g-signal-connect ok "clicked"
                            (lambda (w)
                              (set! (value-color-of self)
                                    (let1 color (make <gdk-color>)
                                      (gtk-color-selection-get-current-color selw color)
                                      (gdk-color->bn-color color)))
                              (attribute-field-value-changed self)
                              (gtk-widget-hide sel))))
    (set! (ref self 'csel-cancel-handler-id)
          (g-signal-connect cancel "clicked"
                            (lambda (w)
                              (gtk-widget-hide sel))))
    (gtk-widget-show sel)
    ))

(define-method value-color-of ((self <attribute-field-color>))
  ;; NB: this is inefficient
  (let1 color (ref (ref (gtk-widget-get-style (value-widget-of self))
                        'bg)
                   0)
    (gdk-color->bn-color color)))

(define-method (setter value-color-of) ((self <attribute-field-color>) color)
  (for-each (cute gtk-widget-modify-bg (value-widget-of self) <>
                  (bn-color->gdk-color color))
            (list GTK_STATE_NORMAL GTK_STATE_ACTIVE GTK_STATE_INSENSITIVE
                  GTK_STATE_PRELIGHT GTK_STATE_SELECTED)))

;;==================================================================
;; populate attribute editing box
;;

;; Default ---------------------------------------------------------
(define-method add-attribute-field ((slot-type <top>)
                                    store object slot)
  (let ((f (get-attribute-field <attribute-field-label>)))
    (set! (label-text-of f) (x->string (slot-definition-name slot)))
    (set! (value-text-of f) (x->string slot-type))
    (attribute-field-pack f store)
    (list f)))

;; Node ------------------------------------------------------------
;; NB: node slot might be #f.
(define-method add-attribute-field ((slot-type <node-meta>)
                                    store object slot)
  (let ((f (get-attribute-field <attribute-field-node>))
        (node (slot-ref object (slot-definition-name slot))))
    (set! (connection-changed-of f)
          (lambda (newnode)
            (slot-set! object (slot-definition-name slot) newnode)
            (notify-graph-changed (current-scene))
            (set-modified! (part-of object))
            ))
    (set! (label-text-of f) (x->string (slot-definition-name slot)))
    (set! (node-class-of f) slot-type)
    (set! (node-of f) node)
    (set! (value-text-of f) (if node (x->string (name-of node)) "(nil)"))
    (set! (value-changed-of f)
          (lambda (f) (attribute-editor-attach (node-of f))))
    (attribute-field-pack f store)
    (list f)
    ))

;; Boolean ---------------------------------------------------------
(define-method add-attribute-field ((slot-type <boolean-meta>)
                                    store object slot)
  (let ((f (get-attribute-field <attribute-field-boolean>))
        (name (slot-definition-name slot)))
    (set! (label-text-of f) (x->string (slot-definition-name slot)))
    (set! (value-boolean-of f) (slot-ref object name))
    (set! (value-changed-of f)
          (lambda (f)
            (slot-set! object name (value-boolean-of f))
            (set-modified! (part-of object))
            ))
    (attribute-field-pack f store)
    (list f)))

;; Real ------------------------------------------------------------
(define-method add-attribute-field ((slot-type <real-meta>)
                                    store object slot)
  (let ((f (get-attribute-field <attribute-field-real>))
        (name (slot-definition-name slot)))
    (set! (label-text-of f) (x->string (slot-definition-name slot)))
    (set! (value-real-of f) (slot-ref object name))
    (set! (value-changed-of f)
          (lambda (f)
            (slot-set! object name (value-real-of f))
            (set-modified! (part-of object))
            ))
    (attribute-field-pack f store)
    (list f)))

;; Symbol ------------------------------------------------------------
(define-method add-attribute-field ((slot-type <symbol-meta>)
                                    store object slot)
  (let ((f (get-attribute-field <attribute-field-entry>))
        (name (slot-definition-name slot)))
    (set! (label-text-of f) (x->string (slot-definition-name slot)))
    (set! (value-text-of f) (x->string (slot-ref object name)))
    (set! (value-changed-of f)
          (lambda (f)
            (let1 v (string-trim-both (value-text-of f))
              (if (string-null? v)
                  (slot-set! object name #f)
                  (slot-set! object name (string->symbol v))))
            (set-modified! (part-of object))
            ))
    (attribute-field-pack f store)
    (list f)))

;; Color ------------------------------------------------------------
(define (make-color-element-updator elt node slot f-all)
  (lambda (f) 
    (set! (ref (ref node slot) elt) (value-real-of f))
    (set! (value-color-of f-all) (ref node slot))
    (set-modified! (part-of node))
    (viewports-post-redisplay)))

(define-method add-attribute-field ((slot-type <bn-color-meta>)
                                    store object slot)
  (let* ((f-all (get-attribute-field <attribute-field-color>))
         (f-rgb (list (get-attribute-field <attribute-field-clamp>)
                      (get-attribute-field <attribute-field-clamp>)
                      (get-attribute-field <attribute-field-clamp>)))
         (ind   (list 0 1 2))
         (sufx  (list "R" "G" "B"))
         (name  (slot-definition-name slot))
         (cvec  (slot-ref object name)))
    (set! (label-text-of f-all) (x->string name))
    (set! (value-color-of f-all) cvec)
    (set! (value-changed-of f-all)
          (lambda (f)
            (let ((newcol (value-color-of f)))
              (for-each (lambda (f ind)
                          (set! (value-real-of f) (ref newcol ind)))
                        f-rgb ind)
              (set! (slot-ref object name) newcol)
              (set-modified! (part-of object))
              )))
    (attribute-field-pack f-all store)
    (for-each
     (lambda (f ind suf)
       (set! (label-text-of f) #`"  ,|name|: ,|suf|")
       (set! (value-real-of f) (ref cvec ind))
       (set! (value-changed-of f)
             (make-color-element-updator ind object name f-all))
       (attribute-field-pack f store))
     f-rgb ind sufx)
    (cons f-all f-rgb)
    ))

;; Vector ------------------------------------------------------------
(define (make-vector-element-updator elt node slot)
  (lambda (f)
    (set! (ref (ref node slot) elt) (value-real-of f))
    (set-modified! (part-of node))
    (viewports-post-redisplay)))

(define-method add-attribute-field ((slot-type <bn-vec3-meta>)
                                    store object slot)
  (let* ((ind   (list 0 1 2))
         (fxyz  (list (get-attribute-field <attribute-field-real>)
                      (get-attribute-field <attribute-field-real>)
                      (get-attribute-field <attribute-field-real>)))
         (sufx  (list "X" "Y" "Z"))
         (name  (slot-definition-name slot))
         (vec3  (slot-ref object name)))
    (for-each
     (lambda (f ind suf)
       (set! (label-text-of f) #`",|name|: ,|suf|")
       (set! (value-changed-of f)
             (make-vector-element-updator ind object name))
       (attribute-field-pack f store)
       (set! (value-real-of f) (ref vec3 ind)))
     fxyz ind sufx)
    (let1 observer (lambda (obj slot value)
                     (when (eq? slot name)
                       (let1 vec3 (slot-ref obj name)
                         (for-each (lambda (f ind)
                                     (set! (value-real-of f) (ref vec3 ind)))
                                   fxyz ind))))
      (push! (observer-alist-of (car fxyz)) (cons object observer))
      (node-observer-add! object observer))
    fxyz))

(provide "banyan/gui/attribed")
