;;;
;;; banyan/gui/database - database interface
;;;
;;;  Copyright(C) 2002-2003 by Shiro Kawai (shiro@acm.org)
;;;
;;;  Permission to use, copy, modify, distribute this software and
;;;  accompanying documentation for any purpose is hereby granted,
;;;  provided that existing copyright notices are retained in all
;;;  copies and that this notice is included verbatim in all
;;;  distributions.
;;;  This software is provided as is, without express or implied
;;;  warranty.  In no circumstances the author(s) shall be liable
;;;  for any damages arising out of the use of this software.
;;;
;;;  $Id: database.scm,v 1.23 2003/01/22 03:15:20 shirok Exp $
;;;

(select-module banyan.gui)
(use gauche.parameter)
(export <repository-selector>
        <repository-display>)

;; PROTOTYPE IMPLEMENTATION

;; This module provides GUI to access the repository.

;;==========================================================
;; Repository selector
;;   Change the current repository

(define-glim-application-frame <repository-selector-dialog> ()
  ()
  (:panes
   (selector (make <repository-selector>)))
  (:layouts
   (default
     (glim-vertically ()
       (:expand (glim-outlining (:style :none :thickness 8)
                  selector))
       (----)
       (dialog-buttons
        (dialog-action-gadget "OK" 'ok)
        (dialog-action-gadget "Cancel" 'cancel))
       )))
  )

(define-method initialize ((self <repository-selector-dialog>) initargs)
  (next-method self
               (list* :name 'repository-selector
                      :pretty-name "Repository Selector"
                      initargs))
  )

(define-method dialog-action ((self <repository-selector-dialog>) action)
  (when (eq? action 'ok)
    (and-let* ((uri (glim-gadget-value (glim-get-frame-pane self 'selector)))
               (repo (or (find (lambda (r) (equal? uri (ref r 'uri)))
                               (standard-repositories))
                         (open-repository uri))))
      (unless (repository-exists? repo)
        (error "No such repository" uri))
      (repository-check-logged-in repo)
      (current-repository repo)))
  (glim-frame-exit self))

;;-------------------------------------------------------
;; Combo box widget to select repository
;;
(define-class <repository-selector> (<glim-value-gadget>)
  ((w-combo)
   (w-entry      :allocation :virtual
                 :slot-ref (lambda (o) (ref (ref o 'w-combo) 'entry))
                 :slot-set! (lambda _ #f))
   (parent-frame :init-keyword :parent-frame)
   (visited-uris :init-value '())
   ))

(define-method initialize ((self <repository-selector>) initargs)
  (let ((hbox  (gtk-hbox-new #f 0))
        (repos (map uri-of
                    (or (and-let* ((cur (current-repository))
                                   ((not (memq cur (standard-repositories)))))
                          (cons cur (standard-repositories)))
                        (standard-repositories)))))
    (set! (widget-of self) hbox)
    (set! (ref self 'visited-uris) repos)
    (let1 lab (gtk-label-new "Repository: ")
      (gtk-box-pack-start hbox lab #f #f 2))
    (let1 combo (gtk-combo-new)
      (set! (ref self 'w-combo) combo)
      (gtk-combo-set-popdown-strings combo repos)
      (gtk-combo-set-value-in-list combo #f #f)
      (gtk-box-pack-start hbox combo #t #t 2)
      (gtk-widget-set-size-request combo 300 -1)
      (gtk-combo-disable-activate combo)
      (unless (null? repos)
        (gtk-entry-set-text (ref combo 'entry) (car repos)))
      )
    (next-method)))

(define-method add-repository-uri ((self <repository-selector>) uri)
  (set! (ref self 'visited-uris)
        (cons uri
              (delete uri (ref self 'visited-uris) equal?))))

(define-method glim-gadget-value ((self <repository-selector>))
  (gtk-entry-get-text (ref self 'w-entry)))

(define-method (setter glim-gadget-value) ((self <repository-selector>) value)
  (gtk-entry-set-text (ref self 'w-entry) (x->string value)))

;;=========================================================
;; A display widget to show the current selected repository
;;

(define-class <repository-display> (<glim-gadget>)
  ())

(define-method initialize ((self <repository-display>) initargs)
  (let* ((w   (gtk-hbox-new #f 2))
         (lab (gtk-label-new "Repository: "))
         (f   (gtk-frame-new #f))
         (but (gtk-button-new-with-label
               (or (and-let* ((repo (current-repository))) (uri-of repo))
                   "")))
         )
    (gtk-box-pack-start w lab #f #f 2)
    (gtk-box-pack-start w f #t #t 2)
    (gtk-container-add f but)
    (gtk-frame-set-shadow-type f GTK_SHADOW_ETCHED_IN)
    (gtk-button-set-relief but GTK_RELIEF_NONE)
    (g-signal-connect but "clicked"
                      (dialog-show+ <repository-selector-dialog>))
    (parameter-observer-add! current-repository
                             (lambda (old repo)
                               (gtk-label-set-text (gtk-bin-get-child but)
                                                   (uri-of repo))))
    (set! (widget-of self) w)
    (next-method)))


;;=========================================================
;; Node name query dialog
;;

(define-class <db-part-name-query-dialog> (<singleton-mixin>)
  ((dialog)
   (label)
   (entry)
   (response-callback :init-value #f)  ; (proc dialog entered-text ok?)
   ))

(define-method initialize ((self <db-part-name-query-dialog>) initargs)
  (next-method)
  (let* ((dialog
          (gtk-dialog-new-with-buttons "Name part"
                                       #f ; must be main app window?
                                       GTK_DIALOG_MODAL
                                       GTK_STOCK_OK
                                       GTK_RESPONSE_OK
                                       GTK_STOCK_CANCEL
                                       GTK_RESPONSE_CANCEL))
         (vbox   (ref dialog 'vbox))
         (label   (gtk-label-new ""))
         (entry  (gtk-entry-new)))
    (gtk-box-pack-start vbox label #t #f 10)
    (gtk-box-pack-start vbox entry #t #f 10)
    (set! (ref self 'dialog) dialog)
    (set! (ref self 'label) label)
    (set! (ref self 'entry) entry)
    (g-signal-connect dialog "response"
                      (lambda (w res)
                        (gtk-widget-hide-all dialog)
                        (and-let* ((cb (ref self 'response-callback)))
                          (cb self
                              (gtk-entry-get-text entry)
                              (= res GTK_RESPONSE_OK)))))
    ))

(define (db-query-part-name text callback . args)
  (let* ((query  (instance-of <db-part-name-query-dialog>))
         (dialog (ref query 'dialog))
         (label  (ref query 'label))
         (default-name (get-optional args #f)))
    (gtk-label-set-text (ref query 'label) text)
    (when default-name
      (gtk-entry-set-text (ref query 'entry) default-name))
    (set! (ref query 'response-callback) callback)
    (gtk-widget-show-all dialog)))

;;=========================================================
;; Log message entry dialog
;;

(define-glim-application-frame <db-log-entry-dialog> ()
  ((response-callback :init-value #f))
  (:panes
   (entry (make <db-log-entry-pane>))
   )
  (:layouts
   (default
     (glim-vertically (:thickness 2 :spacing 2)
       (make <glim-label-pane> :text "Enter log:" :align 'left)
       (:expand
        (glim-outlining (:style :in :height 130 :width 200)
                        entry))
       (----)
       (dialog-buttons
        (dialog-action-gadget "OK" 'ok)
        (dialog-action-gadget "Cancel" 'cancel))
       )))
  )

(define-method initialize ((self <db-log-entry-dialog>) initargs)
  (next-method self
               (list* :name 'db-log-entry
                      :pretty-name "Commit Log Entry"
                      initargs))
  )

(define-method dialog-action ((self <db-log-entry-dialog>) action)
  (when (eq? action 'ok)
    (and-let* ((cb (ref self 'response-callback)))
      (cb (glim-gadget-value (glim-get-frame-pane self 'entry)))))
  (glim-frame-exit self))

(define-class <db-log-entry-pane> (<glim-value-gadget>)
  ())

(define-method initialize ((self <db-log-entry-pane>) initargs)
  (let* ((view (gtk-text-view-new))
         (buf  (gtk-text-view-get-buffer view)))
    (gtk-text-buffer-set-text buf "" -1)
    (set! (widget-of self) view)
    (next-method)
    ))

(define-method glim-gadget-value ((self <db-log-entry-pane>))
  (let* ((buf   (gtk-text-view-get-buffer (widget-of self)))
         (start (gtk-text-buffer-get-start-iter buf))
         (end   (gtk-text-buffer-get-end-iter buf))
         )
    (gtk-text-buffer-get-text buf start end #f)))

(define-method (setter glim-gadget-value) ((self <db-log-entry-pane>) value)
  (when (string? value)
    (let ((buf (gtk-text-view-get-buffer (widget-of self))))
      (gtk-text-buffer-set-text buf value (string-size value)))))

(define (db-ask-log-entry callback)
  (let ((dialog (instance-of <db-log-entry-dialog>)))
    (set! (ref dialog 'response-callback) callback)
    (glim-run-frame-top-level dialog)))

;;=========================================================
;; Committer interface
;;

(define (db-commit-current-scene-to-repository)

  (define (ask-new-name part)
    (db-query-part-name
     "Commit the untitled part as:"
     (lambda (dialog name ok?)
       (when ok?
         (if (valid-part-name? name)
             (begin (make-part-current (current-scene)
                                       (default-part->named-part name))
                    (ask-log-message))
             (ask-new-name part))))))

  (define (ask-log-message)
    (db-ask-log-entry commit))

  (define (commit log)
    (scene-commit (current-scene) log))
  
  ;; Make sure the default part is named
  (let1 unnamed (filter (cut eq? <> (default-part))
                        (ref (current-scene) 'parts))
    (if (null? unnamed)
        (ask-log-message)
        (ask-new-name (car unnamed))))
  )

;;=========================================================
;; "New part" interface
;;

(define (db-create-new-part)

  (define (ask-new-name)
    (db-query-part-name
     "New part name:"
     (lambda (dialog name ok?)
       (when ok?
         (if (valid-part-name? name)
             (add-part (current-scene) (make <part> :name name))
             (ask-new-name))))))

  (ask-new-name))

(provide "banyan/gui/database")
