;;;
;;; banyan/gui/manipulator - 3D manipulator
;;;
;;;  Copyright(C) 2002-2003 by Shiro Kawai (shiro@acm.org)
;;;
;;;  Permission to use, copy, modify, distribute this software and
;;;  accompanying documentation for any purpose is hereby granted,
;;;  provided that existing copyright notices are retained in all
;;;  copies and that this notice is included verbatim in all
;;;  distributions.
;;;  This software is provided as is, without express or implied
;;;  warranty.  In no circumstances the author(s) shall be liable
;;;  for any damages arising out of the use of this software.
;;;
;;;  $Id: manipulator.scm,v 1.11 2003/02/20 08:25:28 shirok Exp $
;;;

(select-module banyan.gui)
(use srfi-1)
(use srfi-2)
(use srfi-11)
(use gauche.uvector)
(use gauche.sequence)
(use gauche.singleton)
(use math.const)
(export <manipulator>)

;; PROTOTYPE IMPLEMENTATION

(define-class <manipulator> (<transient-node> <singleton-mixin>)
  ((target   :accessor target-of :init-value '())
   ;; - list of transforms.
   (mode     :accessor mode-of   :init-value #f)
   ;; - mode may be #f 'translate, 'rotate, or 'scale
   (submode  :accessor submode-of :init-value #f)
   ;; - submode for selection.  #f, 'x, 'y, 'z
   (active?  :accessor active?   :init-value #f)
   ;; - true if active
   (saved-target :accessor saved-target-of)
   ;; - list of saved attributes of the target
   (translation  :accessor translation-of :initform (make-matrix4f))
   ;; - position of the manipulator
   (rotation     :accessor rotation-of :initform (make-matrix4f))
   ;; - rotation of the manipulator
   xcolor
   ycolor
   zcolor
   scolor  ;selected
   gcolor
   tx-shape
   ty-shape
   tz-shape
   rx-shape
   ry-shape
   rz-shape
   sx-shape
   sy-shape
   sz-shape
   ))

(define-method initialize ((self <manipulator>) initargs)
  (next-method)
  (let ((xmat (matrix4f 0 -1 0 0 1 0 0 0 0 0 1 0 0 0 0 1))
        (ymat (matrix4f 1 0 0 0 0 1 0 0 0 0 1 0 0 0 0 1))
        (zmat (matrix4f 1 0 0 0 0 0 1 0 0 -1 0 0 0 0 0 1)))
    (set! (ref self 'xcolor) '(0.7 0.0 0.1 0.0))
    (set! (ref self 'ycolor) '(0.0 0.5 0.2 0.0))
    (set! (ref self 'zcolor) '(0.0 0.0 0.6 0.0))
    (set! (ref self 'scolor) '(0.7 0.7 0.0 0.0))
    (set! (ref self 'tx-shape) (make-translation-manipulator-shape xmat))
    (set! (ref self 'ty-shape) (make-translation-manipulator-shape ymat))
    (set! (ref self 'tz-shape) (make-translation-manipulator-shape zmat))
    (set! (ref self 'rx-shape) (make-rotation-manipulator-shape xmat))
    (set! (ref self 'ry-shape) (make-rotation-manipulator-shape ymat))
    (set! (ref self 'rz-shape) (make-rotation-manipulator-shape zmat))
    (set! (ref self 'sx-shape) (make-scale-manipulator-shape xmat))
    (set! (ref self 'sy-shape) (make-scale-manipulator-shape ymat))
    (set! (ref self 'sz-shape) (make-scale-manipulator-shape zmat))
    (add-hook! selection-watchers manipulator-adjust-selection)
    ))

;;============================================================
;; External protocol
;;

;;------------------------------------------------------------
;; Render & Selection
;;
(define (manipulator-render scale-factor)
  (let1 self (instance-of <manipulator>)
    (and-let* (((active? self))
               (mode (mode-of self)))
      (manipulator-update-gmatrix self)
      (gl-load-matrix (translation-of self))
      (gl-scale scale-factor scale-factor scale-factor)
      (gl-polygon-mode GL_FRONT_AND_BACK GL_FILL)
      (gl-disable GL_LIGHTING)
      (gl-enable-client-state GL_VERTEX_ARRAY)
      (let* ((smode  (submode-of self))
             (scol (ref self 'scolor))
             (xcol (if (eq? smode 'x) scol (ref self 'xcolor)))
             (ycol (if (eq? smode 'y) scol (ref self 'ycolor)))
             (zcol (if (eq? smode 'z) scol (ref self 'zcolor))))
        (case mode
          ((translate)
           (apply gl-color xcol) ((ref self 'tx-shape))
           (apply gl-color ycol) ((ref self 'ty-shape))
           (apply gl-color zcol) ((ref self 'tz-shape)))
          ((rotate)
           (gl-mult-matrix (rotation-of self))
           (apply gl-color xcol) ((ref self 'rx-shape))
           (apply gl-color ycol) ((ref self 'ry-shape))
           (apply gl-color zcol) ((ref self 'rz-shape)))
          ((scale)
           (gl-mult-matrix (rotation-of self))
           (apply gl-color xcol) ((ref self 'sx-shape))
           (apply gl-color ycol) ((ref self 'sy-shape))
           (apply gl-color zcol) ((ref self 'sz-shape)))
          ))
      )
    ))

(define (manipulator-selection-draw scale-factor)
  (let1 self (instance-of <manipulator>)
    (and-let* (((active? self))
               (mode (mode-of self)))
      (manipulator-update-gmatrix self)
      (gl-load-matrix (translation-of self))
      (gl-scale scale-factor scale-factor scale-factor)
      (gl-enable-client-state GL_VERTEX_ARRAY)
      (case mode
        ((translate)
         (gl-load-name 0) ((ref self 'tx-shape))
         (gl-load-name 1) ((ref self 'ty-shape))
         (gl-load-name 2) ((ref self 'tz-shape)))
        ((rotate)
         (gl-mult-matrix (rotation-of self))
         (gl-load-name 0) ((ref self 'rx-shape))
         (gl-load-name 1) ((ref self 'ry-shape))
         (gl-load-name 2) ((ref self 'rz-shape)))
        ((scale)
         (gl-mult-matrix (rotation-of self))
         (gl-load-name 0) ((ref self 'sx-shape))
         (gl-load-name 1) ((ref self 'sy-shape))
         (gl-load-name 2) ((ref self 'sz-shape)))
        )
      )
    ))

;;------------------------------------------------------------
;; Other operations
;;
(define (manipulator-show)
  (set! (active? (instance-of <manipulator>)) #t))

(define (manipulator-hide)
  (set! (active? (instance-of <manipulator>)) #f))

(define (manipulator-active?)
  (active? (instance-of <manipulator>)))

(define (manipulator-mode-set! mode)
  (set! (mode-of (instance-of <manipulator>)) mode))

(define (manipulator-attach target-xform . more-xforms)
  (let1 self (instance-of <manipulator>)
    (set! (target-of self) (cons target-xform more-xforms))
    ))

;; Called when scene's selection is changed.
;; If the attached target is unselected, remove it.
(define (manipulator-adjust-selection scene selections)
  (let ((self (instance-of <manipulator>))
        (new-xfs (fold-right (lambda (s r)
                               (if (is-a? s <instance>)
                                   (cons (transform-of s) r)
                                   r))
                             '()
                             selections)))
    (set! (target-of self) new-xfs)
    (set! (active? self) (not (null? new-xfs)))
    ))

;;------------------------------------------------------------
;; Selection action
;;

(define (manipulator-grab id)
  (let1 self (instance-of <manipulator>)
    (case id
      ((0) (set! (submode-of self) 'x))
      ((1) (set! (submode-of self) 'y))
      ((2) (set! (submode-of self) 'z))
      (else (set! (submode-of self) #f)))))

(define (manipulator-grabbed?)
  (not (not (submode-of (instance-of <manipulator>)))))

;; Called within GL context.  wx1, wy1, wx0, and wy0 are in widget coord.
(define manipulator-drag
  (let ((manipw (make-point4f)) ;static buffer for manipulator in window coord.
        (pos1   (make-point4f)) ;static buffer for cursor pos in world coord.
        (pos0   (make-point4f)) ;static buffer for cursor pos in world coord.
        )
    
    ;; Change the target matrix TARGET.   SELF is the manipulator instance.
    ;; MODE is manipulator mode (translate, scale or rotate).
    ;; MANIP is the center of manipulator in world coordintate.
    ;; POS1 and POS2 contains the position of current and previous pointer,
    ;; in the world coordinate.
    ;; Translation manipulator is aligned to the world coord axis, so
    ;; Scale and rotation manipulator is aligned to the local coord,
    ;; so further translation is required.
    (define (drag-a-target self target mode manip)
      (case mode
        ((translate)
         (let1 dv (point4f-sub pos1 pos0)
           (case (submode-of self)
             ((x) (inc! (ref (translation-of target) 0) (ref dv 0)))
             ((y) (inc! (ref (translation-of target) 1) (ref dv 1)))
             ((z) (inc! (ref (translation-of target) 2) (ref dv 2)))))
         (attribute-change-notify target 'translation))
        ((scale)
         (let* ((dv (point4f-sub pos1 pos0))
                (q  (quatf-conjugate (rotation-of target)))
                (wv (quatf-transform q dv)))
           (case (submode-of self)
             ((x) (inc! (ref (scale-of target) 0) (ref wv 0)))
             ((y) (inc! (ref (scale-of target) 1) (ref wv 1)))
             ((z) (inc! (ref (scale-of target) 2) (ref wv 2)))))
         (attribute-change-notify target 'scale))
        ((rotate)
         (let* ((dv1 (point4f-sub pos1 manip))
                (dv0 (point4f-sub pos0 manip))
                (q   (quatf-conjugate (rotation-of target)))
                (wv1 (quatf-transform q dv1))
                (wv0 (quatf-transform q dv0)))
           (case (submode-of self)
             ((x)
              (let* ((t (vector4f-normalize!
                         (vector4f 0 (ref wv1 1) (ref wv1 2))))
                     (s (vector4f-normalize!
                         (vector4f 0 (ref wv0 1) (ref wv0 2))))
                     (e (vector4f-dot s t))
                     (sxt (vector4f-cross s t))
                     (sint (if (> e 0.9999) 0 (sqrt (/ (- 1 e) 2))))
                     (cost (if (< e -0.9999) 0 (sqrt (/ (+ 1 e) 2))))
                     (q (quatf (if (positive? (ref sxt 0)) sint (- sint))
                               0 0 cost)))
                (quatf-mul! (rotation-of target) q)))
             ((y)
              (let* ((t (vector4f-normalize!
                         (vector4f (ref wv1 0) 0 (ref wv1 2))))
                     (s (vector4f-normalize!
                         (vector4f (ref wv0 0) 0 (ref wv0 2))))
                     (e (vector4f-dot s t))
                     (sxt (vector4f-cross s t))
                     (sint (if (> e 0.9999) 0 (sqrt (/ (- 1 e) 2))))
                     (cost (if (< e -0.9999) 0 (sqrt (/ (+ 1 e) 2))))
                     (q (quatf 0
                               (if (positive? (ref sxt 1)) sint (- sint))
                               0 cost)))
                (quatf-mul! (rotation-of target) q)))
             ((z)
              (let* ((t (vector4f-normalize!
                         (vector4f (ref wv1 0) (ref wv1 1) 0)))
                     (s (vector4f-normalize!
                         (vector4f (ref wv0 0) (ref wv0 1) 0)))
                     (e (vector4f-dot s t))
                     (sxt (vector4f-cross s t))
                     (sint (if (> e 0.9999) 0 (sqrt (/ (- 1 e) 2))))
                     (cost (if (< e -0.9999) 0 (sqrt (/ (+ 1 e) 2))))
                     (q (quatf 0
                               0
                               (if (positive? (ref sxt 2)) sint (- sint))
                               cost)))
                (quatf-mul! (rotation-of target) q)))
             ))
         (attribute-change-notify target 'rotation))
        ))

    ;; Unproject mouse pointer to world coord, on the plane that is
    ;; perpendicular to the camera direction and crosses the point
    ;; of the center of manipulator.  MANIPW is the center of manipulator
    ;; projected to the window coord; it is only used to obtain z value.
    ;; The result of unprojection of current pointer position and
    ;; previous pointer position are stored in pos1 and pos0, respectively.
    ;; Returns the center of manipulator in world coord.
    (define (unproject self wx1 wy1 wx0 wy0)
      (let ((vp        (gl-get-integer GL_VIEWPORT))
            (model-mat '#f64(1 0 0 0 0 1 0 0 0 0 1 0 0 0 0 1))
            (proj-mat  (gl-get-double GL_PROJECTION_MATRIX))
            (manip     (matrix4f-column/shared (translation-of self) 3)))
        (glu-project! manipw manip model-mat proj-mat vp)
        (glu-un-project! pos1
                         (vector4f wx1 (- (ref vp 3) wy1 1) (ref manipw 2))
                         model-mat proj-mat vp)
        (glu-un-project! pos0
                         (vector4f wx0 (- (ref vp 3) wy0 1) (ref manipw 2))
                         model-mat proj-mat vp)
        manip))

    ;; Body of manipulator-drag.
    ;;   (wx1, wy1) - current mouse pointer in window coord
    ;;   (wx0, wy0) - previous mouse pointer in window coord
    (lambda (wx1 wy1 wx0 wy0)
      (let* ((self      (instance-of <manipulator>))
             (target    (target-of self))
             (mode      (mode-of self)))
        (when (and (not (null? target)) mode)
          (for-each (cute drag-a-target self <> mode
                          (unproject self wx1 wy1 wx0 wy0))
                    target))
        ))
    ))

;;============================================================
;; Internal stuff
;;

;; Update gmatrix
;;   For the time being, manipulator is positioned to the first selected
;;   transform.
(define (manipulator-update-gmatrix self)
  (unless (null? (target-of self))
    (let ((t (make-vector4f))
          (h (make-vector4f))
          (s (make-vector4f)))
      (matrix4f-decompose! (gmatrix-of (car (target-of self)))
                           t
                           (rotation-of self)
                           h s)
      (translation->matrix4f! (translation-of self) t)
      )))

;; creates a 3D arrow for translation manipulator
(define make-translation-manipulator-shape
  (let* ((arrow-divs 8)                  ;# of divs
         (arrow-length 0.6)
         (arrow-width 0.2)
         (theta  (lambda (n) (/ (* -2 pi n) arrow-divs)))
         (points (list* (point4f 0 0 0)
                        (point4f 0 arrow-length 0)
                        (map (lambda (n)
                               (point4f (* (sin (theta n)) arrow-width)
                                        0
                                        (* (cos (theta n)) arrow-width)))
                             (iota arrow-divs))))
         (face0  (make-u32vector (+ arrow-divs 2)))
         (face1  (make-u32vector (+ arrow-divs 2))))
    (let1 cnt 1
      (set! (ref face0 0) 0)
      (set! (ref face1 0) 1)
      (dotimes (n arrow-divs)
        (set! (ref face0 cnt) (+ n 2))
        (set! (ref face1 cnt) (+ n 2))
        (inc! cnt))
      (set! (ref face0 cnt) 2)
      (set! (ref face1 cnt) 2))
    ;; Body of make-translation-manipulator-shape
    (lambda (mat)
      (let ((vertex  (make-point4f-array (length points)))
            (lineptr (* mat (point4f 0 1.2 0))))
        (for-each (lambda (point ind)
                    (set! (ref vertex ind)
                          (* mat
                             (* (matrix4f 1 0 0 0 0 1 0 0 0 0 1 0 0 1.2 0 1)
                                point))))
                  points
                  (iota (length points)))
        (lambda ()
          (gl-vertex-pointer 4 vertex)
          (gl-draw-elements GL_TRIANGLE_FAN face0)
          (gl-draw-elements GL_TRIANGLE_FAN face1)
          (gl-line-width 3.0)
          (gl-begin* GL_LINES
            (gl-vertex '#f32(0.0 0.0 0.0))
            (gl-vertex lineptr))
          (gl-line-width 1.0)
          )))
    ))

;; creates a circle
(define make-rotation-manipulator-shape
  (let* ((circle-divs 32)
         (circle-radius 2.5)
         (theta (lambda (n) (/ (* -2 pi n) circle-divs)))
         (circle-points   (map (lambda (n)
                                 (point4f (* (sin (theta n)) circle-radius)
                                          0
                                          (* (cos (theta n)) circle-radius)))
                               (iota circle-divs)))
         (half   (/ circle-divs 2))
         )
    ;; body of make-rotation-manipulator-shape
    (lambda (mat)
      (let ((vertex (make-point4f-array circle-divs))
            (line   (make-u32vector circle-divs)))
        (for-each (lambda (point ind)
                    (set! (ref vertex ind) (* mat point)))
                  circle-points
                  (iota circle-divs))
        (dotimes (n circle-divs) (set! (ref line n) n))
        (lambda ()
          (gl-vertex-pointer 4 vertex)
          (gl-draw-elements GL_LINE_LOOP line)
          )))
    ))

;; creates an offset cube (for scale manipulator)
(define make-scale-manipulator-shape
  (let* ((size   0.2)
         (cube-points (list (point4f size size size)
                            (point4f size (- size) size)
                            (point4f size (- size) (- size))
                            (point4f size size (- size))
                            (point4f (- size) size size)
                            (point4f (- size) (- size) size)
                            (point4f (- size) (- size) (- size))
                            (point4f (- size) size (- size))))
         (cube-faces (u32vector 0 1 2 3
                                0 3 7 4
                                0 4 5 1
                                3 2 6 7
                                7 6 5 4
                                1 2 6 5))
         (off-mat    (matrix4f 1 0 0 0 0 1 0 0 0 0 1 0 0 1.4 0 1))
         )
    ;; body of make-rotation-manipulator-shape
    (lambda (mat)
      (let ((vertex (make-point4f-array (length cube-points)))
            (lineptr (point4f->f32vector (* mat (point4f 0 1.4 0)))))
        (for-each (lambda (point ind)
                    (set! (ref vertex ind) (* mat (* off-mat point))))
                  cube-points
                  (iota (length cube-points)))
        (lambda ()
          (gl-vertex-pointer 4 vertex)
          (gl-draw-elements GL_QUADS cube-faces)
          (gl-color 0.3 0.3 0.3 0.0)
          (gl-begin* GL_LINES
            (gl-vertex '#f32(0.0 0.0 0.0))
            (gl-vertex lineptr))
          )))
    ))

(provide "banyan/gui/manipulator")
