;;;
;;; banyan/gui/nodeselector - node selection
;;;
;;;  Copyright(C) 2002-2003 by Shiro Kawai (shiro@acm.org)
;;;
;;;  Permission to use, copy, modify, distribute this software and
;;;  accompanying documentation for any purpose is hereby granted,
;;;  provided that existing copyright notices are retained in all
;;;  copies and that this notice is included verbatim in all
;;;  distributions.
;;;  This software is provided as is, without express or implied
;;;  warranty.  In no circumstances the author(s) shall be liable
;;;  for any damages arising out of the use of this software.
;;;
;;;  $Id: nodeselector.scm,v 1.9 2003/01/01 00:59:50 shirok Exp $
;;;

(define-module banyan.gui.nodeselector
  (use srfi-1)
  (use srfi-2)
  (use gauche.mop.instance-pool)
  (extend banyan.gui)
  (export node-selector-activate!))
(select-module banyan.gui.nodeselector)

;; A UI to select a node

;; Number of nodes can potentially be very big, so the interface
;; needs a good filtering mechanism.
;;
;; The selector provides three criteria of filtering: part, class,
;; and name.   One or more of those criteria can be fixed by the
;; caller of the dialog.
;;
;; The caller can also specify if it allows the user to create
;; a new node or not.

;;==================================================================
;; Node selector
;;

(define-glim-application-frame <node-selector-dialog> ()
  ((required-class      :init-value <node>)
   (selection-callback  :init-value #f)
   (enable-part-change  :init-value #f)
   (enable-all-parts    :init-value #f)
   (enable-creation     :init-value #f)
   )
  (:panes
   (class-filter (make-class-filter .frame.))
   (part-filter  (make-part-filter .frame.))
   (list-pane    (make-list-pane .frame.))
   (create-node  (make <glim-push-button>
                   :label "Create new node"
                   :activate-callback (lambda _ (create-node .frame.))))
   )
  (:layouts
   (default
     (glim-vertically (:height 250)
       (glim-tabling (:spacing 2
                      :column-options '((0 :fill)
                                        ($ :fill :expand)))
         (list (make <glim-label-pane> :text "Class:" :align 'left)
               (glim-get-frame-pane .frame. 'class-filter))
         (list (make <glim-label-pane> :text "Part:" :align 'left)
               (glim-get-frame-pane .frame. 'part-filter)))
       (glim-spacing (:thickness 2)
         (make <glim-label-pane> :text "Select node:" :align 'left))
       (:expand (glim-spacing (:thickness 2) list-pane))
       (glim-spacing (:thickness 4) create-node)
       (----)
       (dialog-buttons
        (dialog-action-gadget "OK" 'ok)
        (dialog-action-gadget "Apply" 'apply)
        (dialog-action-gadget "Close" 'close))
       )))
  )

(define-method initialize ((self <node-selector-dialog>) initargs)
  (next-method self
               (list* :name 'node-selector
                      :pretty-name "Node selector"
                      initargs))
  )

;; node-selector-activate! callback
;;    &keyword title class enable-class-change
;;             enable-part-change enable-all-parts
;;             enable-name-filter
;;             enable-creation
(define (node-selector-activate! callback . args)
  (let-keywords* args ((class <node>)
                       (enable-class-change #t)
                       (enable-part-change #t)
                       (enable-all-parts #t)
                       (enable-name-filter #t)
                       (enable-creation #t))
    (let* ((self (instance-of <node-selector-dialog>))
           (class-filter (glim-get-frame-pane self 'class-filter))
           )
      (set! (ref self 'required-class) class)
      (set! (ref self 'enable-all-parts)   enable-all-parts)
      (set! (ref self 'enable-part-change) enable-part-change)
      (set! (ref self 'enable-creation)    enable-creation)
      (set! (ref self 'selection-callback)  callback)
      (class-filter-set-class self class)
      (part-filter-set-part self (current-part))
      (glim-run-frame-top-level self))))

;;------------------------------------------------------------------
;; Actions
;;

;; Dialog action
(define-method dialog-action ((self <node-selector-dialog>) action)
  (and-let* (((memq action '(ok apply)))
             (callback (ref self 'selection-callback)))
    (let ((listp (glim-get-frame-pane self 'list-pane)))
      (callback (glim-gadget-value listp))))
  (when (and (memq action '(ok close)))
    (glim-frame-exit self)))

;; create node
(define (create-node dialog)
  (let ((class (ref dialog 'required-class)))
    (node-creator-activate!
     (lambda (class name)
       (let ((new-node
              (make class :name (get-unique-name (current-part) name)))
             (callback (ref dialog 'selection-callback)))
         (callback new-node)
         (glim-frame-exit dialog)))
     class)))

;;------------------------------------------------------------------
;; Class filter
;;

(define (make-class-filter dialog)
  (make <glim-option-pane>
    :name-key (lambda (x) (x->string (class-name x)))
    :value-changed-callback (lambda (g v) (refresh-list dialog))))

(define (class-filter-set-class dialog class)
  (let ((f (glim-get-frame-pane dialog 'class-filter)))
    (set! (ref f 'items) (node-meta-all-subclasses class))))

(define (class-filter-value dialog)
  (glim-gadget-value (glim-get-frame-pane dialog 'class-filter)))

;;------------------------------------------------------------------
;; Part filter
;;

(define (make-part-filter dialog)
  (make <glim-option-pane>
    :name-key (lambda (p) (if (is-a? p <part>) (part-name p) "*"))
    :value-changed-callback (lambda (g v) (refresh-list dialog))))

(define (part-filter-set-part dialog part)
  (let ((pane  (glim-get-frame-pane dialog 'part-filter))
        (all-parts (instance-pool->list <part>)))
    (set! (ref pane 'items)
          (if (ref dialog 'enable-all-parts)
              (cons #t all-parts)
              all-parts))
    (set! (glim-gadget-value pane) part)))

(define (part-filter-value dialog)
  (glim-gadget-value (glim-get-frame-pane dialog 'part-filter)))

;;------------------------------------------------------------------
;; List pane
;;

(define (make-list-pane dialog)
  (make <glim-list-pane> :name-key name-of))

(define refresh-list
  ;; Oof, non-MT safe!  but effective..
  (let ((oclass #f) (opart #f))
    (lambda (dialog)
      (let ((class (class-filter-value dialog))
            (part  (part-filter-value dialog)))
        (when (and class part
                   (or (not (eq? oclass class))
                       (not (eq? opart part))))
          (let* ((listp (glim-get-frame-pane dialog 'list-pane))
                 (nodes (filter (cut is-a? <> class)
                                (append-map part-nodes
                                            (if (is-a? part <part>)
                                                (list part)
                                                (instance-pool->list <part>)))))
                 )
            (set! (ref listp 'items) nodes)))
        ))
    ))

(provide "banyan/gui/nodeselector")
