;;;
;;; banyan/gui/partbrowser - part browser
;;;
;;;  Copyright(C) 2002-2003 by Shiro Kawai (shiro@acm.org)
;;;
;;;  Permission to use, copy, modify, distribute this software and
;;;  accompanying documentation for any purpose is hereby granted,
;;;  provided that existing copyright notices are retained in all
;;;  copies and that this notice is included verbatim in all
;;;  distributions.
;;;  This software is provided as is, without express or implied
;;;  warranty.  In no circumstances the author(s) shall be liable
;;;  for any damages arising out of the use of this software.
;;;
;;;  $Id: partbrowser.scm,v 1.18 2003/02/22 23:19:07 shirok Exp $
;;;

(define-module banyan.gui.partbrowser
  (use srfi-1)
  (use srfi-2)
  (use gauche.mop.instance-pool)
  (use gauche.parameter)
  (extend banyan.gui)
  (export <part-browser>))
(select-module banyan.gui.partbrowser)

;; Part browser - browse parts in the scene/repository

;;==================================================================
;; Part browser main window
;;

(define-glim-application-frame <part-browser> ()
  ()
  (:panes
   (display  (make <part-display>))
   (treeview (make <part-browser-tree-pane> :parent .frame.))
   )
  (:layouts
   (default
     (glim-vertically (:width 400)
       (glim-outlining (:style :none :thickness 6)
         display)
       (:expand
        (glim-horizontally (:thickness 4)
          (:expand
           (glim-outlining (:label "Parts in the scene" :thickness 4)
             (glim-scrolling (:scroll-bar :both :width 150 :height 150
                              :thickness 4)
               treeview)))
          (glim-vertically ()
            (make <glim-push-button> :label "New Part"
                  :activate-callback (cut new-part .frame. <>))
            (make <glim-push-button> :label "Edit Part information"
                  :activate-callback (cut edit-part .frame. <>))
            (make <glim-push-button> :label "Make Part Current"
                  :activate-callback (cut select-part .frame. <>))
            (make <glim-push-button> :label "Select Part's Nodes"
                  :activate-callback (cut select-nodes-in-part .frame. <>))
            (make <glim-push-button> :label "Make Part Visible"
                  :activate-callback (cut make-part-visibile  .frame. <>))
            (make <glim-push-button> :label "Make Part Invisible"
                  :activate-callback (cut make-part-invisibile  .frame. <>))
            ;(----)
            ;(make <glim-push-button> :label "Delete Part"
            ;      :activate-callback (cut dummy-part .frame. <>))
            )))
       (----)
       (dialog-buttons
        (dialog-action-gadget "Close" 'ok))
       )))
  )

(define-method initialize ((self <part-browser>) initargs)
  (next-method self
               (list* :name 'part-browser
                      :pretty-name "Part Browser"
                      initargs))
  (parameter-observer-add! current-scene
                           (lambda _ (refresh-view self)))
  (add-hook! part-watchers (lambda _ (refresh-view self)))
  )

(define-method refresh-view ((self <part-browser>))
  (refresh-view (glim-get-frame-pane self 'treeview)))

;;------------------------------------------------------------------
;; Actions
;;

(define-method dialog-action ((self <part-browser>) action)
  (glim-frame-exit self))

(define-method new-part ((self <part-browser>) gadget)
  (part-editor-activate! #f
                         (lambda (part)
                           (add-part (current-scene) part))))

(define-method edit-part ((self <part-browser>) gadget)
  (and-let* ((parts (part-browser-selected-parts
                     (glim-get-frame-pane self 'treeview)))
             ((not (null? parts))))
    (part-editor-activate! (car parts)
                           (lambda (part) #f))))

(define-method select-part ((self <part-browser>) gadget)
  (let ((parts (part-browser-selected-parts (glim-get-frame-pane self 'treeview))))
    (unless (null? parts)
      (make-part-current (current-scene) (car parts)))
    ))

(define-method select-nodes-in-part ((self <part-browser>) gadget)
  (let* ((parts (part-browser-selected-parts (glim-get-frame-pane self 'treeview)))
         (instances (append-map (lambda (p)
                                  (filter (cut is-a? <> <instance>)
                                          (part-nodes p)))
                                parts)))
    (reset-selections (current-scene))
    (apply set-selection (current-scene) instances)
    ))

(define-method make-part-visibile ((self <part-browser>) gadget)
  (for-each (cut slot-set! <> 'visibility #t)
            (part-browser-selected-parts (glim-get-frame-pane self 'treeview)))
  (viewports-post-redisplay)
  )

(define-method make-part-invisibile ((self <part-browser>) gadget)
  (for-each (cut slot-set! <> 'visibility #f)
            (part-browser-selected-parts (glim-get-frame-pane self 'treeview)))
  (viewports-post-redisplay)
  )

(define-method dummy-part ((self <part-browser>) gadget)
  (print (part-browser-selected-parts (glim-get-frame-pane self 'treeview))))

;;------------------------------------------------------------------
;; Treeview
;;

(define-class <part-browser-tree-pane> (<glim-gadget>)
  ((parent :init-keyword :parent)
   (store)
   ))

(define-method initialize ((self <part-browser-tree-pane>) initargs)
  ;; Usage of tree store columns:
  ;;   0: the basename of the part (displayed)
  ;;   1: the fullname of the part
  (let* ((store (gtk-tree-store-new <string> <string>))
         (view  (gtk-tree-view-new-with-model store))
         (col0  (gtk-tree-view-column-new-with-attributes
                 "name" (gtk-cell-renderer-text-new) "text" 0))
         )
    (gtk-tree-view-append-column view col0)
    (gtk-tree-view-set-headers-visible view #f)
    (g-signal-connect view "row-activated"
                      (lambda (w p i) (print (gtk-tree-path-get-indices p))))
    (set! (widget-of self) view)
    (set! (ref self 'store) store)
    (next-method)
    (refresh-view self)))

(define-method refresh-view ((self <part-browser-tree-pane>))
  (let* ((view  (widget-of self))
         (store (ref self 'store))
         (parts (ref (current-scene) 'parts))
         (current-name (part-name (current-part)))
         (p-alist '())
         (selected-path #f))

    (define (add-part name)
      (let* ((dir  (sys-dirname name)))
        (cond ((assoc name p-alist) => cadr)
              ((equal? dir ".")
               (let1 iter (gtk-tree-store-append store #f)
                 (gtk-tree-store-set-value store iter 0 name)
                 (gtk-tree-store-set-value store iter 1 name)
                 (push! p-alist (list name iter))
                 (when (equal? name current-name)
                   (set! selected-path (gtk-tree-model-get-path store iter)))
                 iter))
              (else
               (let1 iter (gtk-tree-store-append store (add-part dir))
                 (gtk-tree-store-set-value store iter 0 (sys-basename name))
                 (gtk-tree-store-set-value store iter 1 name)
                 (push! p-alist (list name iter))
                 (when (equal? name current-name)
                   (set! selected-path (gtk-tree-model-get-path store iter)))
                 iter)))
        ))
    (gtk-tree-store-clear store)
    (for-each (lambda (p) (add-part (part-name p))) parts)
    (gtk-tree-view-expand-all view)
    (when selected-path
      (gtk-tree-view-set-cursor view selected-path #f #f))
    ))

(define-method part-browser-selected-parts ((self <part-browser-tree-pane>))
  (and-let* ((sel (gtk-tree-view-get-selection (widget-of self)))
             (store (ref self 'store))
             (paths '()))
    (gtk-tree-selection-selected-foreach sel
                                         (lambda (mod path iter)
                                           (push! paths path)))
    (fold (lambda (path r)
            (let* ((iter (gtk-tree-model-get-iter store path))
                   (name (gtk-tree-model-get-value store iter 1))
                   (part (instance-pool-find <part>
                                             (lambda (p)
                                               (equal? (part-name p) name)))))
              (if part (cons part r) r)))
          '()
          paths)
    ))

(provide "banyan/gui/partbrowser")
