;;;
;;; banyan/gui/parteditor - part editor
;;;
;;;  Copyright(C) 2002-2003 by Shiro Kawai (shiro@acm.org)
;;;
;;;  Permission to use, copy, modify, distribute this software and
;;;  accompanying documentation for any purpose is hereby granted,
;;;  provided that existing copyright notices are retained in all
;;;  copies and that this notice is included verbatim in all
;;;  distributions.
;;;  This software is provided as is, without express or implied
;;;  warranty.  In no circumstances the author(s) shall be liable
;;;  for any damages arising out of the use of this software.
;;;
;;;  $Id: parteditor.scm,v 1.18 2003/02/20 14:20:28 shirok Exp $
;;;

(define-module banyan.gui.parteditor
  (use srfi-1)
  (use srfi-2)
  (use gauche.parameter)
  (extend banyan.gui)
  (export part-editor-activate!))
(select-module banyan.gui.parteditor)

;; Part editor - view/edit info about a part
;;  Use part-editor-activate! to pop the dialog up.

;;==================================================================
;; Part editor main window
;;

(define-glim-application-frame <part-editor> ()
  ((target-part)
   (apply-callback))
  (:panes
   (name-entry (make <part-editor-name-entry> :parent .frame.))
   (info-pane  (make <part-editor-info-pane>))
   (description (make <part-editor-description>))
   )
  (:layouts
   (default
     (glim-vertically ()
       (glim-outlining (:thickness 4 :style :none)
         name-entry)
       (:expand
        (glim-horizontally (:thickness 4)
          info-pane
          (:expand
           (glim-outlining (:label "Description:" :thickness 5 :width 300)
             (glim-spacing (:thickness 5)
               description)))
          ))
       (----)
       (dialog-buttons
        (dialog-action-gadget "OK" 'ok)
        (dialog-action-gadget "Apply" 'apply)
        (dialog-action-gadget "Cancel" 'cancel))
       )))
  )

(define-method initialize ((self <part-editor>) initargs)
  (next-method self
               (list* :name 'part-editor
                      :pretty-name "Part Editor"
                      initargs))
  (parameter-observer-add! current-scene (lambda _ (refresh-view self)))
  (add-hook! part-watchers (lambda _ (refresh-view self)))
  )

;; External interface
;;   To create a new part, pass #f to target-part.
(define (part-editor-activate! target-part apply-callback)
  (let1 self (instance-of <part-editor>)
    (set! (ref self 'target-part) target-part)
    (set! (ref self 'apply-callback) apply-callback)
    (refresh-view self)
    (glim-run-frame-top-level self)))

;;------------------------------------------------------------------
;; Action
;;

(define-method dialog-action ((self <part-editor>) action)
  (case action
    ((ok apply)
     (let ((name (glim-gadget-value (glim-get-frame-pane self 'name-entry)))
           (desc (glim-gadget-value (glim-get-frame-pane self 'description)))
           (part (ref self 'target-part)))
       (cond
        ((not (valid-part-name? name))
         (error-dialog "part name contains invalid character: ~s" name))
        ;; TODO uniqueness check
        (else
         (if part
             (set! (part-name part) name)
             (begin (set! part (make <part> :name name))
                    (set! (ref self 'target-part) part)))
         (set! (ref part 'description) desc)
         (cond ((ref self 'apply-callback) => (cut <> part)))
         (when (eq? action 'ok) (glim-frame-exit self)))
        )))
    ((cancel)
     (glim-frame-exit self))
    ))

;;==================================================================
;; Panes
;;

;;------------------------------------------------------------------
;; <part-editor-name-entry>
;;
(define-class <part-editor-name-entry> (<glim-value-gadget>)
  ((parent :init-keyword :parent)
   (entry)
   ))

(define-method initialize ((self <part-editor-name-entry>) initargs)
  (let* ((box   (gtk-hbox-new #f 0))
         (lab   (gtk-label-new "Part name: "))
         (entry (gtk-entry-new)))
    (gtk-box-pack-start box lab #f #f 4)
    (gtk-box-pack-start box entry #t #t 4)
    (set! (widget-of self) box)
    (set! (ref self 'entry) entry)
    (next-method)
    ))

(define-method glim-gadget-value ((self <part-editor-name-entry>))
  (gtk-entry-get-text (ref self 'entry)))

(define-method (setter glim-gadget-value) ((self <part-editor-name-entry>) value)
  (gtk-entry-set-text (ref self 'entry) (x->string value)))

(define-method editable? ((self <part-editor-name-entry>))
  (gtk-entry-get-editable (ref self 'entry)))

(define-method (setter editable?) ((self <part-editor-name-entry>) v)
  (gtk-entry-set-editable (ref self 'entry) v))

;;------------------------------------------------------------------
;; <part-editor-info-pane>
;;
(define-class <part-editor-info-pane> (<glim-gadget>)
  ((parent :init-keyword :parent)
   (labels)                             ;alist of (tag . gtk-label)
   ))

(define-method initialize ((self <part-editor-info-pane>) initargs)
  (define (label-new str)
    (let1 l (gtk-label-new str)
      (gtk-misc-set-alignment l 0 0.5)
      l))

  (let* ((tags   '((revision    "Revision: ")
                   (modified-by "Last modified by: ")
                   (modified-at "Last modified at: ")
                   (created-by  "Created by: ")
                   (created-at  "Created at: ")
                   (license     "License: ")
                   ))
         (tab    (gtk-table-new (length tags) 2 #f))
         (labels (map (lambda (tag) (cons (car tag) (label-new "----")))
                      tags))
         )
    (set! (ref self 'labels) labels)
    (for-each (lambda (tag label ind)
                (gtk-table-attach tab (label-new (cadr tag))
                                  0 1 ind (+ ind 1) GTK_FILL GTK_FILL 2 2)
                (gtk-table-attach tab (cdr label)
                                  1 2 ind (+ ind 1) GTK_FILL GTK_FILL 2 2))
              tags labels (iota (length tags)))
    (set! (widget-of self) tab)
    (next-method)
    ))

(define-method part-editor-info-set ((self <part-editor-info-pane>) part)
  (define (get-label tag)
    (cdr (assq tag (ref self 'labels))))
  (gtk-label-set-text (get-label 'revision)
                      (or (and-let* ((part)
                                     (rev (ref part 'revision)))
                            (string-join (map x->string rev) "."))
                          "----"))
  (gtk-label-set-text (get-label 'modified-by)
                      (if part (x->string (ref part 'modified-by)) "----"))
  (gtk-label-set-text (get-label 'modified-at)
                      (or (and-let* ((part)
                                     (time (ref part 'modification-time)))
                            (sys-strftime "%Y/%m/%d %H:%M:%S"
                                          (sys-localtime time)))
                          "----"))
  (gtk-label-set-text (get-label 'created-by)
                      (if part (x->string (ref part 'created-by)) "----"))
  (gtk-label-set-text (get-label 'created-at)
                      (or (and-let* ((part)
                                     (time (ref part 'creation-time)))
                            (sys-strftime "%Y/%m/%d %H:%M:%S"
                                          (sys-localtime time)))
                          "----"))
  (gtk-label-set-text (get-label 'license)
                      (or (and-let* ((part)
                                     (lic (part-license part)))
                            (ref lic 'cc-license-name))
                          "----"))
  )

;;------------------------------------------------------------------
;; <part-editor-description>
;;
(define-class <part-editor-description> (<glim-value-gadget>)
  ((parent :init-keyword :parent)
   ))

(define-method initialize ((self <part-editor-description>) initargs)
  (let* ((view  (gtk-text-view-new))
         (buf   (gtk-text-view-get-buffer view)))
    (gtk-text-buffer-set-text buf "" -1)
    (set! (widget-of self) view)
    (next-method)
    ))

(define-method glim-gadget-value ((self <part-editor-description>))
  (let* ((buf   (gtk-text-view-get-buffer (widget-of self)))
         (start (gtk-text-buffer-get-start-iter buf))
         (end   (gtk-text-buffer-get-end-iter buf))
         )
    (gtk-text-buffer-get-text buf start end #f)))

(define-method (setter glim-gadget-value) ((self <part-editor-description>) value)
  (when (string? value)
    (let ((buf (gtk-text-view-get-buffer (widget-of self))))
      (gtk-text-buffer-set-text buf value (string-size value)))))

;;==================================================================
;; Refresh
;;

(define-method refresh-view ((self <part-editor>))
  (let ((entry (glim-get-frame-pane self 'name-entry))
        (info  (glim-get-frame-pane self 'info-pane))
        (desc  (glim-get-frame-pane self 'description)))
    (or (and-let* ((part (ref self 'target-part)))
          (set! (glim-gadget-value entry) (part-name part))
          (set! (glim-gadget-value desc) (ref part 'description))
          (set! (editable? entry) (if (ref part 'revision) #f #t))
          (part-editor-info-set info part)
          #t)
        (begin
          (set! (glim-gadget-value entry) "")
          (set! (glim-gadget-value desc) "")
          (set! (editable? entry) #t)
          (part-editor-info-set info #f))))
  )

(provide "banyan/gui/parteditor")

