;;;
;;; banyan/gui/shapedialog - shape-creator dialogs
;;;
;;;  Copyright(C) 2002-2003 by Shiro Kawai (shiro@acm.org)
;;;
;;;  Permission to use, copy, modify, distribute this software and
;;;  accompanying documentation for any purpose is hereby granted,
;;;  provided that existing copyright notices are retained in all
;;;  copies and that this notice is included verbatim in all
;;;  distributions.
;;;  This software is provided as is, without express or implied
;;;  warranty.  In no circumstances the author(s) shall be liable
;;;  for any damages arising out of the use of this software.
;;;
;;;  $Id: shapedialog.scm,v 1.12 2003/12/17 10:50:09 shirok Exp $
;;;

(define-module banyan.gui.shapedialog
  (use srfi-1)
  (use banyan.gui)
  (export <cube-creator-dialog>
          <sphere-creator-dialog>
          <cylinder-creator-dialog>
          <cone-creator-dialog>
          <plane-creator-dialog>
          <lambda-creator-dialog>
          <procedural-creator-dialog>
          <fracterrain-creator-dialog>
          <lsystem-creator-dialog>
          ))
(select-module banyan.gui.shapedialog)

;; PROTOTYPE IMPLEMENTATION

(define-macro (define-shape-creator-dialog class name pretty-name
                parameters creator)
  (define value-types '((spin   . <glim-spin-button>)
                        (option . <glim-option-pane>)))
  `(begin
     (define-glim-application-frame ,class ()
       ()
       (:panes
        ,@(map (lambda (param)
                 (let* ((name  (car param))
                        (type  (caddr param))
                        (inits (cdddr param))
                        (class (cdr (or (assq type value-types)
                                        (error "unknown widget value type" type)))))
                   `(,name (make ,class ,@inits))))
               parameters))
       (:layouts
        (default
          (glim-vertically (:spacing 4)
            (glim-tabling (:spacing 4
                           :column-options '((0 :fill)
                                             ($ :expand :fill)))
              ,@(map (lambda (param)
                       `(list (make <glim-label-pane>
                                :text ,(string-append (cadr param) ": ")
                                :align 'left)
                              (glim-get-frame-pane .frame. ',(car param))))
                     parameters))
            (----)
            (dialog-buttons
             (dialog-action-gadget "OK" 'ok)
             (dialog-action-gadget "Cancel" 'cancel)))
          )
        ))
     (define-method initialize ((self ,class) initargs)
       (next-method self
                    (list* :name ',name
                           :pretty-name ',pretty-name
                           initargs))
       )
     (define-method dialog-action ((self ,class) action)
       (when (memq action '(ok 'apply))
         (,creator
          ,@(append-map (lambda (params)
                          `(,(make-keyword (car params))
                            (glim-gadget-value (glim-get-frame-pane self
                                                                    ',(car params)))))
                        parameters)))
       (when (memq action '(ok cancel))
         (glim-frame-exit self))
       )
     )
  )

;;==========================================================
;; Cube creator
;;

(define-shape-creator-dialog <cube-creator-dialog>
  cube-creator "Create Cube"
  ((size "Size" spin :min 0.001 :max 100.0 :step 0.01 :value 2 :digits 2)
   )
  make-cube)

;;==========================================================
;; Sphere creator
;;

(define-shape-creator-dialog <sphere-creator-dialog>
  sphere-creator "Create Sphere"
  ((radius "Radius"            spin :min 0.001 :max 100 :step 0.01
                                    :value 1 :digits 2)
   (t-div  "Division in theta" spin :min 3 :max 256 :step 1 :value 20)
   (p-div  "Division in phi"   spin :min 2 :max 256 :step 1 :value 10)
   )
  make-sphere)

;;==========================================================
;; Cylinder creator
;;

(define-shape-creator-dialog <cylinder-creator-dialog>
  cylinder-creator "Create Cylinder"
  ((radius "Radius"            spin :min 0.001 :max 100 :step 0.01
                                    :value 1 :digits 2)
   (height "Height"            spin :min 0.001 :max 100 :step 0.01
                                    :value 2 :digits 2)
   (t-div  "Division in theta" spin :min 3 :max 256 :step 1 :value 20)
   (p-div  "Division in y"     spin :min 1 :max 256 :step 1 :value 10)
   )
  make-cylinder)

;;==========================================================
;; Cone creator
;;

(define-shape-creator-dialog <cone-creator-dialog>
  cone-creator "Create Cone"
  ((radius "Radius"            spin :min 0.001 :max 100 :step 0.01
                                    :value 1 :digits 2)
   (height "Height"            spin :min 0.001 :max 100 :step 0.01
                                    :value 2 :digits 2)
   (t-div  "Division in theta" spin :min 3 :max 256 :step 1 :value 20)
   (p-div  "Division in y"     spin :min 1 :max 256 :step 1 :value 10)
   )
  make-cone)

;;==========================================================
;; Plane creator
;;

(define-shape-creator-dialog <plane-creator-dialog>
  plane-creator "Create Plane"
  ((x-size "Size X"        spin :min 0.01 :max 100 :step 0.1
                                :value 2 :digits 2)
   (z-size "Size Z"        spin :min 0.01 :max 100 :step 0.1
                                :value 2 :digits 2)
   (x-div  "Division in x" spin :min 1 :max 256 :step 1 :value 5)
   (z-div  "Division in z" spin :min 1 :max 256 :step 1 :value 5)
   )
  make-plane)

;;==========================================================
;; Procedural shape creators (temporary interface)
;;

(define-glim-application-frame <procedural-creator-dialog> ()
  () 
  (:panes
   (selector (make <glim-option-pane>
               :name-key car :value-key cdr
               :items `(("Lsystem plant" . ,<lsystem-creator-dialog>)
                        ("Fractal terrain" . ,<fracterrain-creator-dialog>)
                        ))
             ))
  (:layouts
   (default
     (glim-vertically ()
       (glim-vertically (:spacing 2)
         (make <glim-label-pane> :text "Procedural Shape Creators")
         selector)
       (----)
       (dialog-buttons
        (dialog-action-gadget "Open" 'open)
        (dialog-action-gadget "Close" 'close))
       )))
  )

(define-method initialize ((self <procedural-creator-dialog>) initargs)
  (next-method self
               (list* :name 'procedural-creator
                      :pretty-name "Procedural shape creator"
                      initargs))
  )

(define-method dialog-action ((self <procedural-creator-dialog>) action)
  (case action
    ((open)
     (let* ((class (glim-gadget-value
                    (glim-get-frame-pane self 'selector))))
       (glim-run-frame-top-level (instance-of class))
       (glim-frame-exit self)))
    (else 
     (glim-frame-exit self))))

;; Fractal terrain
(define-shape-creator-dialog <fracterrain-creator-dialog>
  fracterrain-creator "Create LSystem"
  ((size  "Size"  spin :min 1 :max 100 :step 1 :value 5 :digits 0)
   (seed  "Seed"  spin :min 0 :max 100 :step 1 :value 0 :digits 0)
   (roughness "Roughness" spin :min 1.0 :max 5.0 :step 0.1
                               :value 2.0 :digits 1)
   (perturb   "Perturbation" spin :min 1.0 :max 10.0 :step 0.2
                                  :value 4.0 :digits 1)
   (iter  "Iteration" spin :min 1 :max 32 :step 1 :value 3 :digits 0)
   )
  make-fractal-terrain)

;; Lsystem
(define-shape-creator-dialog <lsystem-creator-dialog>
  lsystem-creator "Create LSystem"
  ((type  "Type"  option :items '(default test))
   (depth "Depth" spin :min 1 :max 8 :step 1 :value 3)
   (alpha "Alpha" spin :min 0 :max 90 :step 1 :value 22.5 :digits 1)
   (delta "Delta" spin :min 0 :max 90 :step 1 :value 22.5 :digits 1)
   )
  make-lsystem)

(provide "banyan/gui/shapedialog")
