;;;
;;; banyan/gui/viewport - viewport
;;;
;;;  Copyright(C) 2002-2003 by Shiro Kawai (shiro@acm.org)
;;;
;;;  Permission to use, copy, modify, distribute this software and
;;;  accompanying documentation for any purpose is hereby granted,
;;;  provided that existing copyright notices are retained in all
;;;  copies and that this notice is included verbatim in all
;;;  distributions.
;;;  This software is provided as is, without express or implied
;;;  warranty.  In no circumstances the author(s) shall be liable
;;;  for any damages arising out of the use of this software.
;;;
;;;  $Id: viewport.scm,v 1.40 2003/01/18 14:47:33 shirok Exp $
;;;

(select-module banyan.gui)
(use srfi-2)
(use math.const)
(use gauche.uvector)
(use gauche.sequence)
(use gauche.mop.instance-pool)
(use gauche.mop.validator)

(export <viewport>
        gl-config-of grid-size-of
        render-mode-of polygon-mode-of
        set-render-mode set-polygon-mode
        set-current-camera reset-current-camera
        viewports-post-redisplay active-viewport)

;; PROTOTYPE IMPLEMENTATION

(define-class <viewport> (<instance-pool-mixin> <validator-mixin>
                          <glim-gadget>)
  ((gl-config  :init-keyword :gl-config :accessor gl-config-of
               :initform (glconfig))
   (grid-size  :init-keyword :grid-size :accessor grid-size-of
               :initform 5
               :validator (lambda (self v)
                            (cond ((integer? v)
                                   (set! (grid-of self) (make-grid v))
                                   v)
                                  (else (set! (grid-of self) #f) #f))))
;   (mode        :init-keyword :mode :init-value 'view
;                :validator (lambda (self v)
;                             (print v)
;                             v))
   (render-mode :init-keyword :render-mode :accessor render-mode-of
                  :initform 'noshade)
   (polygon-mode :init-keyword :polygon-mode :accessor polygon-mode-of
                 :initform 'wire)
   (grid-mode   :init-keyword :grid-mode :accessor grid-mode-of
                :initform #t)
   ;; private
   (viewport)   ; viewport widget
   (camera     :accessor camera-of)
   (camera-attrs :accessor camera-attrs-of)
   (menu       :accessor menu-of :init-value #f
               :init-keyword :menu
               :validator (lambda (o v)
                            (cond ((is-a? v <glim-menu>)
                                   (glim-popup-menu v))
                                  ((is-a? v <glim-popup-menu>) v)
                                  ((not v) v)
                                  (else
                                   (error "<glim-menu>, <glim-popup-menu> or #f is required, but got" v)))))
   (grid       :accessor grid-of :init-value #f)
   (background :accessor background-of :initform (f32vector 0.5 0.5 0.5 0.0))
   (edit-mode  :accessor edit-mode-of :init-value #f)
   (selection-handler :accessor selection-handler-of :init-value #f)

   ;; class slot to keep active viewport
   (active-viewport :allocation :class :init-value #f)
   ))

;; Redisplay all active viewports
(define (viewports-post-redisplay)
  (for-each (lambda (v) (gtk-widget-queue-draw (ref v 'viewport)))
            (instance-pool->list <viewport>)))

;; Returns active viewport
(define (active-viewport)
  (class-slot-ref <viewport> 'active-viewport))

;; A convenience macro
(define-syntax with-gl
  (syntax-rules ()
    ((_ widget () . body)
     (and-let* ((w widget)
                (context (gtk-widget-get-gl-context w))
                (drawable (gtk-widget-get-gl-drawable w))
                ((gdk-gl-drawable-gl-begin drawable context)))
       (begin . body)
       (gdk-gl-drawable-gl-end drawable)))
    ((_ widget (drawable) . body)
     (and-let* ((w widget)
                (context (gtk-widget-get-gl-context w))
                (drawable (gtk-widget-get-gl-drawable w))
                ((gdk-gl-drawable-gl-begin drawable context)))
       (begin . body)
       (gdk-gl-drawable-gl-end drawable)))
    ))

(define-syntax with-viewport-gl
  (syntax-rules ()
    ((_ self . foo)
     (with-gl (ref self 'viewport) . foo))))

(define-syntax Q
  (syntax-rules ()
    ((_ self) (gtk-widget-queue-draw (ref self 'viewport)))))

;;-------------------------------------------------------------
;; Initialize
;;
;;   - make and set default objects
;;   - set up event masks and callbacks
;;
(define-method initialize ((self <viewport>) initargs)
  (let ((w (gtk-drawing-area-new)))
    (set! (widget-of self) w)
    (set! (ref self 'viewport) w)
    (set! (ref self 'active-viewport) self) ;; Fixme
    (gtk-widget-set-size-request w 600 600)
    (next-method)
    (set-current-camera self (car (cameras-of (current-scene))))
    (gtk-widget-set-gl-capability w
                                  (gl-config-of self)
                                  #f #t GDK_GL_RGBA_TYPE)
    ;; necessary to receive keypress event
    (gtk-widget-set-flags w GTK_CAN_FOCUS)

    (gtk-widget-set-events w
                           (logior GDK_EXPOSURE_MASK
                                   GDK_BUTTON_PRESS_MASK
                                   GDK_BUTTON_RELEASE_MASK
                                   GDK_BUTTON1_MOTION_MASK
                                   GDK_BUTTON2_MOTION_MASK
                                   GDK_KEY_PRESS_MASK
                                   GDK_VISIBILITY_NOTIFY_MASK))
    (letrec ((x0 0)
             (y0 0)
             (mouse-modifier 0)
             (mouse-dragged #f)
             (mouse-click-count 0)
             (selection-buffer (make-u32vector 1024))
             (button-selection
              (lambda (event press?)
                (selection-process self event selection-buffer press?
                                   mouse-modifier mouse-click-count)
                (setup-view self #f)
                #t))
             (button-press
              (lambda (w event)
                (let ((x (ref event 'x))
                      (y (ref event 'y)))
                  (set! mouse-modifier (get-event-modifier event))
                  (if (and (not mouse-dragged)
                           (not (eq? mouse-modifier 'camera-control))
                           (= x x0)
                           (= y y0))
                      (inc! mouse-click-count)
                      (set! mouse-click-count 0))
                  (set! x0 x)
                  (set! y0 y)
                  (set! mouse-dragged #f)
                  (cond ((= (ref event 'button) 3)
                         (viewport-menu-popup self event))
                        ((and (= (ref event 'button) 1)
                              (not (eq? mouse-modifier 'camera-control)))
                         (button-selection event #t))
                        ))
                #t))
             (button-release
              (lambda (w event)
                (unless (manipulator-grabbed?)
                  (unless (eq? mouse-modifier 'camera-control)
                    (button-selection event #f)))
                (manipulator-grab -1)     ;release manipulator
                (Q self)
                #t))
             (key-press
              ;; NB: not working
              (lambda (w event)
                (process-key-press self event)))
             (motion-notify
              (lambda (widget event)
                (let* ((x     (ref event 'x))
                       (y     (ref event 'y))
                       (state (ref event 'state)))
                  (cond ((eq? mouse-modifier 'camera-control)
                         (mouse-camera-control self state (- x x0) (- y0 y)))
                        ((manipulator-active?)
                         (mouse-manipulator-control self state x y x0 y0))
                        (else
                         ;; mouse-selection-control
                         #f))
                  (set! x0 x)
                  (set! y0 y)
                  (set! mouse-dragged #t)
                  #t)))
             )
      (g-signal-connect w "realize" (lambda _ (init-gl self) #t))
      (g-signal-connect w "configure_event"
                        (lambda _
                          (with-viewport-gl self () (setup-view self #f))
                          #t))
      (g-signal-connect w "expose_event" (lambda _ (draw-gl self) #t))
      (g-signal-connect w "button_press_event" button-press)
      (g-signal-connect w "button_release_event" button-release)
      (g-signal-connect w "motion_notify_event" motion-notify)
      (g-signal-connect w "key_press_event" key-press)
      ))
  (add-hook! selection-watchers (lambda (scene sels) (Q self)))
  )

;;-----------------------------------------------------------------
;; State manipulation
;;
(define-method set-render-mode ((self <viewport>) mode)
  (set! (render-mode-of self) mode) (Q self))

(define-method set-polygon-mode ((self <viewport>) mode)
  (set! (polygon-mode-of self) mode) (Q self))

(define-method set-grid-mode ((self <viewport>) mode)
  (set! (grid-mode-of self) mode) (Q self))

;; This is to be called from menu handler.
(define-method set-render-flag ((self <viewport>) flag)
  (set-polygon-mode self (if (memq flag '(wire wirel)) 'wire 'fill))
  (set-render-mode self (if (eq? flag 'wire) 'noshade 'shade))
  (set-shade-model self (if (eq? flag 'smooth) GL_SMOOTH GL_FLAT)))

(define-method set-current-camera ((self <viewport>) camera)
  (set! (camera-of self) camera)
  (set! (camera-attrs-of self) (extract-attributes camera))
  (Q self))

(define-method set-shade-model ((self <viewport>) model)
  (with-viewport-gl self () (gl-shade-model model)))

(define-method set-background-color ((self <viewport>) color)
  (set! (background-of self) color)
  (with-viewport-gl self ()
    (gl-clear-color (ref color 0) (ref color 1) (ref color 2) 0.0)))

(define-method set-background-color-interactively ((self <viewport>))
  (define bgcolor-sel
    (let ((sel (gtk-color-selection-dialog-new "Background Color")))
      (g-signal-connect (ref sel 'ok-button) "clicked"
                        (lambda (w)
                          (let1 color (make <gdk-color>)
                            (gtk-color-selection-get-current-color (ref sel 'colorsel) color)
                            (set-background-color self
                                                  (gdk-color->bn-color color)))
                          (Q self)
                          (gtk-widget-destroy sel)))
      (g-signal-connect (ref sel 'cancel-button) "clicked"
                        (lambda (w) (gtk-widget-destroy sel)))
      sel))

  (let* ((color (bn-color->gdk-color (background-of self)))
         (sel   (ref bgcolor-sel 'colorsel)))
    (gtk-color-selection-set-current-color sel color)
    (gtk-color-selection-set-previous-color sel color)
    (gtk-widget-show-all bgcolor-sel))
  )

(define-method reset-current-camera ((self <viewport>))
  (restore-attributes (camera-of self) (camera-attrs-of self))
  (Q self))

(define-method aim-current-camera ((self <viewport>))
  (let1 aim (- #,(vector4f 0 0 0) (center-of-selections (current-scene)))
    (set! (ref (aim-point-of (camera-of self)) 0) (ref aim 0))
    (set! (ref (aim-point-of (camera-of self)) 1) (ref aim 1))
    (set! (ref (aim-point-of (camera-of self)) 2) (ref aim 2))
    ))

;;-----------------------------------------------------------------
;; Callback actions
;;
(define-method init-gl ((self <viewport>))
  (with-viewport-gl self ()
    (let1 bg (background-of self)
      (gl-clear-color (ref bg 0) (ref bg 1) (ref bg 2) 0.0))
    (gl-enable GL_DEPTH_TEST)
    (gl-shade-model GL_FLAT)  ;will be changed by viewport settings
    (gl-enable GL_NORMALIZE)  ;should be configurable
    (gl-polygon-mode GL_FRONT_AND_BACK GL_LINE) ;wireframe default
    (and-let* ((grid-size (grid-size-of self)))
      (set! (grid-of self) (make-grid grid-size)))
    ))

(define-method draw-gl ((self <viewport>))
  (and-let* ((scene (current-scene)))
    (node-update scene)
    (with-viewport-gl self (gldrawable)
      (gl-clear (logior GL_COLOR_BUFFER_BIT GL_DEPTH_BUFFER_BIT))
      (setup-view self #f)
      (when (grid-mode-of self) ((grid-of self)))
      (if (eq? (render-mode-of self) 'shade)
          (begin
            (gl-enable  GL_LIGHTING)
            (node-render default-light-group 'shade))
          (gl-disable GL_LIGHTING))
      (if (eq? (polygon-mode-of self) 'fill)
          (gl-polygon-mode GL_FRONT_AND_BACK GL_FILL)
          (gl-polygon-mode GL_FRONT_AND_BACK GL_LINE))
      (node-render scene (render-mode-of self))
      (manipulator-render (get-manipulator-scale-factor self))
      (gdk-gl-drawable-swap-buffers gldrawable)
      )))

;; Set up the projection matrix.
;; For drawing, pass #f to selection-event.
;; For selection, selection-event contains mouse event.
(define (setup-view viewport selection-event)
  (let* ((wsize (ref (ref viewport 'viewport) 'allocation))
         (w (ref wsize 'width))
         (h (ref wsize 'height))
         (camera (camera-of viewport)))
    (gl-matrix-mode GL_PROJECTION)
    (gl-load-identity)
    (when selection-event
      (let1 viewport (gl-get-integer GL_VIEWPORT)
        (glu-pick-matrix (ref selection-event 'x)
                         (- (ref viewport 3) (ref selection-event 'y))
                         5.0 5.0 viewport)))
    (gl-viewport 0 0 w h)
    (if (perspective? camera)
        (glu-perspective (fov-of camera) (/ w h)
                         (near-plane-of camera)
                         (far-plane-of camera))
        (let* ((w/2 (* (distance-of camera)
                       (%tan (* (/ (fov-of camera) 2) pi/180))))
               (h/2 (* (/ h w) w/2)))
          (gl-ortho (- w/2) w/2 (- h/2) h/2
                    -100 100)))
    (gl-mult-matrix (camera-update-cmatrix camera))
    (gl-matrix-mode GL_MODELVIEW))
  )

;; Camera control
(define (mouse-camera-control viewport state dx dy)
  (let* ((widget (ref viewport 'viewport))
         (wsize (ref widget 'allocation))
         (w     (ref wsize 'width))
         (h     (ref wsize 'height))
         (mode (+ (if (zero? (logand state GDK_BUTTON1_MASK)) 0 1)
                  (if (zero? (logand state GDK_BUTTON2_MASK)) 0 2)))
         (camera (camera-of viewport)))
    (case mode
      ((1) ;; Rotation
       (when (rotatable? camera)
         ;; The factor 200 is arbitrary chosen.

         (dec! (y-rotation-of camera) (* (/ dx w) 200))
         (inc! (x-rotation-of camera) (* (/ dy h) 200))
         (Q viewport)))
      ((2) ;; Translation
       ;; TODO: avoid allocation of quaternions!
       (let* ((mag (/ (/ (distance-of camera) (near-plane-of camera)) h))
              (dv  (quatf-transform (* (make-quatf #,(vector4f 0 1 0)
                                                   (* (y-rotation-of camera)
                                                      pi/180))
                                       (make-quatf #,(vector4f 1 0 0)
                                                   (* (x-rotation-of camera)
                                                      pi/180)))
                                    (vector4f (* mag dx) (* mag dy) 0)))
              (ap  (aim-point-of camera)))
         (point4f-add! ap dv))
       (Q viewport))
      ((3) ;; Zoom
       (dec! (distance-of camera) (* (/ dx w) 20))
       (Q viewport))
      )
    ))

;; Calculate scale factor of manipulator from camera
;; 20 is an arbitrary factor - depends on how big you want to see
;; the manipulators on screen.
(define (get-manipulator-scale-factor viewport)
  (/ (distance-of (camera-of viewport)) 20))

;; get event modifier signature
(define (get-event-modifier event)
  (let1 state (ref event 'state)
    (cond ((not (zero? (logand state GDK_CONTROL_MASK)))
           'camera-control)
          ((not (zero? (logand state GDK_SHIFT_MASK)))
           'incremental)
          (else #f))))

;; Mouse manipulator control
(define (mouse-manipulator-control viewport state x y x0 y0)
  (with-viewport-gl viewport ()
    (setup-view viewport #f)            ;reset projection matrix
    (manipulator-drag x y x0 y0))
  (Q viewport))

;; keypress control
(define (process-key-press viewport event)
  (let ((kv (ref event 'keyval)))
    (cond
     ((= kv GDK_Delete)
      (delete-instance-selected (current-scene))
      (Q viewport))
     )
    #t))

;;-----------------------------------------------------------
;; Selection
;;

(define-method selection-draw ((self <viewport>) event buffer)
  (let ((hits #f))
    (with-viewport-gl self ()
      (gl-select-buffer buffer)
      (gl-render-mode GL_SELECT)
      (gl-init-names)
      (setup-view self event)
      (gl-push-name 0) ; the top stack = 0 : normal objects
      (gl-push-name 0)
      (and-let* ((scene (current-scene))) (node-render scene 'selection))
      (gl-pop-name)
      (gl-load-name 1) ; the top stack = 1 : manipulator
      (gl-push-name 0)
      (manipulator-selection-draw (get-manipulator-scale-factor self))
      (gl-pop-name)
      (set! hits (gl-render-mode GL_RENDER)))
    hits))

(define-method selection-get ((self <viewport>) num-hits scene buffer count)
  (let ((obj-min-z #xffffffff)
        (manip-min-z #xffffffff)
        (major -1)
        (minor -1)
        (cnt 0))
    (dotimes (i num-hits)
      (let ((num-names (ref buffer cnt))
            (z (ref buffer (+ cnt 1)))
            (maj (ref buffer (+ cnt 3))))
        (case maj
          ((0) ;; object
           (when (and (<= major 0) (< z obj-min-z))
             (set! obj-min-z z)
             (set! major 0)
             (set! minor (ref buffer (+ cnt 4)))))
          ((1) ;; manip
           (when (< z manip-min-z)
             (set! manip-min-z z)
             (set! major 1)
             (set! minor (ref buffer (+ cnt 4))))))
        (inc! cnt (+ num-names 3))))
    (values major minor)))

(define-method default-selection-handler ((self <viewport>) target object
                                          press? modifier)
  (case target
    ((clear)
     (when press? (reset-selections (current-scene))))
    ((instance)
     (unless press?
       (unless (eq? modifier 'incremental)
         (reset-selections (current-scene)))
       (add-selection (current-scene) object)))
    ((manipulator)
     (when press? (manipulator-grab object)))
    ))

(define-method selection-process ((self <viewport>) event buffer press?
                                  modifier count)
  (let* ((num-hits (selection-draw self event buffer))
         (scene (current-scene))
         (handler (or (selection-handler-of self)
                      default-selection-handler)))
    (receive (major minor)
        (selection-get self num-hits scene buffer count)
      (case major
        ((-1) ; clicked empty space
         (handler self 'clear #f press? modifier))
        ((0)  ; object selection
         (handler self 'instance (id->instance (current-scene) minor)
                  press? modifier))
        ((1)  ; manipulator selection
         (handler self 'manipulator minor press? modifier))))
    ))

;;-----------------------------------------------------------
;; Make grid
;;

(define (make-grid size)
  (let* ((2size   (* 2 size))
         (2size+1 (+ 2size 1))
         (nvtx    (* 2size+1 4))
         (nlines  (* 2size 2))
         (vtx     (make-point4f-array nvtx))
         (lines   (make-u32vector (* nlines 2)))
         (axes    #f))
    ;; Vertices
    (dotimes (nx 2size+1)
      (set! (ref vtx (* nx 2)) (point4f (- nx size) 0 (- size)))
      (set! (ref vtx (+ (* nx 2) 1)) (point4f (- nx size) 0 size)))
    (dotimes (nz 2size+1)
      (set! (ref vtx (+ (* 2size+1 2) (* nz 2))) (point4f (- size) 0 (- nz size)))
      (set! (ref vtx (+ (* 2size+1 2) (* nz 2) 1)) (point4f size 0 (- nz size))))
    ;; Grid lines
    (dotimes (nx (* 2size 2))
      (if (< nx 2size)
          (set! (ref lines nx) nx)
          (set! (ref lines nx) (+ nx 2))))
    (dotimes (nz (* 2size 2))
      (if (< nz 2size)
          (set! (ref lines (+ (* 2size 2) nz)) (+ (* 2size+1 2) nz))
          (set! (ref lines (+ (* 2size 2) nz)) (+ (* 2size+1 2) nz 2))))
    ;; Axis lines
    (set! axes
          (u32vector 2size 2size+1
                     (+ (* 2size+1 2) 2size) (+ (* 2size+1 2) 2size+1)))
    (lambda ()
      (gl-load-identity)
      (gl-disable GL_LIGHTING)
      (gl-enable-client-state GL_VERTEX_ARRAY)
      (gl-vertex-pointer 4 vtx)
      (gl-color 0.3 0.3 0.3 0.0)
      (gl-draw-elements GL_LINES lines)
      (gl-line-width 2.0)
      (gl-color 0.0 0.0 0.0 0.0)
      (gl-draw-elements GL_LINES axes)
      (gl-line-width 1.0)
      )
    ))

;;-----------------------------------------------------------------
;; Menu
;;  The definition of menu is in mainmenu.scm

(define (viewport-menu-popup viewport event)
  (and-let* ((menu (ref viewport 'menu)))
    (glim-popup-menu-popup menu)))

(provide "banyan/gui/viewport")
