;;;
;;; banyan/instance 
;;;
;;;  Copyright(C) 2002-2003 by Shiro Kawai (shiro@acm.org)
;;;
;;;  Permission to use, copy, modify, distribute this software and
;;;  accompanying documentation for any purpose is hereby granted,
;;;  provided that existing copyright notices are retained in all
;;;  copies and that this notice is included verbatim in all
;;;  distributions.
;;;  This software is provided as is, without express or implied
;;;  warranty.  In no circumstances the author(s) shall be liable
;;;  for any damages arising out of the use of this software.
;;;
;;;  $Id: instance.scm,v 1.30 2003/04/13 04:39:14 shirok Exp $
;;;

(select-module banyan)
(use srfi-2)

;; PROTOTYPE IMPLEMENTATION

;;============================================================
;; Instance
;;

;; Instances are what the user "sees" on the screen as objects.

(define-class <instance> (<renderable-node>)
  ((transform :init-keyword :transform :accessor transform-of
              :initform (make <trs-transform>)
              :attribute-type <transform>)
   (shader    :init-keyword :shader :accessor shader-of
              :initform (default-shader)
              :attribute-type <shader>)
   (object    :init-keyword :object :accessor object-of
              :initform #f
              :attribute-type <node>)
   (selected? :accessor selected? :initform #f)
   ;; Parent/child relationship.  For the time being, these are
   ;; derived from hierarchy of transform nodes.
   (parents   :allocation :virtual :accessor parents-of
              :slot-ref  (lambda (o) (instance-find-parents o))
              :slot-set! (lambda (o v) #f))
   (children  :allocation :virtual :accessor children-of
              :slot-ref  (lambda (o) (instance-find-children o))
              :slot-set! (lambda (o v) #f))
   ))

;; Initiation - a special initialize method at deserialization

(define-method initiate ((self <instance>) initargs)
  (initialize self (list* :transform #f :shader #f :object #f initargs)))

(define (instance-find-children self)
  ;; clumsy, but for now ...
  (let1 r '()
    (for-each (lambda (child-xform)
                (for-each (lambda (node&slot)
                            (when (and (is-a? (car node&slot) <instance>)
                                       (eq? (cdr node&slot) 'transform))
                              (push! r (car node&slot))))
                          (node-collect-reference child-xform)))
              (children-of (transform-of self)))
    r))

(define (instance-find-parents self)
  ;; clumsy, but for now ...
  (and-let* ((parent (parent-of (transform-of self)))
             (r '()))
    (for-each (lambda (node&slot)
                (when (and (is-a? (car node&slot) <instance>)
                           (eq? (cdr node&slot) 'transform))
                  (push! r (car node&slot))))
              (node-collect-reference parent))
    r))

;;------------------------------------------------------------
;; Render protocol
;;
(define-method node-render ((self <instance>) op)
  (and-let* (((visible? self))
             (object (object-of self)))
    (let1 xform (transform-of self)
      (gl-load-matrix (gmatrix-of xform)))
    (when (eq? op 'selection)
      (and-let* ((id (instance->id (current-scene) self)))
        (gl-load-name id)))
    (if (eq? op 'shade)
        (node-render (shader-of self) object)
        (case (selected? self)
          ((#f) (node-render *wire-shader-unselected* object))
          ((0)  (node-render *wire-shader-selected* object))
          ((1)  (node-render *wire-shader-selected1* object))
          (else (node-render *wire-shader-selectedn* object))))
    ))

;;------------------------------------------------------------
;; Update protocol
;;

(define-method node-depends ((self <instance>))
  (let1 xform (transform-of self)
    (if xform (list xform) '())))

;;------------------------------------------------------------
;; Selection protocol
;;
(define-method instance-select ((self <instance>) . args)
  (set! (selected? self) (get-optional args 0)))

(define-method instance-deselect ((self <instance>))
  (set! (selected? self) #f))

;;------------------------------------------------------------
;; Miscellaneous operations
;;

;; Parenting.
(define-method instance-parent ((self <instance>) (parent <instance>))
  (let ((s-xform (transform-of self))
        (p-xform (transform-of parent)))
    (set! (parent-of s-xform) p-xform)
    ))

(define-method instance-unparent ((self <instance>))
  (set! (parent-of (transform-of self)) #f))

;; Freeze local transformation
;;   WHICH is any combination of 'translation, 'rotation or 'scale.
(define-method freeze-transform! ((self <instance>) which)
  (freeze-transform-sub! self which (make-matrix4f)))

(define (freeze-transform-sub! self which parent-matrix)
  (let ((xform (transform-of self))
        (shape (object-of self)))
    ;; NB: non-TRS-transform is not supported (yet).
    (unless (is-a? xform <trs-transform>)
      (error "freezing non-TRS-transform is not supported yet"))
    (let1 fmatrix (matrix4f-copy parent-matrix)
      (when (memq 'translation which)
        (matrix4f-mul! fmatrix (translation->matrix4f (translation-of xform)))
        (set! (translation-of xform) (vector4f 0 0 0))
        (attribute-change-notify xform 'translation))
      (when (memq 'rotation which)
        (matrix4f-mul! fmatrix (quatf->matrix4f (rotation-of xform)))
        (set! (rotation-of xform) (make-quatf))
        (attribute-change-notify xform 'rotation))
      (when (memq 'scale which)
        (matrix4f-mul! fmatrix (scale->matrix4f (scale-of xform)))
        (set! (scale-of xform) (vector4f 1 1 1))
        (attribute-change-notify xform 'scale))
      (freeze-transform! shape fmatrix)
      (for-each (cut freeze-transform-sub! <> which fmatrix)
                (children-of self))
      )
    ))

;; Duplicate instance.
;;   This method has more control than the simple node-duplicate method.
(define-method instance-duplicate ((self <instance>) . args)
  (let-keywords* args ((duplicate-shader #f)
                       (duplicate-object #f)
                       (transform #f)
                       (select-duplicated #f))
    (let* ((shader (if duplicate-shader
                       (node-duplicate (shader-of self) #t)
                       (shader-of self)))
           (object (if duplicate-object
                       (node-duplicate (object-of self) #t)
                       (object-of self)))
           (xform  (node-duplicate (transform-of self) #f))
           (dup    (make <instance>
                     :name (node-duplicate-default-name self)
                     :part (part-of self)
                     :transform xform
                     :shader shader
                     :object object))
           (scene  (current-scene)))
      (add-instance scene dup)
      (when select-duplicated
        (when (selected? self)
          (remove-selection scene self)
        (add-selection scene dup)))
      dup)))

(define-method node-duplicate ((self <instance>) deep-copy?)
  (instance-duplicate self
                      :duplicate-shader deep-copy?
                      :duplicate-object deep-copy?))

;; Switch part of instance
(define-method instance-change-parts ((self <instance>) newpart . args)
  (let-keywords* args ((with-shader #t)
                       (with-object #t))
    (set! (part-of self) newpart)
    (set! (part-of (transform-of self)) newpart)
    (when with-shader
      (set! (part-of (shader-of self)) newpart))
    (when with-object
      (set! (part-of (object-of self)) newpart))
    (notify-graph-changed (current-scene))
    ))

;;------------------------------------------------------------
;; Pre-defined instance
;;
(define-syntax %predef-instance-maker
  (syntax-rules ()
    ((_ args optionals object-creator name-hint)
     (let-keywords* args optionals
       (let ((instance (make <instance>
                         :object object-creator
                         :name name-hint))
             (scene    (current-scene)))
         (add-instance scene instance)
         (when (get-keyword :select args #t)
           (reset-selections scene)
           (add-selection scene instance))
         instance)))))

(define (make-cube . args)
  (%predef-instance-maker args ((size 2.0))
                          (make-cube-shape size)
                          "cube-1"))

(define (make-sphere . args)
  (%predef-instance-maker args ((radius 1.0)
                                (t-div  20.0)
                                (p-div  10.0))
                          (make-sphere-shape radius t-div p-div)
                          "sphere-1"))

(define (make-cylinder . args)
  (%predef-instance-maker args ((radius 1.0)
                                (height 2.0)
                                (t-div  20.0)
                                (p-div  10.0))
                          (make-cylinder-shape radius height t-div p-div)
                          "cylinder-1"))

(define (make-cone . args)
  (%predef-instance-maker args ((radius 1.0)
                                (height 2.0)
                                (t-div  20.0)
                                (p-div  10.0))
                          (make-cone-shape radius height t-div p-div)
                          "cone-1"))

(define (make-plane . args)
  (%predef-instance-maker args ((x-size 2.0)
                                (z-size 2.0)
                                (x-div  1.0)
                                (z-div  1.0))
                          (make-plane-shape x-size z-size x-div z-div)
                          "plane-1"))

(define (make-teapot . args)
  (%predef-instance-maker args ((size 2.0))
                          (make <teapot-shape> :size-of size)
                          "teapot-1"))

(define (make-locator . args)
  (%predef-instance-maker args ()
                          (make <locator-shape>)
                          "locator-1"))

;; NB: this is a temporary interface
(define (make-fractal-terrain . args)
  (%predef-instance-maker args ((size 5.0)
                                (seed 0)
                                (roughness 2.0)
                                (perturb   4.0)
                                (iter      3))
                          (make-fractal-terrain-shape size seed roughness
                                                      perturb iter)
                          "terrain-1"))

;; NB: this is a temporary interface
(define (make-lsystem . args)
  (let-keywords* args ((type 'default)
                       (depth 3.0)
                       (alpha 22.5)
                       (delta 22.5))
    (let ((instances (map (lambda (shape)
                            (make <instance>
                              :object shape :name "lsystem-1"))
                          (make-lsystem-shapes type depth alpha delta)))
          (scene (current-scene))
          (parent-locator (make-locator)))
      (add-instance scene parent-locator)
      (for-each (cut add-instance scene <>) instances)
      (for-each (cut instance-parent <> parent-locator) instances)
      (when (get-keyword :select args #t)
        (apply set-selection scene instances))
      instances)))

(provide "banyan/instance")
