;;;
;;; banyan/license - handles license of created contents
;;;
;;;  Copyright(C) 2003 by Shiro Kawai (shiro@acm.org)
;;;
;;;  Permission to use, copy, modify, distribute this software and
;;;  accompanying documentation for any purpose is hereby granted,
;;;  provided that existing copyright notices are retained in all
;;;  copies and that this notice is included verbatim in all
;;;  distributions.
;;;  This software is provided as is, without express or implied
;;;  warranty.  In no circumstances the author(s) shall be liable
;;;  for any damages arising out of the use of this software.
;;;
;;;  $Id: license.scm,v 1.3 2003/02/20 03:09:52 shirok Exp $
;;;

;; A license object can be used to maintain licence of the part data
;; on the server.   It can be derived from part's metadata.
;; We use Creative Commons licence model.
;; Cf. http://creativecommons.org/

(define-module banyan.license
  (use srfi-1)
  (use srfi-2)
  (use srfi-19)
  (use text.html-lite)
  (use text.tree)
  (use file.util)
  (use banyan)
  (use gauche.version)
  (export <cc:license>
          part-license
          cc:license->rdf)
  )
(select-module banyan.license)

(define-class <cc:license> ()
  ((uri         :init-keyword :uri :init-value "")
   (title       :init-keyword :title :init-value "untitled")
   (description :init-keyword :description :init-value "")
   (publisher   :init-keyword :publisher :init-value '())
   (creator     :init-keyword :creator :init-value '())
   (contributor :init-keyword :contributor :init-value '())
   (copyright-holder :init-keyword :copyright-holder :init-value '())
   (copyright-date :init-keyword :copyright-date :init-value (sys-time))
   (derived-from :init-keyword :derived-from :init-value '())
   (cc-license   :init-keyword :cc-license :init-value 'by)
   (cc-license-name :allocation :virtual
                    :slot-ref  (lambda (o)
                                 (cc-license-name (ref o 'cc-license)))
                    :slot-set! (lambda (o v) #f))
   ))

;; Convert part metadata to license structure.
(define-method part-license ((part <part>))
  (make <cc:license>
    :uri (or (and-let* ((repo (part-repository part)))
               (build-path (ref repo 'uri) (part-name part)))
             "")
    :title (or (ref part 'title) (part-name part))
    :description (or (ref part 'description) "")
    :publisher '()
    :creator (cond ((ref part 'created-by) => list) (else '()))
    :contributor '()
    :copyright-holder (cond ((ref part 'created-by) => list) (else '()))
    :copyright-date (ref part 'creation-time)
    :derived-from (or (ref part 'derived-from) '())
    :cc-license (or (ref part 'cc-license) 'by)))

;; Creative Commons license types
(define-constant *cc-licenses*
  '(
    (by       "Attribution"
              (requires Attribution Notice)
              (permits  Reproduction Distribution DerivativeWorks))
    (by-nd    "Attribution-NoDerivs"
              (requires Attribution Notice)
              (permits  Reproduction Distribution))
    (by-nd-nc "Attribution-NoDerivs-NonCommercial"
              (requires Attribution Notice)
              (permits  Reproduction Distribution)
              (prohibits CommercialUse))
    (by-nc    "Attribution-NonCommercial"
              (requires Attribution Notice)
              (permits  Reproduction Distribution DerivativeWorks)
              (prohibits CommercialUse))
    (by-nc-sa "Attribution-NonCommercial-ShareAlike"
              (requires Attribution ShareAlike Notice)
              (permits  Reproduction Distribution DerivativeWorks)
              (prohibits CommercialUse))
    (by-sa    "Attribution-ShareAlike"
              (requires Attribution ShareAlike Notice)
              (permits Reproduction Distribution DerivativeWorks))
    (nd       "NoDerivs"
              (requires Notice)
              (permits  Reproduction Distribution))
    (nd-nc    "NoDerivs-NonCommercial"
              (requires Notice)
              (permits  Reproduction Distribution)
              (prohibits CommercialUse))
    (nc       "NonCommercial"
              (requires Notice)
              (permits  Reproduction Distribution DerivativeWorks)
              (prohibits CommercialUse))
    (nc-sa    "NonCommercial-ShareAlike"
              (requires ShareAlike Notice)
              (permits  Reproduction Distribution DerivativeWorks)
              (prohibits CommercialUse))
    (sa       "ShareAlike"
              (requires ShareAlike Notice)
              (permits  Reproduction Distribution DerivativeWorks))
    (pub      "PublicDomain"
              (permits  Reproduction Distribution DerivativeWorks))
    ))

(define (cc-license-uri lic)
  #`"http://creativecommons.org/licenses/,|lic|/1.0")

(define (cc-license-name lic)
  (cond ((assq lic *cc-licenses*) => cadr)
        (else "No License Attached")))

(define (cc-license-conditions lic)
  (define (url item)
    #`"http://web.resource.org/cc/,item")
  (or (and-let* ((entry (assq lic *cc-licenses*)))
        (append-map (lambda (condition)
                      (map (lambda (item)
                             `(,(car condition) (@ (rdf:resource ,(url item)))))
                           (cdr condition)))
                    (cddr entry)))
      '()))

;; NB: srfi-19 module had a bug in make-time until Gauche-0.6.7.1.
;; This is workaround.
(define (make-time-fixed type nano sec)
  (make <time> :type type :nanosecond nano :second sec))

(define (cc-date-format time)
  (date->string
   (time-utc->date (make-time-fixed 'time-utc 0 time))
   "~Y-~m-~dT~H:~M:~S~z"))

;; NB: srdf is RDF structure in SXML.  I can't figure out how to
;; place the default namespace in SXML, so I put namespace stuff
;; in attributes for now.
(define-method cc:license->srdf ((self <cc:license>))
  `(rdf:RDF
    (@ (xmlns "http://web.resource.org/cc/")
       (xmlns:dc "http://purl.org/dc/elements/1.1/")
       (xmlns:rdf "http://www.w3.org/1999/02/22-rdf-syntax-ns#"))
    (Work
     (@ (rdf:about ,(ref self 'uri)))
     (dc:title ,(ref self 'title))
     (dc:description ,(ref self 'description))
     ,@(map (lambda (entry) `(dc:creator (Agent (dc:title ,entry))))
            (ref self 'creator))
     ,@(map (lambda (entry) `(dc:rights (Agent (dc:title ,entry))))
            (ref self 'copyright-holder))
     ,@(if (ref self 'copyright-date)
           `((dc:date ,(cc-date-format (ref self 'copyright-date))))
           '())
     ,@(map (lambda (entry) `(dc:contributor (Agent (dc:title ,entry))))
            (ref self 'contributor))
     ,@(map (lambda (entry) `(dc:publisher (Agent (dc:title ,entry))))
            (ref self 'publisher))
     ,@(map (lambda (entry) `(dc:source (@ (rdf:resource ,entry))))
            (ref self 'derived-from))
     (license (@ (rdf:resouce ,(cc-license-uri (ref self 'cc-license)))))
     )
    (License
     (@ (rdf:about ,(cc-license-uri (ref self 'cc-license))))
     ,@(cc-license-conditions (ref self 'cc-license)))
    ))

;; Should be in global xml module, but for now we do by ourselves.
(define (srdf->rdf tree port)

  (define (attr attrs)
    (tree->string
     (map (lambda (e)
            `(,(car e) "=\"" ,(html-escape-string (cadr e)) "\" "))
          (cdr attrs))))
  
  (if (pair? tree)
      (cond ((null? (cdr tree)) (display #`"<,(car tree) />\n" port))
            ((and (null? (cddr tree))
                  (pair? (cadr tree)) (eq? (caadr tree) '@))
             (display #`"<,(car tree) ,(attr (cadr tree)) />\n" port))
            ((and (pair? (cadr tree)) (eq? (caadr tree) '@))
             (display #`"<,(car tree) ,(attr (cadr tree))>\n" port)
             (for-each (cut srdf->rdf <> port) (cddr tree))
             (display #`"</,(car tree)>\n" port))
            (else
             (display #`"<,(car tree)>\n" port)
             (for-each (cut srdf->rdf <> port) (cdr tree))
             (display #`"</,(car tree)>\n" port)))
      (display tree port)))

(define-method cc:license->rdf ((self <cc:license>))
  (call-with-output-string
    (cut srdf->rdf (cc:license->srdf self) <>)))
      
(provide "banyan/license")

