;;;
;;; banyan.meta - metaclass
;;;
;;;  Copyright(C) 2002-2003 by Shiro Kawai (shiro@acm.org)
;;;
;;;  Permission to use, copy, modify, distribute this software and
;;;  accompanying documentation for any purpose is hereby granted,
;;;  provided that existing copyright notices are retained in all
;;;  copies and that this notice is included verbatim in all
;;;  distributions.
;;;  This software is provided as is, without express or implied
;;;  warranty.  In no circumstances the author(s) shall be liable
;;;  for any damages arising out of the use of this software.
;;;
;;;  $Id: meta.scm,v 1.31 2003/11/11 23:50:05 shirok Exp $
;;;

(select-module banyan)
(use srfi-1)
(use srfi-2)
(use srfi-13)
(use util.tree)
(use gauche.charconv)
(use gauche.hook)
(use gauche.mop.validator)
(use gauche.mop.instance-pool)

;; PROTOTYPE IMPLEMENTATION

;;==================================================================
;; <node>
;;

;; Node is something that can be named, saved, and edited by
;; the attribute editor.
;;
;; A node always belongs to some part. (See <part> definition
;; below).
;;
;; In some node classes, a node may be accompanied by C-defined structure.
;; Such C-structure is constructed at the initialize method of the
;; node base class, and stored in %c-node slot, if node class's c-class
;; slot is set.
;; The node metaclass recognizes :c-node slot allocation, which accesses
;; the C structure slots.

(define-class <node-meta> (<instance-pool-meta> <validator-meta> <class>)
  ((abstract :init-value #f :accessor abstract?
             :init-keyword :abstract)
   ;(subclasses :init-value '())         ;alternative of direct-subclasses
   (c-class  :init-value <%node> :init-keyword :c-class) ;C-counterpart
   ))

;(define-method initialize ((class <node-meta>) initargs)
;  (next-method)
;  (for-each (lambda (parent)
;              (when (is-a? parent <node-meta>)
;                (slot-push! parent 'subclasses class)))
;            (class-direct-supers class)))

(define-method compute-get-n-set ((class <node-meta>) slot)
  (if (eq? (slot-definition-allocation slot) :c-node)
      (let ((name (slot-definition-name slot)))
        (list (lambda (o)
                (slot-ref (slot-ref o '%c-node) name))
              (lambda (o v)
                (slot-set! (slot-ref o '%c-node) name v))
              #t))
      (next-method)))

;; Initiate is much like make, except calls instance's initiate 
;; method instead of initialize method.  It is called when deserializing
;; the node graph.  Instance's initiate method can do some trick in that
;; special occasion, rather than standard initialization.
(define-method initiate ((class <node-meta>) . initargs)
  (let ((obj (allocate-instance class initargs)))
    (initiate obj initargs)
    obj))

(define-class <node> (<instance-pool-mixin>)
  ((%c-node)                            ;c-counterpart
   (name   :init-keyword :name   :accessor name-of)
   (part   :init-keyword :part   :accessor part-of)
   (observers :initform (make-hook 3))
   (modified? :init-value #f     :accessor modified?)
   )
  :metaclass <node-meta>
  :abstract #t)

(define (%c-node node) (slot-ref node '%c-node))

;; Default initiate method is just initialize.
(define-method initiate ((self <node>) initargs)
  (initialize self initargs))

;;----------------------------------------------------------------
;; Common base methods - predicate, initializer, printer, etc.
;;
(define (node? obj) (is-a? obj <node>))

(define-method initialize ((self <node>) initargs)
  ;; instantiate and connect C-counterpart
  (let ((c-class (slot-ref (class-of self) 'c-class)))
    (unless (eq? c-class <%node>)
      (let ((c-node (make c-class)))
        (slot-set! self '%c-node c-node)
        (slot-set! c-node 'node self))))
  ;; standard initialize
  (next-method)
  ;; Make sure the node belongs to a part, and have unique name within
  ;; the part.  The order of giving the node a name and adding it to
  ;; a part is important.
  (let ((part (if (slot-bound? self 'part)
                  (part-of self)
                  (current-part)))
        (name (if (slot-bound? self 'name)
                  (ref self 'name)
                  (string-append (string-trim-both
                                  (x->string
                                   (class-name (class-of self)))
                                  #[<>])
                                 "-1"))))
    (set! (ref self 'name) (get-unique-name part name))
    (set! (ref self 'part) part)
    (add-node part self)
    ))

(define-method write-object ((self <node>) out)
  (format out "#<~a ~s>"
          (string-trim-both (x->string (class-name (class-of self))) #[<>])
          (node-path self)))

(define (node-path node)
  #`",(part-name (part-of node)):,(name-of node)")

(define (valid-node-name? name)
  (#/^[-_\w]+$/ name))

(define (name->node name)
  (let1 part&name (string-split name #\:)
    (case (length part&name)
      ((1) ;; find all nodes of the given name
       (instance-pool-fold <node>
                           (lambda (n r)
                             (if (equal? (name-of n) name)
                                 (cons n r)
                                 r))
                           '()))
      ((2) ;; find the node that matches
       (and-let* ((part (name->part (car part&name))))
         (find (lambda (n) (equal? (name-of n) (cadr part&name)))
               (part-nodes part))))
      (else #f))))

;;----------------------------------------------------------------
;; Metainformation utility
;;

(define-method node-meta-subclasses ((class <node-meta>))
  (class-direct-subclasses class))

(define-method node-meta-all-subclasses ((class <node-meta>))
  (reverse!
   (tree-fold class
              cons
              '()
              (lambda (node) (null? (class-direct-subclasses node)))
              (lambda (proc knil node)
                (fold proc (cons node knil)
                      (class-direct-subclasses node))))))

(define-method node-meta-all-concrete-subclasses ((class <node-meta>))
  (reverse!
   (tree-fold class
              (lambda (node knil)
                (if (abstract? node)
                    knil
                    (cons node knil)))
              '()
              (lambda (node) (null? (class-direct-subclasses node)))
              (lambda (proc knil node)
                (fold proc
                      (if (abstract? node)
                          knil
                          (cons node knil))
                      (class-direct-subclasses node))))))

;;----------------------------------------------------------------
;; Node attribute protocol
;;
(define (attribute-type slot)
  (slot-definition-option slot :attribute-type #f))

(define-method node-attribute-slots ((self <node>))
  (filter-map (lambda (slot)
                (and (attribute-type slot) slot))
              (class-slots (class-of self))))

(define-method extract-attributes ((self <node>))
  (map (lambda (slot)
         (let* ((name (slot-definition-name slot))
                (val  (slot-ref self name)))
           (cons name (attribute-value-dup val))))
       (node-attribute-slots self)))

(define-method restore-attributes ((self <node>) data)
  (for-each (lambda (p) (slot-set! self (car p) (cdr p))) data))

(define-method attribute-value-dup ((v <top>)) v)
(define-method attribute-value-dup ((v <list>)) (list-copy v))
(define-method attribute-value-dup ((v <vector>)) (vector-copy v))
(define-method attribute-value-dup ((v <f32vector>)) (f32vector-copy v))
(define-method attribute-value-dup ((v <vector4f>)) (vector4f-copy v))
(define-method attribute-value-dup ((v <point4f>)) (point4f-copy v))
(define-method attribute-value-dup ((v <quatf>)) (quatf-copy v))
(define-method attribute-value-dup ((v <matrix4f>)) (matrix4f-copy v))

;; Collect references that points to this node.  returns a list of
;; (<node> . slot).
;; This would be too slow if we have tens of thousands of nodes,
;; but it works for now.
(define-method node-collect-reference ((self <node>))
  (let ((r '()))
    (instance-pool-for-each
     <node>
     (lambda (node)
       (for-each
        (lambda (slot)
          (let ((atype (slot-definition-option slot :attribute-type #f))
                (aname (slot-definition-name slot)))
            (when (and atype
                       (is-a? atype <class>)
                       (or (memq <node> (class-precedence-list atype))
                           (memq atype (class-precedence-list <node>)))
                       (eq? (slot-ref node aname) self))
              (push! r (cons node aname)))))
        (class-slots (class-of node)))))
    r))

;; attribute observer

(define-method node-observer-add! ((self <node>) observer)
  (add-hook! (ref self 'observers) observer))

(define-method node-observer-delete! ((self <node>) observer)
  (delete-hook! (ref self 'observers) observer))

(define-method attribute-set! ((self <node>) slot value)
  (set! (ref self 'slot) value)
  (set! (modified? self) #t)
  (run-hook (ref self 'observers) self slot value)
  (when (part-of self) (set-modified! (part-of self))))

(define-method attribute-change-notify ((self <node>) slot)
  (run-hook (ref self 'observers) self slot (ref self slot))
  (when (part-of self) (set-modified! (part-of self))))

;;----------------------------------------------------------------
;; Node duplicate protocol
;;

;; default naming convention of duplicated node:
;;    foo     ->  foo-copy
;;    foo-copy -> foo-copy-1
;;    foo-copy-1 -> foo-copy-2 ...
(define-method node-duplicate-default-name ((self <node>))
  (get-unique-name (part-of self)
                   (if (#/-copy(-\d+)?$/ (name-of self))
                       (name-of self)
                       #`",(name-of self)-copy")))

;; The default method sets name and part slot, and copies all
;; attribute slots.  If it is not appropriate, subclass has to
;; overload this method.
;; If deep-copy? is true, the nodes pointed by this node are
;; copied recursively.
(define-method node-duplicate ((self <node>) deep-copy?)
  (let ((new (make (class-of self)
               :name (node-duplicate-default-name self)
               :part (part-of self))))
    (for-each (lambda (slot)
                (let* ((name (slot-definition-name slot))
                       (value (attribute-value-dup (slot-ref self name))))
                  (slot-set! new name
                             (if (and deep-copy? (is-a? value <node>))
                                 (node-duplicate value #t)
                                 value))))
              (node-attribute-slots self))
    new))

;;----------------------------------------------------------------
;; Node delete protocol
;;

;; 'Deleting' node should make it 'invisible' from the scene,
;; which can be done by removing all references to the node.
(define-method node-delete ((self <node>))
  (let ((refs (node-collect-reference self)))
    (for-each (lambda (ref)
                ;; TODO: some slots might require default node to be set.
                (slot-set! (car ref) (cdr ref) #f))
              refs))
  (cond ((part-of self) => (delete-node <> self)))
  (instance-pool-remove! self (cut eq? self <>))
  )

;;----------------------------------------------------------------
;; Node render protocol
;;
;;  OP might be:
;;    noshade
;;    shade
;;    selection

(define-class <renderable-node> (<node>)
  ((visibility :init-keyword :visibility :init-value #t
               :attribute-type <boolean>)
   (temporary-visibility :init-value #t)
   ))

(define-method node-render ((self <renderable-node>) op) #f)

(define-method visible? ((self <renderable-node>))
  (and (ref self 'visibility) (ref self 'temporary-visibility)))

;;----------------------------------------------------------------
;; Node update protocol
;;

(define-class <updatable-node> (<node>)
  ())

(define-method initialize ((self <updatable-node>) initargs)
  (next-method)
  (and-let* ((scene (current-scene)))
    (notify-graph-changed (current-scene))))

(define-method node-update ((self <updatable-node>)) #f)

;; Method node-depends is used to construct scene update list.
;; Each subclass has to overload this method, if its output
;; depends on some of its input.
(define-method node-depends ((self <node>)) '())

(provide "banyan/meta")
