;;;
;;; banyan.part - part
;;;
;;;  Copyright(C) 2002-2003 by Shiro Kawai (shiro@acm.org)
;;;
;;;  Permission to use, copy, modify, distribute this software and
;;;  accompanying documentation for any purpose is hereby granted,
;;;  provided that existing copyright notices are retained in all
;;;  copies and that this notice is included verbatim in all
;;;  distributions.
;;;  This software is provided as is, without express or implied
;;;  warranty.  In no circumstances the author(s) shall be liable
;;;  for any damages arising out of the use of this software.
;;;
;;;  $Id: part.scm,v 1.19 2003/02/16 14:02:26 shirok Exp $
;;;

(select-module banyan)
(use gauche.mop.instance-pool)

;;----------------------------------------------------------------
;; Part
;;

;; Part has a list of nodes.
;;   * A node always belongs to one and only one part.  It is
;;     enforced by the initialize method of the node.
;;   * Part is the unit of check-in/out.
;;   * Part defines a namespace.  No two nodes in the part
;;     can have the same name.
;;   * Part can be constructed hierarchically, using '/' as a
;;     separator.  In a client, the hierarchical part names are
;;     just treated as a single string (like file pathname).
;;     It is a database side that interprets the hierarchy.
;;   * When a client is started, a special part "Untitled"
;;     is created as a default part, to keep new nodes.
;;     The untitled part will never saved as is---it has to be
;;     named properly before checked in.

;; Part revision:
;;   '() - this part is not committed yet.
;;   '(<integer> ...) - revisions.

;; Part imports:
;;   The slot 'imports keeps a list of parts that are unconditionally
;;   imported from the part, regardless whether the part's scene graph
;;   has a reference to the listed parts or not.  The method part-imports
;;   returns an actual list of external parts to be imported, which is
;;   a union of (ref part 'imports) and the result of scanning inter-part
;;   reference within the part's scene graph.
;;   Part's metadata has also :imports key, which lists _names_ of
;;   imported parts.

(define-class <part-meta> (<instance-pool-meta> <validator-meta>)
  ())

(define-method compute-get-n-set ((class <part-meta>) slot)
  (cond ((eq? (slot-definition-allocation slot) :metadata)
         (let* ((name (slot-definition-name slot))
                (meta-key (make-keyword name)))
           (next-method class
                        (list name
                              :allocation :virtual
                              :slot-ref
                              (lambda (o)
                                (get-keyword meta-key (ref o 'metadata) #f))
                              :slot-set!
                              (lambda (o v)
                                (update! (ref o 'metadata)
                                         (cut set-keyword meta-key <> v)))
                              ))))
        (else (next-method))))

(define-class <part> (<instance-pool-mixin>)
  ((name        :init-keyword :name   :accessor part-name)
   (nodes       :init-value ()        :accessor part-nodes)
   (repository  :init-keyword :repository :accessor  part-repository)
   (source-name :init-value #f :accessor part-source-name)
   (metadata    :init-value '())  ;kv-list, not includes 'metadata header
   (imports     :init-value '())  ;see the comment above
   (visibility  :init-value #t :accessor part-visibility
                :observer (lambda (o v)
                            (for-each
                             (lambda (n)
                               (when (is-a? n <renderable-node>)
                                 (slot-set! n 'temporary-visibility v)))
                             (part-nodes o))))
   (revision    :allocation :metadata)
   (exports     :allocation :metadata)
   (modified?   :init-value #f :getter modified?)
   (title       :allocation :metadata)
   (description :allocation :metadata)
   (modified-by :allocation :metadata)
   (modification-time :allocation :metadata)
   (created-by  :allocation :metadata)
   (creation-time :allocation :metadata)
   (contributed-by :allocation :metadata)
   (derived-from :allocation :metadata)
   (cc-license :allocation :metadata)
   )
  :metaclass <part-meta>)

(define-method initialize ((part <part>) initargs)
  (next-method)
  (unless (slot-bound? part 'repository)
    (set! (part-repository part) (current-repository)))
  )

(define (part? obj) (is-a? obj <part>))

(define-method write-object ((part <part>) out)
  (format out "#<part ~s>" (part-name part)))

;; used to display part names with additional information
(define-method part-descriptive-name ((part <part>))
  (if (ref part 'modified?)
      (format #f "~a (modified)" (part-name part))
      (part-name part)))

;;--------------------------------------------------------------
;; Parameters
;;

;; Default part
(define default-part
  (make-parameter (make <part> :name "*Untitled*" :repository #f)))

(define (default-part? part) (eq? part (default-part)))

(define (default-part->named-part newname)
  (let ((opart (default-part)))
    (set! (part-name opart) newname)
    (set! (part-repository opart) (current-repository))
    (default-part (make <part> :name "*Untitled*"))
    opart))

;; Current part
;;   NB: no observer should be attached directly here.  Use scene's
;;   part watcher hook.
(define current-part (make-parameter (default-part)))

;;---------------------------------------------------------------
;; Modification tracking
;;

;; Whenever a node is added or removed from the part, or the attribute
;; of the node within the part is modified, the part should be notified
;; by set-modified!.   It subsequently calls hook part-modified-watchers.
;; However, if the parameter part-track-modified is false, set-modified!
;; and unset-modified! does nothing.

;; Hook: part-modified-watchers
;;   Invoked as (HOOK PART) when PART is modified.
(define part-modified-watchers (make-hook 1))

;; Parameter: part-track-modified
(define part-track-modified (make-parameter #t))

(define-method set-modified! ((self <part>))
  (when (part-track-modified)
    (set! (ref self 'modified?) #t)
    (part-modified-watchers self)))

(define-method unset-modified! ((self <part>))
  (when (part-track-modified)
    (set! (ref self 'modified?) #f)
    (for-each (cut slot-set! <> 'modified? #f) (part-nodes self))
    (part-modified-watchers self)))

;;---------------------------------------------------------------
;; Node management
;;
(define-method add-node ((self <part>) node)
  (unless (memq node (part-nodes self))
    (push! (part-nodes self) node)
    (set-modified! self)))

(define-method delete-node ((self <part>) node)
  (update! (part-nodes self) (cut delete node <> eq?))
  (set-modified! self))

(define-syntax with-part
  (syntax-rules ()
    ((_ part . body)
     (parameterize ((current-part part)) . body))))

(define-method part-imports ((self <part>))
  (fold (lambda (node r)
          (fold (lambda (slot r)
                  (cond ((and-let* ((s (slot-definition-name slot))
                                    ((slot-bound? node s))
                                    (n (slot-ref node s))
                                    ((is-a? n <node>))
                                    (p (part-of n))
                                    ((not (eq? p self))))
                           (lset-adjoin eq? r p)))
                        (else r)))
                r
                (node-attribute-slots node)))
        (ref self 'imports)
        (part-nodes self)))

(define-method part-basename ((self <part>))
  (sys-basename (part-name self)))

(define (valid-part-name? string)
  (#/^[-_\w]+(\/[-_\w]+)*$/ string))

(define (name->part name)
  (instance-pool-find <part> (lambda (p) (equal? (part-name p) name))))

;;----------------------------------------------------------------
;; Unique name for node
;;

;; returns unique name within part, using given name as a candidate.
(define-method get-unique-name ((self <part>) name)
  (let1 name (if (string-null? name) "node-1" name) ;filter out null name
    (if (find (lambda (n) (equal? name (name-of n)))
              (part-nodes self))
        ;; Need to rename the name.
        ;; Temporary---this is awfully inefficient if the number of nodes
        ;; goes beyond 10^2 or so.
        (cond ((#/-(\d+)$/ name)
               => (lambda (m)
                    (get-unique-name self
                                     #`",(m 'before)-,(+ 1 (string->number (m 1)))")))
              (else (get-unique-name self #`",|name|-1")))
        name)))

;;----------------------------------------------------------------
;; Node serialization and deserialization protocol
;;

;; Serialization
;;  Note that the serialization doesn't cross the boundary of the
;;  part.
;;
;;  If the part contains other part, or the reference
;;  to the node which belongs to other part, it is written
;;  as "external node reference", or #,(node-ext-ref <part> [<name>]).
;;
;;  Where <part> is in relative path if it's under the current part,
;;  or absolute path otherwise.  <name> is the name of the node,
;;  if the reference is to the node under <part>.

(define-method node-serialize ((part <part>) port)

  (define context (make-hash-table))
  
  (define (write-value value)
    (cond
     ((part? value)
      (format port "#,(part-ref ~s)" (part-name value)))
     ((node? value)
      (cond ((hash-table-get context value #f)
             (format port "#,(node-ref ~s)" (name-of value)))
            ((eq? (part-of value) part)
             (hash-table-put! context value value)
             (write-node value))
            (else
             (format port "#,(node-ext-ref ~s ~s)"
                     (part-name (part-of value))
                     (name-of value)))
            ))
     ((list? value)
      (display "(" port)
      (for-each write-value value)
      (display ")" port))
     ((vector? value)
      (display "#(" port)
      (for-each write-value value)
      (display ")" port))
     (else
      (write value port) (newline port))))

  (define (write-node node)
    (format port "#,(node ~s ~s "
            (class-name (class-of node)) (name-of node))
    (for-each (lambda (slot)
                (format port " ~a " (slot-definition-name slot))
                (write-value (slot-ref node (slot-definition-name slot)))
                (display " " port))
              (node-attribute-slots node))
    (display ")\n" port))

  (format port "#,(part ~s " (part-name part))
  (for-each write-value (part-nodes part))
  (display ")\n" port)
  ;; NB: should it be here?
  (unset-modified! part)
  )

;; Deserialization is done in three passes, due to the inter-part name
;; resolution.
;;
;;   The first pass is just a simple 'read'.  During the first pass,
;;   SRFI-10 mechanism registers the node proxy.  Returns a part-proxy,
;;   which keeps the structure of the scene graph, while all nodes are
;;   still node proxies.   The connection to the server can be dropped
;;   here (it is important that we can drop connection each time after
;;   the first pass, for a scene can be consisted of lots of parts, and
;;   naive recursive fetching may consume large amount of network
;;   resources).
;;
;;   The second pass is intra-part resolution.  The table of node proxies
;;   are traversed and the nodes are realized.  At this point, each node
;;   is not interconnected.
;;
;;   The third pass is inter-part resolution.  The entire tree is
;;   traversed, replacing node proxies to nodes.

;; Dictionary of node proxies
(define node-deserializer-context (make-parameter #f))

;; A dynamic parameter to indicate the stage of deserializer
(define part-deserializing-stage (make-parameter #f))

(define-class <part-proxy> ()
  ((part :init-keyword :part)
   (node-proxies :init-keyword :node-proxies) ; toplevel nodes
   (metadata)
   (context) ; dictionary - maps internal node name -> proxy
             ; external nodes are registered using "part:node-name" as the key
   (remote-proxies :init-value ())
   ))

(define-class <node-proxy> ()
  ((class :init-keyword :class :init-value #f)
   (name  :init-keyword :name)
   (part  :init-value #f) ;; if this part is external
   (initargs :init-keyword :initargs :init-value '())
   (node  :init-value #f)
   (initialized? :init-value #f)
   ))

;; Pass 1 ---------------------------------------------
(define (node-deserialize-read port)
  (parameterize ((node-deserializer-context (make-hash-table 'string=?))
                 (part-deserializing-stage 1))
    (let1 part-proxy (read port)
      (set! (ref part-proxy 'context) (node-deserializer-context))
      part-proxy)))

(define-reader-ctor 'node
  (lambda (class name . args)
    (let1 proxy (get-node-proxy name)
      (set! (ref proxy 'class) (eval class (find-module 'banyan))) ;Arrrggh..
      (set! (ref proxy 'initargs) args)
      proxy)))

(define-reader-ctor 'part
  (lambda (name . args)
    (make <part-proxy>
      :part (make <part> :name name :source-name name)
      :node-proxies args)))

(define-reader-ctor 'node-ref
  (lambda (name)
    (get-node-proxy name)))

(define-reader-ctor 'node-ext-ref
  (lambda (part name)
    (let1 proxy (get-node-proxy #`",|part|:,|name|" name)
      (set! (ref proxy 'class) #f)
      (set! (ref proxy 'part) part)
      proxy)))

(define (get-node-proxy key . maybe-real-name)
  (let-optionals* maybe-real-name ((name key))
    (or (and-let* ((context (node-deserializer-context)))
          (or (hash-table-get context key #f)
              (let1 proxy (make <node-proxy> :name name)
                (hash-table-put! context key proxy)
                proxy)))
        (error "#,(node ...) appears out of context"))))

;; Pass 2 ----------------------------------------------
(define (node-deserialize-realize part-proxy)

  (define (realize-node key proxy)
    (unless (ref proxy 'node)
      (cond
       ((ref proxy 'class)
        => (lambda (class)
             (set! (ref proxy 'node)
                   (initiate class :name (ref proxy 'name)))))
       ((not (ref proxy 'part))
        (errorf "Node ~s doesn't have the definition" key))
       (else
        (slot-push! part-proxy 'remote-proxies proxy))
       )))

  (parameterize ((part-track-modified #f)
                 (part-deserializing-stage 2))
    (with-part (ref part-proxy 'part)
      (hash-table-for-each (ref part-proxy 'context) realize-node)
      (slot-set! part-proxy 'context #f) ;; we don't need context anymore
      )))

;; Pass 3 ---------------------------------------------
(define (node-deserialize-restructure part-proxy)

  (define (realize-remote-node proxy)
    (let1 part (name->part (ref proxy 'part))
      (unless part
        (error "reference from unknwon part:" (ref proxy 'part)))
      (let* ((name (ref proxy 'name))
             (node (find (lambda (n) (equal? (name-of n) name))
                         (part-nodes part))))
        (unless node
          (errorf "reference of unexistent node ~a in part ~a"
                  name (ref proxy 'part)))
        (set! (ref proxy 'node) node))))

  (define (proxy->node proxy)
    (unless (or (ref proxy 'initialized?)
                (ref proxy 'part)) ;; don't initialize remote node reference
      (node-initialize proxy))
    (ref proxy 'node))
  
  (define (node-initialize proxy)
    (let loop ((initargs (ref proxy 'initargs))
               (r '()))
      (if (null? initargs)
          (let1 node (ref proxy 'node)
            (for-each (cut apply slot-set! node <>) (reverse r))
            (slot-set! proxy 'initialized? #t)
            node)
          (loop (cddr initargs)
                (cons (list (car initargs)
                            (rec (cadr initargs)))
                      r)))
      ))

  (define (rec tree)
    (cond ((list? tree) (map rec tree))
          ((vector? tree) (map-to <vector> rec tree))
          ((is-a? tree <node-proxy>) (proxy->node tree))
          (else tree)))

  (parameterize ((part-deserializing-stage 3))
    (for-each realize-remote-node (ref part-proxy 'remote-proxies))
    (let ((part (ref part-proxy 'part))
          (metadata (ref part-proxy 'metadata)))
      (slot-set! part 'nodes (map rec (ref part-proxy 'node-proxies)))
      (slot-set! part 'metadata metadata)
      (slot-set! part 'imports
                 (map (lambda (name)
                        (or (name->part name)
                            (error "reference to unknown part:" name)))
                      (get-keyword :imports metadata '())))
      part)))

(provide "banyan/part")
