;;;
;;; banyan.repository - repository
;;;
;;;  Copyright(C) 2002-2003 by Shiro Kawai (shiro@acm.org)
;;;
;;;  Permission to use, copy, modify, distribute this software and
;;;  accompanying documentation for any purpose is hereby granted,
;;;  provided that existing copyright notices are retained in all
;;;  copies and that this notice is included verbatim in all
;;;  distributions.
;;;  This software is provided as is, without express or implied
;;;  warranty.  In no circumstances the author(s) shall be liable
;;;  for any damages arising out of the use of this software.
;;;
;;;  $Id: repository.scm,v 1.39 2003/02/22 05:04:47 shirok Exp $
;;;

(select-module banyan)
(use srfi-1)
(use srfi-2)
(use rfc.uri)
(use rfc.http)
(use rfc.base64)
(use file.util)
(use gauche.parameter)

;; PROTOTYPE IMPLEMENTATION

;; Provisional file format
;;  File is dumped as two S-exprs.  The first one is metadata,
;;  and the second is a serialized form of a part.
;;  The metadata part doesn't contain any special srfi-10 syntax---
;;  it can be read by plain Scheme, so that the script to process
;;  files can be easily written.
;;
;; Data transfer over HTTP
;;  Media type "application/x-banyan" is used to exchange the format.
;;  Specify the media type to "Accept" header field in GET request to
;;  download the file.  (Otherwise, the repository server returns an
;;  HTML document that describes the property of the data).
;;  Optionally you can attach "info" query tag to the request URI to
;;  get only the metadata part.
;;
;;  To check in the data, give the media type to "Content-Type" header
;;  field in POST request, and sends both the metadata and serialized
;;  data as the message body (not multipart).
;;  The reply of POST request is the updated metadata.
;;  Again, attaching "info" to the request URI allows updating only
;;  the metadata, although it is not used yet.

;;==================================================================
;; <repository>
;;

;; Provides interface for repository.

;; A repository is the root of the tree of the scene database.
;; It's sort of CVS's CVSROOT.  One repository can have arbitrary
;; number of hiearachy of parts.
;; A repository can be located either in local filesystem or on the net.

(define-class <repository> (<instance-pool-mixin>)
  ((uri  :init-keyword :uri :accessor uri-of)
   (user :init-keyword :user :initform "guest") ;; default
   (password :init-keyword :password :initform "guest") ;; default
   (default :init-keyword :default :init-value #f)
   (logged-in :init-value #f)           ;true once logged in
   ))

;; subclass is expected to implement those methods:
;;   repository-login <repository>
;;       Attempt to login to <repository>, using values of user and password
;;       slots.  Calls repository-login-handler repeatedly until login
;;       succeeds.  Returns #t if logged in, or #f if not.
;;   repository-exists? <repository>
;;       Check whether the repository really exists at the uri.
;;   repository-create <repository>
;;       Creates a repository at the uri of <repository> if it doesn't
;;       exist.  This is unlikely to work on http repository.
;;   repository-parts <repository> [<path>]
;;       Returns a list of parts under <path> in <repository>.
;;   repository-commit <repository> <part> <log>
;;       Commit <part> to <repository>, with log message <log>
;;   repository-check-out <repository> <path>
;;       Check-out the part of <path> in <repository>.

;;============================================================
;; Parameters
;;

;; Keep the current repository.
(define current-repository (make-parameter #f))

;; List of pre-defined repositories.
(define standard-repositories (make-parameter '()))

;; Repository-login-handler is called when repository-login method
;; needs to ask the user to get login name and password for the given
;; repository.  If the value is a procedure, it takes a repository
;; and a context, and expected to set user and password slots.
;; If #f, repository-login method exits if it doesn't know login name
;; and password (assuming non-interactive use).
;;
;; The second argument (context) may be #f (normal login procedure),
;; 'bad-user (retry when the user name is not recognized), or 'bad-pass
;; (retry when the password doesn't match).
(define repository-login-handler (make-parameter #f))

;;--------------------------------------------------------------
;; Subclasses
;;
(define-class <file-repository> (<repository>)
  ((path :init-keyword :path :accessor path-of)
   ))

(define-class <http-repository> (<repository>)
  ((host :init-keyword :host :accessor host-of)
   (port :init-keyword :port :accessor port-of)
   (server :init-keyword :server :accessor server-of)  ; host:port
   (path :init-keyword :path :accessor path-of)
   ))

;; Create <repository> instance at uri.  the name 'open-repository' is
;; a bit misleading---this doesn't connect to the actualy repository.
;; merely it checks the syntax of uri and sets up some initial values
;; for the slots.
;; If a repository with the same URI has already been opened, it is
;; returned.
(define (open-repository uri . args)
  (or
   (instance-pool-find <repository>
                       (lambda (r) (string=? uri (ref r 'uri))))
   (receive (proto path) (uri-scheme&specific uri)
     (cond
      ((string=? proto "file")
       (apply make <file-repository> :uri uri :path path args))
      ((string=? proto "http")
       (receive (serv p query label) (uri-decompose-hierarchical path)
         (unless serv (error "bad URI" uri))
         (receive (user host port) (uri-decompose-authority serv)
           (unless host (error "bad URI" uri))
           (apply make <http-repository>
                  :uri uri
                  :path #`",|p|,(or query \"\"),(or label \"\")"
                  :host host :port (or port 80)
                  :server #`",|host|:,(or port 80)"
                  args))))
      (else
       (error "unsupported protocol" uri))
      ))
   ))

;;--------------------------------------------------------------
;; resource path
;;
;; A resource-path is a URI to point a specific part data.
;; It can be relative pathname, an absolute pathname, or
;; a fully-qualified URI.  In the first two cases, the pathname
;; is used within the current repository.

;; Decompose resource path to repository and path
(define (decompose-resource-path rpath . args)
  (let-optionals* args ((repository   (current-repository))
                        (current-path "/"))
    ;;  (split-path-at "foo" "/bar/foo/woo/boo") => "/bar/foo", "woo/boo"
    ;;  (split-path-at "foo" "/bar/woo/boo") => #f, #f
    ;;  (split-path-at "foo" "/bar/woo/foo") => "/bar/woo/foo", ""
    (define (split-path-at component path)
      (receive (before after) (string-scan path #`"/,component" 'after*)
        (cond ((not before) (values #f #f))
              ((string-null? after) (values before ""))
              ((string-prefix? "/" after) (values before (string-drop after 1)))
              (else (values #f #f)))))

    (define (sanitize-path path)
      (let1 cpath (sys-normalize-pathname path :canonicalize #t)
        (when (member ".." (string-split cpath #\/))
          (error "bad resource path" rpath))
        cpath))

    (receive (proto ipath) (uri-scheme&specific rpath)
      (cond
       ((not proto)                     ;relative
        (unless repository
          (error "can't determine repository of resource-path" rpath))
        (if (absolute-path? ipath)
            (values repository (string-drop ipath 1))
            (values repository
                    (string-drop (sanitize-path (build-path current-path ipath)) 1))))
       ((member proto '("http" "https" "file"))
        (receive (repo path) (split-path-at "repository" rpath)
          (unless repo (error "bad resource path" rpath))
          (values (open-repository repo) path)))
       (else
        (error "unsupported URI scheme" proto))
       ))
    ))

;;--------------------------------------------------------------
;; A convenience function to load parts specified by the resource path.
;;

(define (repository-load-parts-by-path resource-path)
  (receive (repo path) (decompose-resource-path resource-path)
    (repository-check-out repo path)))

;;--------------------------------------------------------------
;; common function to store repository
;;

(define-method ensure-repository ((repo <repository>))
  (unless (repository-exists? repo)
    (repository-create repo))
  repo)

(define (next-revision revision)
  (if (pair? revision)
      (let loop ((rev revision)
                 (r '()))
        (if (null? (cdr rev))
            (reverse! (cons (+ (car rev) 1) r))
            (loop (cdr rev) (cons (car rev) r))))
      '(1 1)))

(define (revise-metadata part log)
  (let* ((ometa (ref part 'metadata))
         (now   (sys-time))
         (me    (ref (current-repository) 'user))
         (rev   (next-revision (get-keyword :revision ometa #f)))
         )
    (list
     :name (part-name part)
     :revision rev
     :description (get-keyword :description ometa "")
     :modified-by me
     :modification-time now
     :created-by (or (get-keyword :created-by ometa #f) me)
     :creation-time (or (get-keyword :creation-time ometa #f) now)
     :exports (or (get-keyword :exports ometa #f) #t)
     :imports (map part-name (part-imports part))
     :history (cons (list
                     :revision rev :modified-by me :modification-time now
                     :log log)
                    (get-keyword :history ometa '()))
     )))

(define (repository-write-part out part log)
  (let1 newmeta (revise-metadata part log)
    (write (cons 'metadata newmeta) out)
    (newline out)
    (node-serialize part out)
    (car (list newmeta)) ;;TODO: suspicious compiler bug - need to check
    ))

;; Start from the set of parts, calls WRITER for all the
;; parts imported from the specified parts.
(define (repository-write-all-parts parts writer)
  (let loop ((parts parts)
             (written '()))
    (unless (null? parts)
      (if (memq (car parts) written)
          (loop (cdr parts) written)
          (let1 imports (part-imports (car parts))
            (writer (car parts))
            (loop (lset-union eq? parts imports)
                  (cons (car parts) written))))
      )
    ))

(define (repository-check-logged-in repo)
  (unless (ref repo 'logged-in)
    (set! (ref repo 'logged-in) (repository-login repo))))

;; Returns part proxy and list of imported part names.
(define (repository-read-part in)
  (let* ((metadata (read in))
         (part-proxy (node-deserialize-read in)))
    (set! (ref part-proxy 'metadata) (cdr metadata))
    (values part-proxy
            (get-keyword :imports (cdr metadata) '()))))

;; Start from the set of part names, repeatedly calls READER to read
;; the parts (and their related parts).  Reader takes a part name
;; and returns part-proxy and a list of imported part names.
(define (repository-read-all-parts names reader)
  (let loop ((part-names names)
             (part-proxies '()))
    (cond
     ((null? part-names)
      (for-each node-deserialize-realize part-proxies)
      (map node-deserialize-restructure (reverse part-proxies)))
     (else
      (receive (proxy imports) (reader (car part-names))
        (loop (lset-union equal? (cdr part-names) imports)
              (cons proxy part-proxies)))))
    ))

;;--------------------------------------------------------------
;; <file-repository> implementation
;;

(define-method repository-login ((repo <file-repository>))
  ;; we don't need to login
  #t)

(define-method repository-exists? ((repo <file-repository>))
  (and (file-exists? (path-of repo))
       (file-is-directory? (path-of repo))))

(define-method repository-parts ((repo <file-repository>) . maybe-path)
  (let* ((path (get-optional maybe-path #f))
         (dir  (if path
                   (build-path (path-of repo) (string-drop path 1))
                   (path-of repo))))
    (reverse!
     (fold
      (lambda (entry knil)
        (cond ((or (string=? entry ".") (string=? entry "..")) knil)
              ((not (file-is-directory? (build-path dir entry))) knil)
              (else
               (let ((subdir? (has-subdirectory? (build-path dir entry)))
                     (part? (file-exists? (build-path dir entry "part.bn"))))
                 (if (or subdir? part?)
                     (cons (list entry :part part? :subdir subdir?) knil)
                     knil)))))
      '()
      (sort (sys-readdir dir))))))

(define (has-subdirectory? path)
  (let loop ((entries (sys-readdir path)))
    (cond
     ((null? entries) #f)
     ((or (string=? (car entries) ".") (string=? (car entries) ".."))
      (loop (cdr entries)))
     ((file-is-directory? (build-path path (car entries))) #t)
     (else (loop (cdr entries))))))

(define-method repository-create ((repo <file-repository>))
  (make-directory* (path-of repo)))
                  
(define-method repository-commit ((repo <file-repository>)
                                  (part <part>)
                                  log)
  (ensure-repository repo)
  (let* ((name (part-name part))
         (dir  (build-path (path-of repo) name))
         (file (build-path dir "part.bn")))
    (make-directory* dir)
    (let ((nmeta (call-with-temporary-output-file file
                   (lambda (port)
                     (repository-write-part port part log)))))
      (set! (ref part 'metadata) nmeta))))

(define-method repository-check-out ((repo <file-repository>) name)
  (repository-read-all-parts
   (list name)
   (lambda (name)
     (let ((file (build-path (path-of repo) name "part.bn")))
       (unless (file-exists? file)
         (errorf "no such part ~s in the repository ~s" name (ref repo 'uri)))
       (call-with-input-file file repository-read-part)))))
       
;;--------------------------------------------------------------
;; <http-repository> implementation
;;

(define (http-get-header-field name headers default)
  (cond ((assoc name headers) => cadr)
        (else default)))

(define (http-parse-message-body body reader)
  (if (and body reader)
      (with-error-handler
          (lambda (e)
            (errorf "premature http reply (reader error: ~a)" (ref e 'message)))
        (lambda () (reader body)))
      #t))

(define (http-compose-auth repo)
  (let ((user (ref repo 'user))
        (pass (ref repo 'password)))
    (if (and user pass)
        `(:cookie
          ,(string-append "$Version=1;banyan-auth="
                          (base64-encode-string
                           (string-append user "\x00" pass))))
        '())))

(define (http-repo-access repo proc error-on-fail? reader . args)
  (receive (stat headers body)
      (apply proc (append args
                          `(:user-agent "Banyan/0.0"
                            :accept "application/x-banyan"
                            ,@(http-compose-auth repo))))
    (if (string=? stat "200")
        (let1 status (http-get-header-field "x-banyan-status" headers "ok")
          (if (equal? status "error")
              (if body
                  (error (cadr (read-from-string body)))
                  (error "server error"));; TODO: better handling
              (http-parse-message-body body reader)))
        (if error-on-fail?
            (errorf "http error (code ~a)" stat)
            #f))))

;; Login strategy:
;;   For now, I take easy way.  If username and/or password is not given
;;   in .banyanrc, access as a guest (no attempt to login).
;;   If they both are given, attemt to login with them, and signals an
;;   error if failed (repository-login-handler is not called yet).
(define-method repository-login ((repo <http-repository>))

  (define (try-login)
    (let1 r (http-repo-access repo http-get #f read-from-string
                              (server-of repo)
                              #`",(path-of repo)?login")
      (cond ((null? r) 'ok)
            ((and (pair? r) (eq? (car r) 'bad-auth) (pair? (cdr r)))
             (cadr r))
            (else (error "protocol error during login")))))
  
  (if (and (ref repo 'user) (ref repo 'password))
      (case (try-login)
        ((ok) #t)
        ((bad-user)
         (errorf "login failed: no such user: ~s" (ref repo 'user)))
        ((bad-pass)
         (errorf "login failed: bad password for user: ~s" (ref repo 'user)))
        (else
         (errorf "login failed")))
      #f))

(define-method repository-exists? ((repo <http-repository>))
  (http-repo-access repo http-head #f #f (server-of repo)
                    #`",(path-of repo)?list"))

(define-method repository-parts ((repo <http-repository>) . maybe-path)
  (let1 path (get-optional maybe-path "")
    (http-repo-access repo http-get #t read-from-string
                      (server-of repo)
                      #`",(path-of repo),|path|?list")
    ))

(define-method repository-create ((repo <http-repository>))
  (error "not supported yet"))

(define-method repository-check-out ((repo <http-repository>) name)
  (repository-read-all-parts
   (list name)
   (lambda (name)
     (http-repo-access repo http-get #t
                       (cut call-with-input-string <> repository-read-part)
                       (server-of repo)
                       (build-path (path-of repo) name)))))

(define-method repository-commit ((repo <http-repository>)
                                  (part <part>)
                                  log)
  (let* ((name  (part-name part))
         (out   (open-output-string))
         (nmeta (repository-write-part out part log)))
    (http-repo-access repo http-post #t read-from-string
                      (server-of repo)
                      (build-path (path-of repo) name)
                      (get-output-string out))
    (set! (ref part 'metadata) nmeta)))

(provide "banyan/repository")
