;;;
;;; banyan/sandbox - provides sandbox environment to evaluate node expression.
;;;
;;;  Copyright(C) 2003 by Shiro Kawai (shiro@acm.org)
;;;
;;;  Permission to use, copy, modify, distribute this software and
;;;  accompanying documentation for any purpose is hereby granted,
;;;  provided that existing copyright notices are retained in all
;;;  copies and that this notice is included verbatim in all
;;;  distributions.
;;;  This software is provided as is, without express or implied
;;;  warranty.  In no circumstances the author(s) shall be liable
;;;  for any damages arising out of the use of this software.
;;;
;;;  $Id: sandbox.scm,v 1.4 2003/01/13 03:51:36 shirok Exp $
;;;

(select-module banyan)

;; PROTOTYPE IMPLEMENTATION

;;===============================================================
;; Sandbox module
;; 

;; NB: The current implementation depends on Gauche's feature of 
;; allowing define to transfer syntactic bindings.  It is not a
;; formally supported feature so this code will be required to
;; change when Gauche changes it.

(define-module banyan.sandbox
  (use srfi-1)
  (use srfi-2)
  (use srfi-13)
  (use srfi-14)
  (export hiki hiki-koe kapu)

  ;; (hiki MODULE SYMBOL ...)
  ;; Explicitly enables bindings of SYMBOL from MODULE
  (define-syntax hiki
    (syntax-rules ()
      ((_ module) #f)
      ((_ module name . names)
       (begin
         (define name (with-module module name))
         (export name)
         (hiki module . names)))
      ))

  ;; (hiki-koe MODULE SYMBOL ...)
  ;; Enables all bindings in MODULE, except SYMBOL.
  (define-macro (hiki-koe module . names)
    (define (fold-bindings module proc knil)
      (let* ((mod     (find-module module))
             (exports (module-exports mod)))
        (if (pair? exports)
            (fold (lambda (name knil)
                    (if (memq name names) knil (cons (proc name) knil)))
                  knil
                  exports)
            (hash-table-fold (module-table mod)
                             (lambda (name val knil)
                               (if (memq name names)
                                   knil
                                   (cons (proc name) knil)))
                             knil)
            )))
    `(begin
       ,@(fold-bindings module
                        (lambda (symbol)
                          `(define ,symbol (with-module ,module ,symbol)))
                        '())))
  
  ;; (kapu SYMBOL ...)
  ;; Disable binding of SYMBOL ...
  (define-syntax kapu
    (syntax-rules ()
      ((_) #f)
      ((_ name . names)
       (begin
         (define-macro (name . args)
           (errorf "~a can't be used within sandbox module"
                   (unwrap-syntax 'name)))
         (kapu . names)))
      ))
  )

;; Creates a fresh module, in which one can't do harmful things.
;; All bindings sits inside the created module, 
;; is excluded.

(define (make-sandbox-module)
  (let ((m (make-module #f)))
    (eval
     '(begin
        (import banyan.sandbox)
        (hiki-koe null
                  select-module with-module)
        (hiki-koe scheme
                  open-input-file open-output-file
                  call-with-input-file call-with-output-file
                  with-input-from-file with-output-to-file
                  load eval transcript-on transcript-off
                  null-environment scheme-report-environment 
                  interaction-environment)
        (hiki-koe srfi-1)
        (hiki-koe srfi-2)
        (hiki-koe srfi-13)
        (hiki-koe srfi-14)
        )
     m)
    ;; Now, this resets module precedence list of m to null, voiding
    ;; all bindings but explicitly incorporated ones.
    (eval '(extend) m)
    m))

(provide "banyan/sandbox")
