;;;
;;; banyan/scene - scene manager
;;;
;;;  Copyright(C) 2002-2003 by Shiro Kawai (shiro@acm.org)
;;;
;;;  Permission to use, copy, modify, distribute this software and
;;;  accompanying documentation for any purpose is hereby granted,
;;;  provided that existing copyright notices are retained in all
;;;  copies and that this notice is included verbatim in all
;;;  distributions.
;;;  This software is provided as is, without express or implied
;;;  warranty.  In no circumstances the author(s) shall be liable
;;;  for any damages arising out of the use of this software.
;;;
;;;  $Id: scene.scm,v 1.28 2003/01/22 03:15:20 shirok Exp $
;;;

(select-module banyan)
(use util.toposort)

;; PROTOTYPE IMPLEMENTATION

;;; <scene> is a client-side world in which nodes are realized.
;;; It has a global knowledge of the local rendered world,
;;; bridges between user input and objects in the world
;;; (including selections),  and maintains time in the local world.
;;; In the interactive mode, a scene also maintains the default
;;; client camera with which the user can walkthrough the local world.
;;;
;;; A client can have multiple scenes, but there's
;;; only one 'current' scene at a moment in each client.
;;; A parameter current-scene keeps the current scene.
;;;
;;; [Scene and part]: Scene and part are orthogonal.  A scene
;;; can have multiple parts, and a part can belong to multiple
;;; scenes.  When a new, empty scene is created, a new part is
;;; created and that becomes the default part for the scene.
;;; Scene can import other parts into it, and the scene's default
;;; part switches to the newly imported part.

;;------------------------------------------------------------------
;; Class <scene>
;;

(define-class <scene> (<transient-node> <renderable-node>)
  ((instances  :init-keyword :instances :accessor instances-of
               :attribute-type <list>
               :initform '())
   ;; - a list of instances.
   (cameras    :init-keyword :cameras :accessor cameras-of
               :initform '())
   ;; - a list of cameras.
   (selections :accessor selections-of :initform '())
   ;; - a stack of selection lists.
   (selection-stack :accessor selection-stack-of :initform '())
   ;; - a list of selections
   (updaters   :accessor updaters-of :init-value #f)
   ;; - a topologically sorted updaters list.  #f means need to calculate.
   (id->instance :initform (make-hash-table 'eqv?))
   (instance->id :initform (make-hash-table 'eq?))
   (max-id       :initform 0)
   ;; - slots that maintain the mapping between selection ids and instances
   (parts  :init-keyword :parts)
   ;; - list of parts this scene maintains.
   (default-part :init-keyword :default-part)
   ;; - the default part for this scene.
   (sandbox :init-value #f)
   ;; - The sandbox in which expressions are evaluated.
   ;;   See eval-expr below.
  ))

(define-method initialize ((scene <scene>) initargs)
  (next-method)
  ;; sets up parts.
  (unless (slot-bound? scene 'parts)
    (set! (ref scene 'parts) (list (default-part))))
  (if (slot-bound? scene 'default-part)
      ;; consistency check
      (unless (memq (ref scene 'default-part)
                    (ref scene 'parts))
        (slot-push! scene 'parts
                    (ref scene 'default-part)))
      ;; use the first part as the default one.
      (set! (ref scene 'default-part)
            (car (ref scene 'parts)))))

;;------------------------------------------------------------------
;; current scene parameter
;;

(define current-scene (make-parameter #f))

(parameter-observer-add! current-scene
                         (lambda (old new)
                           (current-part (ref new 'default-part)))
                         'before)

(parameter-observer-add! current-scene
                         (lambda (old new)
                           (unless (eq? old new)
                             (part-watchers new)
                             (selection-watchers new (selections-of new))))
                         'after)

;; Hook part-watchers are called whenever the current scene
;; changes the default part.

(define part-watchers (make-hook 1))

;;-----------------------------------------------------------------
;; New scene
;;

;;; Creates an empty, blank scene.
(define (new-scene . args)
  (let1 scene (apply make <scene> args)
    (current-scene scene)
    (make-default-camera)
    scene))

;;------------------------------------------------------------------
;; Adding and removing instance
;;

(define-method add-instance ((scene <scene>) instance)
  (let1 id (ref scene 'max-id)
    (inc! (ref scene 'max-id))
    (slot-push! scene 'instances instance)
    (hash-table-put! (ref scene 'id->instance) id instance)
    (hash-table-put! (ref scene 'instance->id) instance id)
    ))

(define-method delete-instance ((scene <scene>) instance)
  (update! (instances-of scene) (cut delete instance <> eq?))
  (when (memq instance (selections-of scene))
    (remove-selection scene instance))
  ;; fix id<->instance mapping table
  (let1 id (hash-table-get (ref scene 'instance->id) instance #f)
    (hash-table-delete! (ref scene 'instance->id) instance)
    (when id (hash-table-delete! (ref scene 'id->instance) id)))
  ;; TODO: handle shared transform/shape/material.
  ;; This is just for now...
  (delete-node (part-of instance) instance)
  (cond ((transform-of instance)
         => (lambda (n) (delete-node (part-of n) n))))
  (cond ((object-of instance)
         => (lambda (n) (delete-node (part-of n) n))))
  )

(define-method delete-instance-selected ((scene <scene>))
  (let ((sel (selections-of scene)))
    (reset-selections scene) ;; to call selection callback only once
    (for-each (cut delete-instance scene <>) sel)))

(define-method filter-instances ((scene <scene>) pred)
  (filter pred (instances-of scene)))

;;------------------------------------------------------------------
;; Adding and removing parts
;;

;; internal common routine
(define (scene-add-part-int scene part)
  (slot-push! scene 'parts part)
  (for-each (lambda (node)
              (when (is-a? node <instance>) (add-instance scene node)))
            (part-nodes part))
  ;; special treatment for unnamed part 
  (and-let* ((parts  (ref scene 'parts))
             (unnamed (find (cute eq? <> (default-part)) parts)))
    (slot-set! scene 'parts (delete! unnamed parts eq?)))
  )

(define-method add-part ((scene <scene>) (part <part>))
  (scene-add-part-int scene part)
  (slot-set!  scene 'default-part part)
  (current-part part)
  ;; call observers
  (part-watchers scene)
  )

(define-method add-parts ((scene <scene>) parts)
  (unless (null? parts)
    (cond ((find (lambda (p) (not (is-a? p <part>))) parts)
           => (lambda (e) (error "part required, but got" e))))
    (let1 root (car parts)
      (for-each (cut scene-add-part-int scene <>) parts)
      (slot-set! scene 'default-part root)
      (current-part root)
      (part-watchers scene)
      )))

(define-method make-part-current ((scene <scene>) (part <part>))
  (unless (memq part (ref scene 'parts))
    (error "make-part-current: part ~a is not in scene ~a" part scene))
  (slot-set! scene 'default-part part)
  (current-part part)
  (part-watchers scene)
  )

;;------------------------------------------------------------------
;; Adding and removing camera
;;

(define-method add-camera! ((scene <scene>) camera)
  (slot-push! scene 'cameras camera))

(define-method delete-camera! ((scene <scene>) camera)
  (update! (cameras-of scene) (cut delete camera <> eq?)))

;;------------------------------------------------------------------
;; Managing selections
;;

;; Hook selection-watchers are called whenever
;;   - Selection of the current scene is changed
;;   - The current scene is changed
;; Arguments are scene and new selection.

(define selection-watchers (make-hook 2))

(define-method reset-selections ((scene <scene>))
  (let* ((sel (selections-of scene))
         (change? (not (null? sel))))
    (when change?
      (for-each instance-deselect sel)
      (set! (selections-of scene) '())
      (selection-watchers scene '()))))

(define-method set-selection ((scene <scene>) . instances)
  (let1 old (selections-of scene)
    (set! (selections-of scene) instances)
    (for-each instance-deselect old)
    (for-each instance-select instances)
    (selection-watchers scene instances)))

(define-method add-selection ((scene <scene>) instance)
  (let ((sel (selections-of scene))
        (change? (not (memq instance scene))))
    (when change? (set! (selections-of scene) (cons instance sel)))
    (instance-select instance)
    (when change? (selection-watchers scene (selections-of scene)))))
  
(define-method remove-selection ((scene <scene>) instance)
  (let1 change? (memq instance (selections-of scene))
    (when change? (update! (selections-of scene) (cut delete instance <> eq?)))
    (instance-deselect instance)
    (when change? (selection-watchers scene (selections-of scene)))))

(define-method center-of-selections ((scene <scene>))
  (let1 center (vector4f 0.0 0.0 0.0)
    (for-each (lambda (instance)
                (vector4f-add! center
                               (translation-of
                                (transform-of instance))))
              (selections-of scene))
    center))

;; Selection stack -----------------------------------------------
(define-method selection-stack-depth ((scene <scene>))
  (length (selection-stack-of scene)))

(define (adjust-instance-selections stack)
  (for-each (lambda (level set)
              (for-each (cut instance-select <> level) set))
            (iota (length stack) (length stack) -1)
            (reverse stack)))

(define-method push-selections ((scene <scene>))
  (let ((sel (selections-of scene)))
    (slot-push! scene 'selection-stack sel)
    (adjust-instance-selections (selection-stack-of scene))
    (set! (selections-of scene) '())
    (selection-watchers scene '())))

(define-method pop-selections ((scene <scene>))
  (unless (null? (selection-stack-of scene))
    (for-each instance-deselect (selections-of scene))
    (let ((sel (pop! (selection-stack-of scene))))
      (set! (selections-of scene) sel)
      (adjust-instance-selections (selection-stack-of scene))
      (for-each instance-select sel)
      (selection-watchers scene sel))))

(define-method swap-selections ((scene <scene>))
  (let ((sel (selections-of scene))
        (top (pop! (selection-stack-of scene))))
    (slot-push! scene 'selection-stack sel)
    (set! (selections-of scene) top)
    (for-each instance-deselect sel)
    (adjust-instance-selections (selection-stack-of scene))
    (for-each instance-select top)
    (selection-watchers scene top)))

(define-method peek-selection-stack ((scene <scene>))
  (let1 s (selection-stack-of scene)
    (if (null? s) s (car s))))

(define-method clear-selection-stack ((scene <scene>))
  (for-each (cut for-each instance-deselect <>) (selection-stack-of scene))
  (set! (selection-stack-of scene) '()))

;; Mapping selection id and instance -----------------------------
(define-method id->instance ((scene <scene>) id)
  (hash-table-get (ref scene 'id->instance) id #f))

(define-method instance->id ((scene <scene>) instance)
  (hash-table-get (ref scene 'instance->id) instance #f))

;; Convenience function
(define (selections)
  (selections-of (current-scene)))

(define (current-selected-object)
  (let1 selected (selections-of (current-scene))
    (if (null? selected) #f (ref (car selected) 'object))))

;;------------------------------------------------------------------
;; Render scene
;;

(define-method node-render ((scene <scene>) op)
  (for-each (cut node-render <> op) (instances-of scene)))

;;------------------------------------------------------------------
;; Update scene
;;

(define-method sort-updaters ((scene <scene>))
  ;; OK, this is a sloppy algorithm.  Just for the time being...
  (let* ((all-nodes (append-map part-nodes
                                (cons (transient-part) (ref scene 'parts))))
         (branches (map (lambda (n)
                          (cons n (node-depends n)))
                        all-nodes)))
    (if (null? branches)
        #f
        (reverse! (filter (cut is-a? <> <updatable-node>)
                          (topological-sort branches))))))

(define-method node-update ((scene <scene>))
  (unless (updaters-of scene)
    (set! (updaters-of scene) (sort-updaters scene)))
  (when (updaters-of scene)
    (for-each node-update (updaters-of scene))))

;; Node would notify the scene when the graph has been changed, thus
;; updaters list should be recalculated.  For the time being,
;; this invalidates the current updaters list, forcing the next
;; redraw to reconstruct the list.   Later, more efficient algorithm
;; should be implemented.
(define-method notify-graph-changed ((scene <scene>))
  (set! (updaters-of scene) #f))

(define-method notify-graph-changed ((top <top>)) #f) ;;fallback

;;------------------------------------------------------------------
;; Committing to the repository
;;

(define-method scene-commit ((scene <scene>) log)
  (for-each (lambda (part)
              (when (modified? part)
                (repository-commit (current-repository) part log)))
            (ref scene 'parts)))

;;------------------------------------------------------------------
;; Expression evaluation
;;

(define-method eval-expr ((scene <scene>) expr)
  (let ((sandbox (or (ref scene 'sandbox)
                     (let1 m (make-sandbox-module)
                       (set! (ref scene 'sandbox) m)
                       m))))
    (eval expr m)))

(provide "banyan/scene")
