;;;
;;; banyan.shader - shader
;;;
;;;  Copyright(C) 2002-2003 by Shiro Kawai (shiro@acm.org)
;;;
;;;  Permission to use, copy, modify, distribute this software and
;;;  accompanying documentation for any purpose is hereby granted,
;;;  provided that existing copyright notices are retained in all
;;;  copies and that this notice is included verbatim in all
;;;  distributions.
;;;  This software is provided as is, without express or implied
;;;  warranty.  In no circumstances the author(s) shall be liable
;;;  for any damages arising out of the use of this software.
;;;
;;;  $Id: shader.scm,v 1.8 2003/01/01 00:59:50 shirok Exp $
;;;

(select-module banyan)

;; PROTOTYPE IMPLEMENTATION

;;---------------------------------------------------------------
;; Shader
;;

;; Shader is an entity that "knows" how to draw a given object.
;; The matrices are already set up to object local coodrinate
;; when the 'render' method is called.

(define-class <shader> (<renderable-node>)
  ()
  :abstract #t)

(define-method node-render ((self <shader>) (object <node>))
  ())

;; GL shader ----------------------------------------------------
;;   A standard OpenGL shader.  requires gl-material.

(define-class <gl-shader> (<shader>)
  ((material   :init-keyword :material :accessor material-of
               :initform (make <gl-material>)
               :attribute-type <gl-material>)
   ))

(define-method node-render ((shader <gl-shader>) (object <node>))
  (and-let* ((mat (material-of shader)))
    (gl-material GL_FRONT GL_AMBIENT   (ambient-of mat))
    (gl-material GL_FRONT GL_DIFFUSE   (diffuse-of mat))
    (gl-material GL_FRONT GL_SPECULAR  (specular-of mat))
    (gl-material GL_FRONT GL_SHININESS (shininess-of mat))
    (gl-material GL_FRONT GL_EMISSION  (emission-of mat)))
  (node-render object 'shade))

;; When an instance is created, a default shader is assigned.
;; A default shader is the first instance of GL-shader in the
;; current part---if no shader exists, one is created.

(define (default-shader)
  (or (find (cut is-a? <> <gl-shader>)
            (reverse (part-nodes (current-part))))
      (make <gl-shader> :name "default-shader")))

;; Const shader ---------------------------------------------------
;;   Render the object with constant color.  It's of little
;;   use for users.
;;   Two const-shaders are created as the transient node
;;   in order to render the wireframe editing screen.

(define-class <const-shader> (<shader>)
  ((color     :init-keyword :color :accessor color-of
              :initform (f32vector 0.5 0.5 0.5 0.0)
              :attribute-type <bn-color4>)
   ))

(define-method node-render ((shader <const-shader>) (object <node>))
  (gl-color (color-of shader))
  (node-render object 'noshade))

;; Default wire shaders
(define *wire-shader-unselected*
  (with-part (transient-part)
    (make <const-shader> :color (f32vector 0.15 0.15 0.35 0.0))))

(define *wire-shader-selected*
  (with-part (transient-part)
    (make <const-shader> :color (f32vector 0.60 0.85 0.60 0.0))))

(define *wire-shader-selected1*
  (with-part (transient-part)
    (make <const-shader> :color (f32vector 0.85 0.85 0.85 0.0))))

(define *wire-shader-selectedn*
  (with-part (transient-part)
    (make <const-shader> :color (f32vector 0.75 0.25 0.25 0.0))))

(provide "banyan/shader")
