;;;
;;; banyan/shape - shapes
;;;
;;;  Copyright(C) 2002-2003 by Shiro Kawai (shiro@acm.org)
;;;
;;;  Permission to use, copy, modify, distribute this software and
;;;  accompanying documentation for any purpose is hereby granted,
;;;  provided that existing copyright notices are retained in all
;;;  copies and that this notice is included verbatim in all
;;;  distributions.
;;;  This software is provided as is, without express or implied
;;;  warranty.  In no circumstances the author(s) shall be liable
;;;  for any damages arising out of the use of this software.
;;;
;;;  $Id: shape.scm,v 1.31 2003/12/17 10:50:09 shirok Exp $
;;;

(select-module banyan)
(use srfi-2)
(use math.const)
(use gauche.uvector)
(use gauche.collection)
(use gauche.mop.propagate)

;; PROTOTYPE IMPLEMENTATION

;;===============================================================
;; Shape abstract class
;;
(define-class <shape> (<renderable-node>)
  ()
  :abstract #t)

(define-method freeze-transform! ((shape <shape>) (lmatrix <matrix4f>))
  #f)

;;===============================================================
;; Polygon Shape
;;
(define-class <polygon-shape> (<propagate-mixin> <shape>)
  ((vertices :allocation :c-node
             :init-keyword :vertices :accessor vertices-of
             :attribute-type <point4f-array>)
   (normals  :allocation :c-node
             :init-keyword :normals :accessor normals-of
             :attribute-type <vector4f-array>)
   (faces    :allocation :c-node
             :init-keyword :faces :accessor faces-of
             :attribute-type #t)
   (num-faces :allocation :c-node
             :init-keyword :num-faces :getter num-faces-of)
   )
  :c-class <%shape>)

(define-method node-render ((self <polygon-shape>) op)
  (when (visible? self)
    (%shape-render (%c-node self) op)))

(define-method face-fold-indices ((self <polygon-shape>) proc knil)
  (%shape-face-fold-indices (%c-node self) proc knil))

(define-method face-fold-count&indices ((self <polygon-shape>) proc knil)
  (%shape-face-fold-count&indices (%c-node self) proc knil))

(define-method face-fold-vertices ((self <polygon-shape>) proc knil)
  (%shape-face-fold-vertices (%c-node self) proc knil))

(define-method face-fold-count&vertices ((self <polygon-shape>) proc knil)
  (%shape-face-fold-count&vertices (%c-node self) proc knil))

(define-method face-fold-center ((self <polygon-shape>) proc knil)
  (%shape-face-fold-center (%c-node self) proc knil))

(define (shape-set-facet-normals! shape)
  (%shape-set-facet-normals! (%c-node shape)))

(define (shape-set-vertex-normals! shape)
  (%shape-set-vertex-normals! (%c-node shape)))

(define (shape-set-render-normals! scale)
  (%shape-set-render-normals! scale))

(define-method freeze-transform! ((shape <polygon-shape>) (lmatrix <matrix4f>))
  (let ((var (vertices-of shape)))
    (dotimes (i (point4f-array-length var))
      (set! (ref var i) (* lmatrix (ref var i)))))
  ;; NB: let the next render routine calculate normals.
  ;; TODO: need to deal with user-computed normals
  (set! (normals-of shape) #f)
  )

;;---------------------------------------------------------
;; Basic shapes
;;
(define (make-cube-shape size)
  (let ((size/2  (/ size 2))
        (-size/2 (/ size -2))
        (root3   (sqrt 3))
        (-root3  (- (sqrt 3)))
        (vtx     (make-point4f-array 8)))
    (set! (ref vtx 0) (point4f  size/2  size/2  size/2))
    (set! (ref vtx 1) (point4f  size/2 -size/2  size/2))
    (set! (ref vtx 2) (point4f  size/2 -size/2 -size/2))
    (set! (ref vtx 3) (point4f  size/2  size/2 -size/2))
    (set! (ref vtx 4) (point4f -size/2  size/2  size/2))
    (set! (ref vtx 5) (point4f -size/2 -size/2  size/2))
    (set! (ref vtx 6) (point4f -size/2 -size/2 -size/2))
    (set! (ref vtx 7) (point4f -size/2  size/2 -size/2))
    (make <polygon-shape>
      :name "cube-shape-1"
      :vertices vtx
      :faces (list
              (list GL_QUADS
                    (u8vector 0 1 2 3
                              0 3 7 4
                              4 5 1 0
                              3 2 6 7
                              7 6 5 4
                              2 1 5 6))))
    ))

(define (make-sphere-shape radius t-divx p-divx)
  (let* ((t-div  (inexact->exact (ceiling (clamp t-divx 4.0))))
         (p-div  (inexact->exact (ceiling (clamp p-divx 1.0))))
         (siz    (+ (* t-div (- p-div 1)) 2))
         (vtx    (make-point4f-array siz))
         (faces  '()))
    (dotimes (t t-div)
      (dotimes (p (- p-div 1))
        (let* ((phi   (* (/ (- (+ p 1) (/ p-div 2)) p-div) pi))
               (theta (* (/ t t-div) 2 pi))
               (x     (* (cos phi) (cos theta)))
               (y     (sin phi))
               (z     (* (cos phi) (sin theta))))
          (set! (ref vtx (+ (* p t-div) t))
                (point4f (* radius x) (* radius y) (* radius z))))))
    (set! (ref vtx (- siz 2)) (point4f 0 (- radius) 0))
    (set! (ref vtx (- siz 1)) (point4f 0 radius 0))
    (dotimes (t t-div)
      (with-builder (<u32vector> add! get)
        (add! (- siz 2))                ;bottom
        (dotimes (p (- p-div 1))
          (add! (+ (* p t-div) t))
          (add! (+ (* p t-div) (modulo (+ t 1) t-div))))
        (add! (- siz 1))                ;top
        (push! faces (list GL_TRIANGLE_STRIP (get)))))
    (make <polygon-shape> :name "sphere-shape-1" :vertices vtx :faces faces)))

(define (make-cylinder-shape radius height t-divx y-divx)
  (let* ((t-div  (inexact->exact (ceiling (clamp t-divx 3.0))))
         (y-div  (inexact->exact (ceiling (clamp y-divx 1.0))))
         (siz    (+ (* t-div (+ y-div 2)) 2))
         (vtx    (make-point4f-array siz))
         (faces  '()))
    (dotimes (t t-div)
      (dotimes (y (+ y-div 2))
        (let* ((theta (* (/ t t-div) 2 pi))
               (x     (cos theta))
               (z     (sin theta)))
          (set! (ref vtx (+ (* y t-div) t))
                (point4f (* radius x)
                         (* height (- (/ y (+ y-div 1)) 0.5))
                         (* radius z)))
          )))
    (set! (ref vtx (- siz 2)) (point4f 0 (- (/ height 2)) 0))
    (set! (ref vtx (- siz 1)) (point4f 0 (/ height 2) 0))
    ;; floor
    (with-builder (<u32vector> add! get)
      (add! (- siz 2))                ;floor center
      (dotimes (t t-div) (add! t))
      (add! 0)
      (push! faces (list GL_TRIANGLE_FAN (get))))
    ;; lid
    (with-builder (<u32vector> add! get)
      (add! (- siz 1))                ;lid center
      (dotimes (t t-div) (add! (+ (* (+ y-div 1) t-div) (- t-div t 1))))
      (add! (+ (* (+ y-div 1) t-div) (- t-div 1)))
      (push! faces (list GL_TRIANGLE_FAN (get))))
    ;; walls
    (dotimes (t t-div)
      (with-builder (<u32vector> add! get)
        (dotimes (y (+ y-div 2))
          (add! (+ (* y t-div) t))
          (add! (+ (* y t-div) (modulo (- t 1) t-div))))
        (push! faces (list GL_TRIANGLE_STRIP (get)))))
    (make <polygon-shape> :name "cylinder-shape-1" :vertices vtx :faces faces))
  )

(define (make-cone-shape radius height t-divx y-divx)
  (let* ((t-div (inexact->exact (ceiling (clamp t-divx 4.0))))
         (y-div (inexact->exact (ceiling (clamp y-divx 0.0))))
         (siz   (+ (* t-div (+ y-div 2)) 2))
         (vtx   (make-point4f-array siz))
         (faces '()))
    (dotimes (t t-div)
      (dotimes (y (+ y-div 2))
        (let* ((theta (* (/ t t-div) 2 pi))
               (coeff (- 1 (/ y (+ y-div 1))))
               (x     (* coeff radius (cos theta)))
               (z     (* coeff radius (sin theta))))
          (set! (ref vtx (+ (* y t-div) t))
                (point4f x (* height (- (/ y (+ y-div 1)) 0.5)) z))
          )))
    (set! (ref vtx (- siz 2)) (point4f 0 (- (/ height 2)) 0))
    (set! (ref vtx (- siz 1)) (point4f 0 (/ height 2) 0))
    ;; floor
    (with-builder (<u32vector> add! get)
      (add! (- siz 2))                ;floor center
      (dotimes (t t-div) (add! t))
      (add! 0)
      (push! faces (list GL_TRIANGLE_FAN (get))))
    ;; walls
    (dotimes (t t-div)
      (with-builder (<u32vector> add! get)
        (dotimes (y (+ y-div 2))
          (add! (+ (* y t-div) t))
          (add! (+ (* y t-div) (modulo (- t 1) t-div))))
        (push! faces (list GL_TRIANGLE_STRIP (get)))))
    (make <polygon-shape> :name "cone-shape-1" :vertices vtx :faces faces))
  )

(define (make-plane-shape x-size z-size x-divx z-divx)
  (let* ((x-div  (inexact->exact (ceiling (clamp x-divx 0.0))))
         (z-div  (inexact->exact (ceiling (clamp z-divx 0.0))))
         (siz    (* (+ x-div 2) (+ z-div 2)))
         (vtx    (make-point4f-array siz))
         (faces  '()))
    (dotimes (z (+ z-div 2))
      (dotimes (x (+ x-div 2))
        (set! (ref vtx (+ (* z (+ x-div 2)) x))
              (point4f (* x-size (- (/ x (+ x-div 1)) 0.5))
                       0
                       (* z-size (- (/ z (+ z-div 1)) 0.5))))))
    (dotimes (x (+ x-div 1))
      (with-builder (<u32vector> add! get)
        (dotimes (z (+ z-div 2))
          (add! (+ (* z (+ x-div 2)) (+ x 1)))
          (add! (+ (* z (+ x-div 2)) x)))
        (push! faces (list GL_QUAD_STRIP (get)))))
    (make <polygon-shape> :name "plane-shape-1" :vertices vtx :faces faces)
    ))

;; For testing
;(define-class <teapot-shape> (<shape>)
;  ((size  :init-keyword :size :accessor size-of :init-value 1.0)))

;(define-method node-render ((self <teapot-shape>) op)
;  (glut-solid-teapot (size-of self)))

;;===============================================================
;; Locator Shape
;;

;; A locator is a placeholder for intermediate transform.
;; It is displayed only on the non-shading mode.

(define-class <locator-shape> (<shape>)
  ())

(define-method node-render ((self <locator-shape>) op)
  (when (and (visible? self) (eq? op 'noshade))
    (gl-vertex-pointer 3 '#f32(1.0 0.0 0.0 -1.0 0.0 0.0
                               0.0 1.0 0.0 0.0 -1.0 0.0
                               0.0 0.0 1.0 0.0 0.0 -1.0))
    (gl-draw-elements GL_LINES '#u8(0 1 2 3 4 5))))

;;===============================================================
;; Connector Shape
;;

;; exclusively used by connector instance

(define-class <connector-shape> (<shape>)
  ())

(define-method node-render ((self <connector-shape>) op)
  (when (and (visible? self) (eq? op 'noshade))
    (gl-vertex-pointer 3 '#f32(1.0 0.0 0.0 -1.0 0.0 0.0
                               0.0 1.0 0.0 0.0 -1.0 0.0
                               0.0 0.0 1.0 0.0 0.0 -1.0
                               0.5 0.0 0.0 0.0 0.5 0.0
                               0.0 0.0 0.5 0.0 0.5 0.0
                               -0.5 0.0 0.0 0.0 0.5 0.0
                               0.0 0.0 -0.5 0.0 0.5 0.0
                               0.5 0.0 0.0 0.0 -0.5 0.0
                               0.0 0.0 0.5 0.0 -0.5 0.0
                               -0.5 0.0 0.0 0.0 -0.5 0.0
                               0.0 0.0 -0.5 0.0 -0.5 0.0
                               0.5 0.0 0.0 0.0 0.0 0.5
                               0.5 0.0 0.0 0.0 0.0 -0.5
                               -0.5 0.0 0.0 0.0 0.0 0.5
                               -0.5 0.0 0.0 0.0 0.0 -0.5))
    (gl-draw-elements GL_LINES '#u8(0 1 2 3 4 5
                                    6 7 8 9 10 11 12 13
                                    14 15 16 17 18 19 20 21
                                    22 23 24 25 26 27 28 29))
    ))

;;===============================================================
;; Procedural Shape
;;

;; A procedural shape is an updatable nodes that takes another shape
;; and calculates shape from it.

;; Parametric shape

(define-class <parametric-shape> (<polygon-shape> <updatable-node>)
  ((s-wrap   :init-keyword :s-wrap :init-value #t
             :attribute-type <boolean>)
   (t-wrap   :init-keyword :t-wrap :init-value #t
             :attribute-type <boolean>)
   (generate-vertex :init-keyword :generate-vertex
                    :init-value #f)
   )
  :c-class <%shape>)

;;-----------------------------------------------------------
;; Experimental fractal terrain generator
;;

(use srfi-27)

(define-class <fractal-terrain-shape> (<polygon-shape>)
  ((size       :init-keyword :size  :attribute-type <real>
               :init-value 5.0)
   (seed       :init-keyword :seed  :attribute-type <integer>
               :init-value 0)
   (roughness  :init-keyword :roughness :attribute-type <real>
               :init-value 2.0)
   (perturbation :init-keyword :perturbation
                 :attribute-type <real>
                 :init-value 1.0)
   (iterations :init-keyword :iterations :attribute-type <integer>
               :init-value 4)
   )
  :c-class <%shape>)

(define-method generate-geometry ((shape <fractal-terrain-shape>))
  (let* ((np   (+ 1 (expt 2 (ref shape 'iterations))))
         (pts  (make-point4f-array (* np np)))
         (rsrc (make-random-source))
         (rand (random-source-make-reals rsrc)))

    ;; grid index
    (define (gi x y) (+ (* y np) x))
    ;; index -> coordinate
    (define (xcoord i)
      (- (* i (/ (* (ref shape 'size) 2) (- np 1)))
         (ref shape 'size)))
    (define (zcoord i)
      (- (ref shape 'size)
         (* i (/ (* (ref shape 'size) 2) (- np 1)))))

    ;; init corner points
    (define (init-corners)
      (dotimes (x 2)
        (dotimes (y 2)
          (let ((xx (* x (- np 1)))
                (yy (* y (- np 1))))
            (set! (ref pts (gi xx yy))
                  (point4f (xcoord xx) 0 (zcoord yy)))))))

    ;; set midpoint of (x0,y0) and (x1,y1), with perturbation factor p
    (define (set-mid x0 y0 x1 y1 p)
      (let ((h0 (ref (ref pts (gi x0 y0)) 1))
            (h1 (ref (ref pts (gi x1 y1)) 1))
            (x2 (+ x0 (quotient (- x1 x0) 2)))
            (y2 (+ y0 (quotient (- y1 y0) 2))))
        (set! (ref pts (gi x2 y2))
              (point4f (xcoord x2)
                       (+ (/ (+ h0 h1) 2) (* p (- (rand) 0.5)))
                       (zcoord y2)))))

    ;; calculate midpoints.  step = expt(2, iterations - num-midpts)
    (define (mid num-midpts step p)
      (dotimes (x num-midpts)
        (dotimes (y num-midpts)
          (let* ((x0 (* x step))
                 (y0 (* y step))
                 (x1 (+ x0 step))
                 (y1 (+ y0 step)))
            (when (zero? x) (set-mid x0 y0 x0 y1 p))
            (when (zero? y) (set-mid x0 y0 x1 y0 p))
            (set-mid x0 y0 x1 y1 p)
            (set-mid x1 y0 x1 y1 p)
            (set-mid x0 y1 x1 y1 p)))))

    ;; body 
    (random-source-pseudo-randomize! rsrc (ref shape 'seed) 0)
    (init-corners)
    (let loop ((num-midpts 1)
               (step (- np 1))
               (p    (ref shape 'perturbation)))
      (when (> step 1)
        (mid num-midpts step p)
        (loop (* num-midpts 2) (/ step 2) (/ p (ref shape 'roughness)))))
    (set! (ref shape 'vertices) pts)
    (set! (ref shape 'faces)
          (map (lambda (s)
                 (list GL_TRIANGLE_STRIP
                       (list->u32vector
                        (apply append
                               (map (lambda (t) (list t (+ t 1)))
                                    (iota np s np))))))
               (iota (- np 1))))
    ))

(define (make-fractal-terrain-shape size seed roughness perturb iter)
  (let* ((shape (make <fractal-terrain-shape>
                  :size size :seed (inexact->exact (round seed))
                  :roughness roughness
                  :perturbation perturb
                  :iterations (inexact->exact (round iter)))))
    (generate-geometry shape)
    shape))

;;-----------------------------------------------------------
;; Experimental L-system implementation
;;

;; L-system is realized by a group of interconnected nodes.
;; <lsystem-generator> is a node that keeps the production rules
;; and generates edges from them.
;; A <lsystem-shape> node should be connected to the <lsystem-generator>
;; to realize one type of edge as a shape.
;;
;; Actually, the 'edges' generated by <lsystem-generator> is a set of
;; matrices; <lsystem-shape> can apply the matrices to the unit shape to
;; obtain the actual shape.

;; Production rules:
;;  <rules> : (<rule> ...)
;;  <rule>  : (<edge> => <term> ...)
;;          | (<edge> (<probability> => <term> ...) ...)
;;  <probability> : real number within (0.0, 1.0], or 'else
;;  <term>  : (<term> ...)
;;          | <edge>
;;          | <rotation>
;;          | (* <repeat> <term> ...)
;;  <edge>  : symbol
;;  <rotation> : x+ x- y+ y- z+ z-
;;  <repeat> : integer

(define-class <lsystem-generator> (<updatable-node>)
  ((rules    :init-keyword :rules :init-value '() ;production rules
             :attribute-type #t) ; <sexp>
   (initial-edge  :init-keyword :initial-edge :init-value #f
                  :attribute-type #t) ; <symbol>
   (depth    :init-keyword :depth :init-value 5
             :attribute-type <integer>)
   (alpha    :init-keyword :alpha :init-value 25.7
             :attribute-type <real>) ; angle to diverge from axis
   (delta    :init-keyword :delta :init-value 25.7
             :attribute-type <real>) ; angle to rotage around axis
   (edges    :init-keyword :shapes :init-value '())
   ))

(define (lsystem-production-expand-1 lis rules)

  (define (expand-a-rule rule)
    (cond
     ((not (pair? (cdr rule))) (error "bad rule" rule))
     ((eq? (cadr rule) '=>) (cddr rule))
     ((not (every pair? (cdr rule))) (error "bad rule" rule))
     (else ;; probabilistic state transition - not yet supported
      (error "stocastic L-system is yet to be written"))))

  (define (expand-a-term term)
    (cond ((memq term '(x+ x- y+ y- z+ z-)) (list term))
          ((symbol? term)
           (cond ((assq term rules) => expand-a-rule)
                 (else (error "encountered a term without rule" term))))
          ((not (list? term))
           (error "bad term" term))
          (else
           (list (append-map expand-a-term term)))))

  (append-map expand-a-term lis))

(define (lsystem-production-expand init rules depth)
  (let loop ((lis (list init))
             (n   0))
    (if (= n depth)
        lis
        (loop (lsystem-production-expand-1 lis rules) (+ n 1)))))

(define (lsystem-generate-edges lis alpha delta)
  (define edges (make-hash-table 'eq?))
  (define unit  #,(vector4f 0 1 0))
  (define rotations
    `((x+ ,(rotation->matrix4f #,(vector4f 1 0 0)  alpha))
      (x- ,(rotation->matrix4f #,(vector4f -1 0 0) alpha))
      (y+ ,(rotation->matrix4f #,(vector4f 0 1 0)  delta))
      (y- ,(rotation->matrix4f #,(vector4f 0 -1 0) delta))
      (z+ ,(rotation->matrix4f #,(vector4f 0 0 1)  alpha))
      (z- ,(rotation->matrix4f #,(vector4f 0 0 -1) alpha))))
  (define (turtle pos rot commands)
    (when (pair? commands)
      (let1 c (car commands)
        (cond ((pair? c)
               (turtle pos rot c)
               (turtle pos rot (cdr commands)))
              ((assq c rotations)
               => (lambda (p) (turtle pos (* rot (cadr p)) (cdr commands))))
              (else
               (let ((next-pos (+ pos (* rot unit)))
                     (matrix (* (translation->matrix4f pos) rot)))
                 (hash-table-push! edges c matrix)
                 (turtle next-pos rot (cdr commands))))))))

  (turtle #,(point4f 0 0 0) (make-matrix4f) lis)
  (hash-table-map edges (lambda (key val) (cons key (reverse! val)))))

(define-method node-update ((self <lsystem-generator>))
  (and-let* (((null? (ref self 'edges)))
             (initial (ref self 'initial-edge))
             (rules   (ref self 'rules))
             (commands (lsystem-production-expand initial rules
                                                  (ref self 'depth))))
    (slot-set! self 'edges
               (lsystem-generate-edges commands
                                       (* (ref self 'alpha) pi/180)
                                       (* (ref self 'delta) pi/180)))
    ))

(define-class <lsystem-shape> (<polygon-shape> <updatable-node>)
  ((unit-shape :init-keyword :unit-shape :init-value #f
               :attribute-type <polygon-shape>)
   (edge-name  :init-keyword :edge-name :init-value #f
               :attribute-type <symbol>)
   (generator  :init-keyword :generator :init-value #f
               :attribute-type <lsystem-generator>)
   )
  :c-class <%shape>)

(define-method initialize ((self <lsystem-shape>) initargs)
  (next-method self (list* :vertices #f initargs)))

(define-method node-depends ((self <lsystem-shape>))
  (cond-list ((ref self 'unit-shape))
             ((ref self 'generator))))

(define-method node-update ((self <lsystem-shape>))
  (and-let* (((not (ref self 'vertices)))
             (unit (ref self 'unit-shape))
             (gen  (ref self 'generator))
             (edge-matrices (assq (ref self 'edge-name) (ref gen 'edges)))
             (mtxs (cdr edge-matrices))
             (unit-verts (ref unit 'vertices))
             (unit-faces (ref unit 'faces))
             (unit-num-verts (point4f-array-length unit-verts)))
    (with-builder (<point4f-array> add-vtx! get-vtx
                                   :size (* (length mtxs) unit-num-verts))
      (with-builder (<list> add-face! get-face)
        (let loop ((mtxs mtxs)
                   (offset 0))
          (if (null? mtxs)
              (begin (slot-set! self 'vertices (get-vtx))
                     (slot-set! self 'faces (get-face)))
              (let ((mat (car mtxs)))
                (for-each (lambda (vertex)
                            (add-vtx! (* mat vertex)))
                          unit-verts)
                (for-each (lambda (faces)
                            ;; kludge
                            (let1 facets
                                (if (u32vector? (cadr faces))
                                    (cadr faces)
                                    (coerce-to <u32vector> (cadr faces)))
                              (add-face! (list (car faces)
                                               (u32vector-add facets offset)))
                              ))
                          unit-faces)
                (loop (cdr mtxs) (+ offset unit-num-verts)))))
        )
      )
    ))

;; Convenience fn to create lsystem shape nodes.
;; I don't have enough UI component yet, so many parameters are "pre-set"

(define (%shape-translate shape vec)
  (let* ((vtx (ref shape 'vertices))
         (len (point4f-array-length vtx)))
    (dotimes (i len)
      (let1 pt (point4f-array-ref/shared vtx i)
        (point4f-add! pt vec)))
    )
  shape)

(define (%leaf-shape)
  (let ((vertices (make-point4f-array 6)))
    (set! (ref vertices 0) #,(point4f 0 0 0))
    (set! (ref vertices 1) #,(point4f -0.2 0.4 0))
    (set! (ref vertices 2) #,(point4f 0.2 0.4 0))
    (set! (ref vertices 3) #,(point4f -0.2 0.6 0))
    (set! (ref vertices 4) #,(point4f 0.2 0.6 0))
    (set! (ref vertices 5) #,(point4f 0 1.0 0))
    (make <polygon-shape>
      :name "leaf-shape-1"
      :vertices vertices
      :faces (list (list GL_POLYGON (u32vector 0 1 3 5 4 2))
                   (list GL_POLYGON (u32vector 2 4 5 3 1 0))))
    ))

(define *lsystem-shapes*
  `((C ,(lambda ()
          (%shape-translate (make-cylinder-shape 0.07 1.0 3 1)
                            #,(vector4f 0 0.5 0))))
    (F ,(lambda ()
          (%shape-translate (make-cylinder-shape 0.07 1.0 3 1)
                            #,(vector4f 0 0.5 0))))
    (S ,(lambda () (%leaf-shape)))
    (L ,(lambda () (%leaf-shape)))
    ))

(define *lsystem-inits&rules*
  ;; ((tag init rule ...) ...)
  '((default
     A
     (A => (x+ F (x+ S) A) y+ y+ y+ y+ y+
           (x+ F (x+ S) A) y+ y+ y+ y+ y+ y+ y+
           (x+ F (x+ S) A))
     (F => F (x+ S) y+ y+ y+ y+ y+ y+ y+ (x+ S))
     (S => (x+ x+ L))
     (L => L)
     )
    (test  ;; alpha must be small
     A
     (A => C B)
     (C => C x+ C)
     (B => (x+ x+ x+ x+ x+ x+ x+ x+ x+ x+ x+
               (x+ x+ x+ z+ z+ z+ z+ z+ z+ z+ z+ S)
               (x+ x+ x+ z- z- z- z- z- z- z- z- S) F) y+ y+ y+ y+
           (x+ x+ x+ x+ x+ x+ x+ x+ x+ x+ x+
               (x+ x+ x+ z+ z+ z+ z+ z+ z+ z+ z+ S)
               (x+ x+ x+ z- z- z- z- z- z- z- z- S) F) y+
           (x+ x+ x+ x+ x+ x+ x+ x+ x+ x+ x+
               (x+ x+ x+ z+ z+ z+ z+ z+ z+ z+ z+ S)
               (x+ x+ x+ z- z- z- z- z- z- z- z- S) F) y+ y+ y+
           (x+ x+ x+ x+ x+ x+ x+ x+ x+ x+ x+
               (x+ x+ x+ z+ z+ z+ z+ z+ z+ z+ z+ S)
               (x+ x+ x+ z- z- z- z- z- z- z- z- S) F) y+ y+
           (x+ x+ x+ x+ x+ x+ x+ x+ x+ x+ x+
               (x+ x+ x+ z+ z+ z+ z+ z+ z+ z+ z+ S)
               (x+ x+ x+ z- z- z- z- z- z- z- z- S) F))
     (F => x+ F
        (x+ x+ x+ x+ x+ z+ z+ z+ z+ z+ z+ z+ z+ S)
        (x+ x+ x+ x+ x+ z- z- z- z- z- z- z- z- S)
        x+ F
        (x+ x+ x+ x+ x+ z+ z+ z+ z+ z+ z+ z+ z+ S)
        (x+ x+ x+ x+ x+ z- z- z- z- z- z- z- z- S))
     (S => S)
     )
    ))

(define (make-lsystem-shapes tag depth alpha delta)
  (or (and-let* ((inits&rules (assq tag *lsystem-inits&rules*))
                 (generator (make <lsystem-generator>
                              :rules (cddr inits&rules) 
                              :initial-edge (cadr inits&rules)
                              :depth depth :alpha alpha :delta delta)))
        (node-update generator)
        (let* ((edges  (map car (ref generator 'edges)))
               (shapes (filter-map
                        (lambda (edge)
                          (and-let* ((p (assq edge *lsystem-shapes*)))
                            (make <lsystem-shape>
                              :unit-shape ((cadr p)) :edge-name edge
                              :generator generator)))
                        edges)))
          (for-each node-update shapes)
          shapes))
      (error "undefined L-System tag" tag)))

(provide "banyan/shape")
