;;;
;;; banyan/transform - transforms
;;;
;;;  Copyright(C) 2002-2003 by Shiro Kawai (shiro@acm.org)
;;;
;;;  Permission to use, copy, modify, distribute this software and
;;;  accompanying documentation for any purpose is hereby granted,
;;;  provided that existing copyright notices are retained in all
;;;  copies and that this notice is included verbatim in all
;;;  distributions.
;;;  This software is provided as is, without express or implied
;;;  warranty.  In no circumstances the author(s) shall be liable
;;;  for any damages arising out of the use of this software.
;;;
;;;  $Id: transform.scm,v 1.14 2003/02/23 02:41:58 shirok Exp $
;;;

(select-module banyan)
(use gauche.mop.validator)
(use gauche.mop.propagate)

;; PROTOTYPE IMPLEMENTATION

;; Vector presentation

(define-class <bn-vec3-meta> (<class>)
  ())

(define-class <bn-vec3> ()
  ()
  :metaclass <bn-vec3-meta>)

;;---------------------------------------------------------------------
;; Transform class
;;

;; trick : we can't specify :attribute-type <transform> in the `parent'
;; slot, since the value of slot option is evaluated before <transform>
;; is bound.
;; For now, we define a special metaclass and fixes the slot option
;; when the class is created.

(define-class <transform-meta> (<node-meta>)
  ())

(define-method initialize ((class <transform-meta>) initargs)
  (next-method)
  (let ((slot (class-slot-definition class 'parent)))
    ;; DIRTY: depends on the internal rep. of slot definition
    (set-cdr! slot (list* :attribute-type class (cdr slot))))
  )

;; Base transform class

(define-class <transform> (<updatable-node>)
  ((gmatrix     :allocation :c-node     :accessor gmatrix-of)
   (lmatrix     :allocation :c-node     :accessor lmatrix-of)
   (cmatrix     :allocation :c-node     :accessor cmatrix-of)
   (imatrix     :allocation :c-node     :accessor imatrix-of)
   (parent      :accessor parent-of :init-value #f
                :validator (lambda (o v)
                             (let ((old (parent-of o)))
                               (unless (eq? old v)
                                 (transform-update-parent o old v)
                                 (slot-set! (%c-node o) 'parent
                                            (and (is-a? v <transform>)
                                                 (%c-node v)))
                                 (notify-graph-changed (current-scene))))
                             v))
   (children    :accessor children-of :init-value '())
   )
  :abstract #t
  :c-class <%transform>
  :metaclass <transform-meta>)

;; And this is another trick... When 'parent' slot is initialized
;; at the creation of instance, transform-update-parent is called,
;; but the old value of the slot is unbound.  So I patch it.
(define-method slot-unbound ((class <transform-meta>) object slot)
  (if (eq? slot 'parent)
      #f
      (next-method)))

(define-method transform-update-parent ((xf <transform>) old new)
  (when old
    (update! (children-of old) (cut delete xf <> eq?)))
  (if new
      (begin
        (push! (children-of new) xf)
        (and-let* ((inv (matrix4f-inverse (gmatrix-of new) #f)))
          (set! (cmatrix-of xf) inv)))
      (begin
        (set! (cmatrix-of xf) #f))
      ))

(define-method node-depends ((self <transform>))
  (let1 p (parent-of self)
    (if p (list p) '())))

(define-method node-update ((self <transform>))
  (%transform-update (%c-node self)))

(define-method node-reset! ((self <transform>))
  #f)

;; Some transform protocols.
;(define-method transform-translation-matrix ((xf <transform>))
;  (let ((tv (matrix4f-column/shared (lmatrix-of xf)

;;---------------------------------------------------------------------
;; TRS transform
;;

(define-class <trs-transform> (<transform>)
  ((translation :init-keyword :translation :accessor translation-of
                :attribute-type <bn-vec3>
                :initform (vector4f 0.0 0.0 0.0))
   (rotation    :init-keyword :rotation :accessor rotation-of
                :attribute-type #t
                :initform (make-quatf))
   (scale       :init-keyword :scale :accessor scale-of
                :attribute-type <bn-vec3>
                :initform (vector4f 1.0 1.0 1.0)))
  :c-class <%transform>)

(define-method node-update ((self <trs-transform>))
  (let ((lm (lmatrix-of self))
        (gm (gmatrix-of self)))
    (tqs->matrix4f! lm
                    (translation-of self)
                    (rotation-of self)
                    (scale-of self))
    (next-method)
    ))

(define-method node-reset! ((self <trs-transform>))
  (set! (translation-of self) (vector4f 0.0 0.0 0.0))
  (set! (rotation-of self) (make-quatf))
  (set! (scale-of self) (vector4f 1.0 1.0 1.0))
  (set! (lmatrix-of self) (make-matrix4f)))

(define-method set-matrix-decomposing ((self <trs-transform>) mat)
  (receive (f t r h s) (matrix4f-decompose mat)
    (set! (translation-of self) t)
    (set! (scale-of self) s)
    (if r
        (receive (axis angle) (matrix4f->rotation r)
          (set! (rotation-of self) (make-quatf axis angle)))
        (set! (rotation-of self) (make-quatf)))))

(provide "banyan/transform")
