;;;
;;; glim/appframe.scm - application frame
;;;
;;;  Copyright(C) 2002-2003 by Shiro Kawai (shiro@acm.org)
;;;
;;;  Permission to use, copy, modify, distribute this software and
;;;  accompanying documentation for any purpose is hereby granted,
;;;  provided that existing copyright notices are retained in all
;;;  copies and that this notice is included verbatim in all
;;;  distributions.
;;;  This software is provided as is, without express or implied
;;;  warranty.  In no circumstances the author(s) shall be liable
;;;  for any damages arising out of the use of this software.
;;;
;;;  $Id: appframe.scm,v 1.11 2003/01/15 12:32:59 shirok Exp $
;;;

(define-module glim.appframe
  (use srfi-1)
  (use gtk)
  (use gauche.parameter)
  (use gauche.mop.singleton)
  (use gauche.mop.validator)
  (use glim.pane)
  (export <glim-application-frame>
          glim-application-frame?
          glim-make-application-frame
          glim-run-frame-top-level
          glim-default-frame-top-level
          glim-frame-exit
          instance-of

          %glim-register-pane-creator
          glim-get-frame-pane
          define-glim-application-frame
          %glim-expand-panes
          %glim-expand-layouts
          )
  )
(select-module glim.appframe)

;;=============================================================
;; Application frame base class
;;

(define-class <glim-application-frame-meta> (<class>) ())
(define-class <glim-application-frame-base> () ()
  :metaclass <glim-application-frame-meta>)

(define-class <glim-application-frame> (<glim-application-frame-base>)
  ((%widget)
   (%name        :init-keyword :name :init-form (name-anon-appframe))
   (%pretty-name :init-keyword :pretty-name :init-value #f)
   (%pane-creators :allocation :each-subclass :init-value '())
   (%layouters   :allocation :each-subclass :init-value '())
   (%panes       :init-value '())
   (%layout      :init-value '())
   (%top-level   :init-form glim-default-frame-top-level)
   (%default-instance :allocation :each-subclass :init-value #f)
   (%active-frames :allocation :class :init-value '())
   ))

(define name-anon-appframe
  (let ((c 0))
    (lambda () (inc! c) #`"application-frame-,c")))

(define-method initialize ((self <glim-application-frame>) initargs)
  (let1 w (gtk-window-new GTK_WINDOW_TOPLEVEL)
    (slot-set! self '%widget w)
    (next-method)
    (unless (ref self '%pretty-name)
      (set! (ref self '%pretty-name) (x->string (ref self '%name))))
    (gtk-window-set-title w (ref self '%pretty-name))
    (g-signal-connect w "delete_event"
                      (lambda _ (glim-frame-delete-callback self) #f))
    (%glim-realize-panes self)
    (glim-layout-frame self)
    ))

(define (glim-application-frame? obj) (is-a? obj <glim-application-frame>))

(define (glim-make-application-frame name . opts)
  (let ((class (get-keyword :frame-class opts <glim-application-frame>)))
    (apply make class :name name opts)))

;; In most cases, application frames can be a singleton.
;; However, just inheriting <singleton-mixin> is not enough, for
;; we need to handle delete callback.  To save work, application frame
;; itself supports singleton-like interface---you can create a 'default
;; instance' by instance-of method, which is handled specially.
(define-method instance-of ((class <glim-application-frame-meta>) . args)
  (or (class-slot-ref class '%default-instance)
      (let ((frame (apply make class args)))
        (class-slot-set! class '%default-instance frame)
        frame)))

;; workaround of gf vs module interaction bug
;;  Cf. http://www.shiro.dreamhost.com/scheme/wiliki/wiliki.cgi?Gauche%3aGenericFunction%a4%c8Module&l=jp
(define instance-of (with-module gauche.mop.singleton instance-of))

(define-method glim-run-frame-top-level ((frame <glim-application-frame>) . opts)
  (apply (ref frame '%top-level) frame opts))

(define-method glim-default-frame-top-level ((frame <glim-application-frame>) . opts)
  (let1 aframes (ref frame '%active-frames)
    (unless (memq frame aframes)
      (gtk-widget-show-all (ref frame '%widget))
      (set! (ref frame '%active-frames) (cons frame aframes))
      (when (null? aframes) (gtk-main)))
    ))

(define-method glim-frame-exit ((frame <glim-application-frame>))
  (let1 aframes (delete frame (ref frame '%active-frames) eq?)
    (gtk-widget-hide-all (ref frame '%widget))
    (set! (ref frame '%active-frames) aframes)
    (when (null? aframes) (gtk-main-quit))
    ))

(define-method glim-frame-delete-callback ((frame <glim-application-frame>))
  (when (eq? frame (ref frame '%default-instance))
    (slot-set! frame '%default-instance #f))
  (glim-frame-exit frame))

;;=============================================================
;; Panes
;;

;; bookkeeping
(define (%glim-expand-panes frame-class descriptions)
  `(set! (class-slot-ref ,frame-class '%pane-creators)
         (list
          ,@(map (lambda (desc)
                   `(cons ',(car desc) ;name of pane
                          (lambda (.frame.) ,(cadr desc))))
                 descriptions))))

(define-method glim-get-frame-pane ((frame <glim-application-frame>) name)
  (cond ((assq name (slot-ref frame '%panes)) => cdr)
        (else (errorf "no such pane ~s in frame ~s" name frame))))

(define-method %glim-realize-panes ((frame <glim-application-frame>))
  (set! (slot-ref frame '%panes)
        (map (lambda (p) (cons (car p) ((cdr p) frame)))
             (slot-ref frame '%pane-creators))))


;;=============================================================
;; Layout
;;

(define (%glim-expand-layouts frame-class descriptions)
  (let ((child  (gensym)))
    `(set! (class-slot-ref ,frame-class '%layouters)
           (list
            ,@(map (lambda (desc)
                     `(cons ',(car desc)
                            (lambda (.frame.)
                              (let ((,child ,(cadr desc)))
                                (gtk-container-add (ref .frame. '%widget)
                                                   (widget-of ,child))))))
                   descriptions)))))

;; glim-layout-frame
;;   unlike CLIM, this method is only called when the application frame
;;   is realized.  Resizing objects are taken care of by the gui widgets.

(define-method glim-layout-frame ((frame <glim-application-frame>))
  ;; TODO: select layout.  currently just the first layout is used.
  (let1 layouters (ref frame '%layouters)
    (when (null? layouters)
      (errorf "don't know how to layout ~s" frame))
    ((cdar layouters) frame)))

;;=============================================================
;; define-glim-application-frame
;;

(define-macro (define-glim-application-frame name supers slots . options)
  `(begin
     (define-class ,name (,@supers <glim-application-frame>)
       ,slots
       ,@(cond ((assv :metaclass options) => values)
               (else '())))
     ,(cond ((assv :panes options)
             => (lambda (p) (%glim-expand-panes name (cdr p))))
            (else '()))
     ,(cond ((assv :layouts options)
             => (lambda (p) (%glim-expand-layouts name (cdr p))))
            (else '()))
     ))

(provide "glim/appframe")
