;;;
;;; glim/menu.scm - menu constructor
;;;
;;;  Copyright(C) 2002-2003 by Shiro Kawai (shiro@acm.org)
;;;
;;;  Permission to use, copy, modify, distribute this software and
;;;  accompanying documentation for any purpose is hereby granted,
;;;  provided that existing copyright notices are retained in all
;;;  copies and that this notice is included verbatim in all
;;;  distributions.
;;;  This software is provided as is, without express or implied
;;;  warranty.  In no circumstances the author(s) shall be liable
;;;  for any damages arising out of the use of this software.
;;;
;;;  $Id: menu.scm,v 1.13 2003/01/18 09:59:51 shirok Exp $
;;;

(define-module glim.menu
  (use srfi-1)
  (use srfi-2)
  (use gtk)
  (use glim.pane)
  (use glim.appframe)
  (export <glim-menu> glim-menu
          <glim-menu-item> glim-menu-item
          glim-menu-find-item glim-menu-add-item glim-menu-delete-item
          <glim-popup-menu> glim-popup-menu glim-popup-menu-popup
          <glim-pulldown-menu> glim-pulldown-menu)
  )
(select-module glim.menu)

;; Glim menu is one-level abstraction from widgets and gadgets,
;; but not as abstract as CLIM's command table.
;; The menu object, <glim-menu> keeps the definition of menu
;; hierarchy.  It can be realized as "popup menu" or "pulldown menu"

;;============================================================
;; <glim-menu> - Abstract menu hierarchy
;;

;; Static definition of menu
;;  (glim-menu <menu-items>) => glim-menu
;;  <menu-item> : (menu   [name] [label] . options)
;;              | (button [name] [label] . options)
;;              | (check  [name] [label] . options)
;;              | (separator [name])
;;              | (radio  [name] [label] . options)
;;
;;  options is a keyword-value list.   Menu and radio item may have
;;  :items option that followed by <menu-items>.

(define-class <glim-menu> ()
  ((items :init-value '() :init-keyword :items)
   (instances :init-value '())
   ))

(define-method add-instance ((self <glim-menu>) instance)
  (slot-push! self 'instances instance))

(define-method delete-instance ((self <glim-menu>) instance)
  (update! (ref self 'instances) (cut delete instance <> eq?)))

;; Abstract menu item.
;; The instances field keeps a list of (<root-gadget> . <gadget>),
;; where <root-gadget> is either <glim-popup-menu> or <glim-pulldown-menu>,
;; and <gadget> is the widget that realizes this item under <root-gadget>.
;;
;; When a new item is added to this item, a message is sent to <root-gadget>,
;; so that it can create a new widgets to the added item.
(define-class <glim-menu-item> ()
  ((type      :init-value :button :init-keyword :type)
   (name      :init-value #f      :init-keyword :name) ;symbol
   (label     :init-value ""      :init-keyword :label)
   (options   :init-value '()     :init-keyword :options)
   (items     :init-value '()     :init-keyword :items)
   (callback  :init-value #f      :init-keyword :callback)
   (value     :init-value #f      :init-keyword :value)
   (parent    :init-value #f      :init-keyword :parent)
   (instances :init-value '()))
  )

(define-method add-instance ((self <glim-menu-item>) root instance)
  (slot-push! self 'instances (cons root instance)))

(define-method delete-instance ((self <glim-menu-item>) instance)
  (update! (ref self 'instances)
           (cut remove (lambda (p) (eq? instance (cdr p))) <>)))

(define-method glim-menu-find-item ((menu <glim-menu>) name)
  (define (rec items)
    (cond ((null? items) #f)
          ((eq? (ref (car items) 'name) name) (car items))
          (else (or (rec (ref (car items) 'items))
                    (rec (cdr items))))))
  (rec (ref menu 'items)))

;; Activate protocol
;;
;;  interaction
;;    -> widget
;;        -> glim-menu-item-activate
;;             -> (call callback)
;;
;;  emitter is the widget (instance)
;;  subitem is used only for radio
;;
(define-method glim-menu-item-activate ((menu <glim-menu-item>) subitem emitter . args)
  (and-let* ((proc (ref menu 'callback)))
    (apply proc menu args))
  ;; Propagate state change to other instances.
  ;; TODO: This part depends on the realization of menu item (gtk widgets), so
  ;; should be separated from <glim-menu>.  But for now..
  ;; NB: I can't use gtk-check-button-set-active to change state of
  ;; other widgets, for it will emit activate signal and cause this method
  ;; called again.  Instead, I directly sets active field of the widget.
  ;; The Gtk document is not very clear about what is the right way.
  (case (ref menu 'type)
    ((check)
     (for-each (lambda (ins)
                 (unless (eq? (cdr ins) emitter)
                   (slot-set! (cdr ins) 'active (if (car args) 1 0))))
               (ref menu 'instances)))
    ((radio)
     (for-each (lambda (item)
                 (for-each (lambda (ins)
                             (slot-set! (cdr ins) 'active
                                        (if (eq? item subitem) 1 0)))
                           (ref item 'instances))
                 )
               (ref menu 'items)))
    ))

;; parse static menu definitions

(define (glim-menu-item def)
  (cond
   ((is-a? def <glim-menu-item>) def)
   ((list? def)
    (receive (type name label opts)
        (cond
         ((null? (cdr def)) (values (car def) #f "" '()))
         ((symbol? (cadr def))
          (if (string? (caddr def))
              (values (car def) (cadr def) (caddr def) (cdddr def))
              (values (car def) (cadr def)
                      (get-keyword :label (cddr def) (x->string (cadr def)))
                      (cddr def))))
         ((string? (cadr def))
          (values (car def) #f (cadr def) (cddr def)))
         (else (values (car def) #f
                       (get-keyword :label (cdr def) "")
                       (cdr def))))
      (unless (memq type '(menu button check separator radio))
        (error "unknown menu type" type))
      (let* ((self (make <glim-menu-item>
                     :type type :name name :options opts :label label
                     :callback (get-keyword :callback opts #f)
                     :value (or (get-keyword :value opts #f) label)))
             (children (map (lambda (def)
                              (let1 child (glim-menu-item def)
                                (slot-set! child 'parent self)
                                child))
                            (get-keyword :items opts '()))))
        (slot-set! self 'items children)
        self)))
   (else (error "bad menu item definition form" def))))

(define (glim-menu items)
  (let* ((children (map glim-menu-item items))
         (root (make <glim-menu> :items children)))
    (for-each (cut slot-set! <> 'parent root) children)
    root))

;; Dynamically adds/deletes menu
;;
;;   glim-menu-add-item parent item
;;      Adds item::<glim-menu-item> to parent.  Item must be an abstract
;;      item without any instances. From this method, glim-menu-add-item
;;      is called for each instance of parent to instantiate item.
;;
;;   glim-menu-delete-item parent item
;;      Deletes item from parent.  If item is not a child or parent,
;;      nothing happens.  If item is a child, then glim-menu-delete-item
;;      is called for each instance of parent, and the method must destroy
;;      the instance and remove it from item.instances.
;;
(define-method glim-menu-add-item ((parent <glim-menu>) item)
  ;; adding toplevel item
  (let1 item (glim-menu-item item) ;make sure item is item
    (set! (ref item 'parent) parent)
    (for-each (cut glim-menu-add-item <> item) (ref parent 'instances))
    (update! (ref parent 'items) (cut append <> (list item)))
    ))

(define-method glim-menu-add-item ((parent <glim-menu-item>) item)
  (let1 item (glim-menu-item item) ;make sure item is item
    (set! (ref item 'parent) parent)
    (for-each (lambda (ins)
                (glim-menu-add-item (car ins) (cdr ins) item))
              (ref parent 'instances))
    (update! (ref parent 'items) (cut append <> (list item)))
    ))

(define-method glim-menu-delete-item ((parent <glim-menu>) item)
  ;; TODO: writeme
  (error "removing toplevel menu is not supported yet"))

(define-method glim-menu-delete-item ((parent <glim-menu-item>) item)
  (when (memq item (ref parent 'items))
    (for-each (lambda (ins)
                (glim-menu-delete-item (car ins) (cdr ins) item))
              (ref parent 'instances))
    (set! (ref item 'parent) #f)
    (update! (ref parent 'items) (cut delete item <> eq?))
    ))

;;============================================================
;; Realization of menus
;;   An abstract <glim-menu> object can be realized by gtk widgets,
;;   using either glim-popup-menu or glim-pulldown-menu.

(define-class <glim-gtk-menu> (<glim-gadget>)
  ((abstract-menu :init-keyword :abstract-menu :init-value #f)))

;;------------------------------------------------------------
;; popup menu
;;

(define-class <glim-popup-menu> (<glim-gtk-menu>)
  ())

(define-method initialize ((self <glim-popup-menu>) initargs)
  (let1 menu (gtk-menu-new)
    (set! (widget-of self) menu)
    (next-method)
    (and-let* ((def (ref self 'abstract-menu)))
      (for-each (cut glim-gtk-menu-item-realize self menu <>) (ref def 'items))
      (add-instance def self))
    ))

;; convenience routine
(define (glim-popup-menu menu-def)
  (make <glim-popup-menu>
    :abstract-menu (if (is-a? menu-def <glim-menu>)
                       menu-def
                       (glim-menu menu-def))))

(define (glim-popup-menu-popup menu)
  (gtk-menu-popup (widget-of menu) #f #f #f 0 0))

;;------------------------------------------------------------
;; pulldown menu
;;

(define-class <glim-pulldown-menu> (<glim-gtk-menu>)
  ())

(define-method initialize ((self <glim-pulldown-menu>) initargs)
  (let1 menubar (gtk-menu-bar-new)
    (set! (widget-of self) menubar)
    (gtk-widget-show menubar)
    (next-method)
    (and-let* ((def (ref self 'abstract-menu)))
      (for-each (lambda (topitem)
                  (glim-gtk-menu-item-realize self menubar topitem))
                (ref def 'items))
      (add-instance def self))
    ))

(define (glim-pulldown-menu menu-def)
  (make <glim-pulldown-menu>
    :abstract-menu (if (is-a? menu-def <glim-menu>)
                       menu-def
                       (glim-menu menu-def))))

;;------------------------------------------------------------
;; common routine for realization
;;

;; placeholder for radiogroup
(define-class <glim-gtk-radio> ()
  ((glim-menu-item :init-keyword :glim-menu-item);point back to glim-menu-item
   ))
;; realize MENU-ITEM::<glim-menu-item> and its children
;; under PARENT::<gtk-menu-item> within ROOT::<glim-gtk-menu>
(define (glim-gtk-menu-item-realize root parent menu-item)

  ;; Connect widgets and menu-item, and set up signals.
  ;; Note that item == menu-item except radio, in which case menu-item
  ;; is the abstract radio item (placeholder) and item is each selection
  ;; in the radio; the widget instance is kept in the item, while callback
  ;; is called on the menu-item.
  (define (wire item widget . callback-args)
    (gtk-menu-shell-append parent widget)
    (gtk-widget-show widget)
    (add-instance item root widget)
    (g-signal-connect widget "destroy"
                      (lambda _ (delete-instance item widget)))
    (case (ref menu-item 'type)
      ((button)
       (g-signal-connect widget "activate"
                         (lambda _
                           (glim-menu-item-activate menu-item item widget))))
      ((check)
       (g-signal-connect widget "activate"
                         (lambda _
                           (glim-menu-item-activate menu-item item widget
                                                    (gtk-check-menu-item-get-active widget)))))
      ((radio)
       (g-signal-connect widget "activate"
                         (lambda _
                           (when (gtk-check-menu-item-get-active widget)
                             (glim-menu-item-activate menu-item item widget (car callback-args))))))
      )
    widget)
  
  (let* ((label (ref menu-item 'label)))
    (case (ref menu-item 'type)
      ((button) (wire menu-item (gtk-menu-item-new-with-label label)))
      ((check)  (wire menu-item (gtk-check-menu-item-new-with-label label)))
      ((separator) (wire menu-item (gtk-separator-menu-item-new)))
      ((menu)   (let* ((t (gtk-menu-item-new-with-label label))
                       (sub (gtk-menu-new)))
                  (gtk-menu-item-set-submenu t sub)
                  (for-each (cut glim-gtk-menu-item-realize root sub <>)
                            (ref menu-item 'items))
                  (wire menu-item t)))
      ((radio)  (let loop ((items (ref menu-item 'items))
                           (widgets '())
                           (group #f))
                  (if (null? items)
                      (add-instance menu-item root
                                    (make <glim-gtk-radio>
                                      :glim-menu-item menu-item))
                      (let* ((radio-item (car items))
                             (neww (wire radio-item
                                         (gtk-radio-menu-item-new-with-label
                                          group
                                          (ref radio-item 'label))
                                         (ref radio-item 'value))))
                        (loop (cdr items)
                              (cons neww widgets)
                              (gtk-radio-menu-item-get-group neww))))))
      (else (errorf "unknown menu item type ~s" (ref menu-item 'type)))))
  )

;; Add item --------------------------------------------

(define-method glim-menu-add-item ((self <glim-gtk-menu>)
                                   (item <glim-menu-item>))
  (glim-gtk-menu-item-realize self (widget-of self) item))

(define-method glim-menu-add-item ((self <glim-gtk-menu>)
                                   (parent <gtk-menu-item>)
                                   (item <glim-menu-item>))
  ;; TODO: writeme
  (error "adding new menu-item to gtk-menu-item is not supported yet")
  )

(define-method glim-menu-add-item ((self <glim-gtk-menu>)
                                   (parent <glim-gtk-radio>)
                                   (item <glim-menu-item>))
  (and-let* ((radio-group-item (ref parent 'glim-menu-item))
             (parent-widget    (get-widget-from-instances
                                self
                                (ref radio-group-item 'parent)))
             (radio-widgets    (map (cut get-widget-from-instances self <>)
                                    (ref radio-group-item 'items)))
             )
    (let* ((group (if (null? radio-widgets)
                      #f
                      (gtk-radio-menu-item-get-group (car radio-widgets))))
           (neww (gtk-radio-menu-item-new-with-label
                  group
                  (ref item 'label))))
      (gtk-menu-shell-append (gtk-menu-item-get-submenu parent-widget)
                             neww)
      (gtk-widget-show neww)
      (add-instance item self neww)
      (g-signal-connect neww "destroy"
                        (lambda _ (delete-instance item neww)))
      (g-signal-connect neww "activate"
                        (lambda _
                          (when (gtk-check-menu-item-get-active neww)
                            (glim-menu-item-activate radio-group-item
                                                     item neww
                                                     (ref item 'value)))))
      )
    ))

;; Remove item --------------------------------------------

(define-method glim-menu-delete-item ((self <glim-gtk-menu>)
                                      (item <glim-menu-item>))
  ;; TODO: writeme
  (error "not supported yet"))

(define-method glim-menu-delete-item ((self <glim-gtk-menu>)
                                      (parent <gtk-menu-item>)
                                      (item <glim-menu-item>))
  ;; TODO: writeme
  (error "adding new menu-item to gtk-menu-item is not supported yet")
  )

(define-method glim-menu-delete-item ((self <glim-gtk-menu>)
                                      (parent <glim-gtk-radio>)
                                      (item <glim-menu-item>))
  (for-each (lambda (ins)
              ;; The destroy signal handler takes care of removing instance
              ;; from item's instance list.
              (gtk-widget-destroy (cdr ins))
              )
            (ref item 'instances)))

;; util -------------------------------------------------

(define (get-widget-from-instances gtk-menu-root glim-menu-item)
  (cond ((assq gtk-menu-root (ref glim-menu-item 'instances)) => cdr)
        (else #f)))

(provide "glim/menu")
