;;;
;;; glim/pane/hselect.scm - hierarchy-selection pane, to be autoloaded.
;;;
;;;  Copyright(C) 2002-2003 by Shiro Kawai (shiro@acm.org)
;;;
;;;  Permission to use, copy, modify, distribute this software and
;;;  accompanying documentation for any purpose is hereby granted,
;;;  provided that existing copyright notices are retained in all
;;;  copies and that this notice is included verbatim in all
;;;  distributions.
;;;  This software is provided as is, without express or implied
;;;  warranty.  In no circumstances the author(s) shall be liable
;;;  for any damages arising out of the use of this software.
;;;
;;;  $Id: hselect.scm,v 1.6 2003/02/20 01:45:02 shirok Exp $
;;;

(select-module glim.pane)
(use srfi-1)
(use srfi-2)
(use srfi-13)
(use srfi-14)
(use file.util)

;;============================================================
;; Hierarchy-selector-mixin
;;
;;  Like File selection dialog, allows users to select an item
;;  from hierarchically organized data.   It can adapt to any
;;  tree-type data structure by providing appropriate callbacks.
;;  In the following explanation, I use terms of filesystem
;;  hierarchy for understandability.
;;
;;  "Path" is a name to refer to the item in the data structure.
;;  Internally it is represented by the list of components.
;;  The value of 'path-separator' is used to display it.
;;
;;  SLOTS:
;;    path-separator (string) : path elements are concatenated by
;;      this separator.  It is used to display paths.
;;
;;    expand-hierarchy-callback (procedure) : called when a new "directory"
;;      is selected.   Arguments are the pane and the path represented in
;;      the list of components.
;;      Should return two lists, one for "subdirectories" and
;;      the other for "regular files".   "subdirectories" shouldn't include
;;      the reference to itself and its parent "directory".
;;

(define-class <glim-hierarchy-selector-mixin> ()
  ((path-separator :init-keyword :path-separator
                   :init-value "/")
   (expand-hierarchy-callback :init-keyword :expand-hierarchy-callback
                              :init-value
                              (lambda (pane path)
                                (directory-list2
                                 (glim-hierarchy-path-string pane path)
                                 :children? #t)))
   ;; private
   (current-directory   :init-keyword :current-directory
                        :init-value '())
   ))

;; subclass has to specialize these methods
(define-method glim-expand-hierarchy ((pane <glim-hierarchy-selector-mixin>) path)
  (cond ((ref pane 'expand-hierarchy-callback) => (cut <> pane path))
        (else (values '() '()))))

(define-method glim-hierarchy-refresh
    ((pane <glim-hierarchy-selector-mixin>))
  #f)

;; getter/setter of current paths.

(define-method glim-hierarchy-current-directory
    ((self <glim-hierarchy-selector-mixin>))
  (ref self 'current-directory))

(define-method (setter glim-hierarchy-current-directory)
    ((self <glim-hierarchy-selector-mixin>) path)
  ;; Check for equality of current directory is necessary,
  ;; for glim-hierarchy-refresh may cause re-set of some parts of
  ;; widgets, which in turn calls this setter again.
  (unless (equal? path (ref self 'current-directory))
    (set! (ref self 'current-directory) path)
    (glim-hierarchy-refresh self)))

(define-method glim-hierarchy-path-string
    ((self <glim-hierarchy-selector-mixin>) . maybe-path)
  (let ((path (get-optional maybe-path (ref self 'current-directory)))
        (sep  (ref self 'path-separator)))
    (if (null? path)
        sep
        (string-join path sep 'prefix))))

(define-method glim-hierarchy-parse-path-string
    ((self <glim-hierarchy-selector-mixin>) path-string)
  (let* ((sep (ref self 'path-separator))
         (lis (string-split path-string sep)))
    (let loop ((lis lis)
               (r '()))
      (cond ((null? lis) (reverse! r))
            ((string-null? (car lis)) (loop (cdr lis) r))
            (else (loop (cdr lis) (cons (car lis) r)))))))

;;============================================================
;; Hierarchy-selector-pane
;;
;;  A realization of hierarchy-selector-mixin, pretty much the
;;  same layout as the gtk's default file selector.

(define-class <glim-hierarchy-selector-pane> (<glim-gtk-mixin>
                                              <glim-hierarchy-selector-mixin>
                                              <glim-value-gadget>
                                              <gtk-vbox>)
  ((dirlist-title  :init-keyword :dirlist-title
                   :init-value "Directories")
   (filelist-title :init-keyword :filelist-title
                   :init-value "Files")
   (path-option-pane)
   (dirlist-store)
   (dirlist-view)
   (filelist-store)
   (filelist-view)
   (selection-pane)
   ))

(define-method initialize ((self <glim-hierarchy-selector-pane>) initargs)
  (next-method)
  (let* ((o (make <glim-option-pane>
              :name-key car :value-key cdr
              :value-changed-callback
              (lambda (gadget value)
                (set! (glim-hierarchy-current-directory self) value))))
         (p (gtk-hpaned-new))
         (dw (gtk-scrolled-window-new #f #f))
         (ds (gtk-list-store-new <string>))
         (dv (gtk-tree-view-new-with-model ds))
         (fw (gtk-scrolled-window-new #f #f))
         (fs (gtk-list-store-new <string>))
         (fv (gtk-tree-view-new-with-model fs))
         (v  (gtk-entry-new)))
    ;; set up widgets
    (set! (ref self 'path-option-pane) o)
    (set! (ref self 'dirlist-store) ds)
    (set! (ref self 'dirlist-view) dv)
    (set! (ref self 'filelist-store) fs)
    (set! (ref self 'filelist-view) fv)
    (set! (ref self 'selection-pane) v)
    ;; layout widgets
    (gtk-paned-pack1 p dw #t #t)
    (gtk-scrolled-window-set-shadow-type dw GTK_SHADOW_IN)
    (gtk-container-add dw dv)
    (gtk-scrolled-window-set-policy dw GTK_POLICY_AUTOMATIC GTK_POLICY_ALWAYS)
    (gtk-widget-set-size-request dw 140 150)
    (gtk-paned-pack2 p fw #t #t)
    (gtk-scrolled-window-set-shadow-type fw GTK_SHADOW_IN)
    (gtk-container-add fw fv)
    (gtk-scrolled-window-set-policy fw GTK_POLICY_AUTOMATIC GTK_POLICY_ALWAYS)
    (gtk-widget-set-size-request fw 140 150)
    (gtk-container-set-border-width self 10)
    (gtk-box-pack-start self (widget-of o) #f #f 4)
    (gtk-box-pack-start self p #t #t 4)
    (gtk-box-pack-end self v #f #f 4)
    ;; tree stuff
    (let ((dirlist-column
           (gtk-tree-view-column-new-with-attributes
            (ref self 'dirlist-title)
            (gtk-cell-renderer-text-new)
            "text" 0))
          (filelist-column
           (gtk-tree-view-column-new-with-attributes
            (ref self 'filelist-title)
            (gtk-cell-renderer-text-new)
            "text" 0))
          )
      (define (dir-activated . _)
        (and-let* ((iter (gtk-tree-selection-get-selected
                          (gtk-tree-view-get-selection dv)))
                   (dir  (gtk-tree-model-get-value ds iter 0))
                   (path (append (ref self 'current-directory) (list dir))))
          (set! (glim-hierarchy-current-directory self) path)
          (gtk-entry-set-text v (glim-hierarchy-path-string self path))
          ))
      (define (file-activated . _)
        (and-let* ((iter (gtk-tree-selection-get-selected
                          (gtk-tree-view-get-selection fv)))
                   (file (gtk-tree-model-get-value fs iter 0))
                   (path (append (ref self 'current-directory) (list file))))
          (gtk-entry-set-text v (glim-hierarchy-path-string self path))))
      (define (selection-activated . _)
        (let1 val (glim-gadget-value self)
          (glim-value-changed-callback self val)))
      
      (gtk-tree-view-append-column dv dirlist-column)
      (gtk-tree-view-append-column fv filelist-column)

      (g-signal-connect dv "row-activated" dir-activated)
      (g-signal-connect fv "row-activated" file-activated)
      (g-signal-connect v "activate" selection-activated)
      )
    (glim-refresh-view self)
    ))

(define-method glim-hierarchy-refresh ((self <glim-hierarchy-selector-pane>))
  (define (list-set lis items)
    (gtk-list-store-clear lis)
    (for-each (lambda (item)
                (let ((iter (gtk-list-store-append lis)))
                  (gtk-list-store-set-value lis iter 0 (x->string item))))
              items))
  (define (break-paths current)
    (pair-fold-right (lambda (cdrs knil)
                       (let1 p (reverse cdrs)
                         (acons (glim-hierarchy-path-string self p) p knil)))
                     (list (list (ref self 'path-separator)))
                     (reverse current)))
  (let ((cur  (ref self 'current-directory)))
    (receive (dirs files)
        (glim-expand-hierarchy self cur)
      (list-set (ref self 'dirlist-store) dirs)
      (list-set (ref self 'filelist-store) files))
    (set! (ref (ref self 'path-option-pane) 'items) (break-paths cur))
    )
  )

(define-method glim-refresh-view ((self <glim-hierarchy-selector-pane>))
  (glim-hierarchy-refresh self)
  )

(define-method glim-gadget-value ((self <glim-hierarchy-selector-pane>))
  (glim-hierarchy-parse-path-string
   self
   (gtk-entry-get-text (ref self 'selection-pane))))

(define-method (setter glim-gadget-value)
    ((self <glim-hierarchy-selector-pane>) path)
  (gtk-entry-set-text (ref self 'selection-pane)
                      (glim-hierarchy-path-string self path)))

(provide "glim/pane/hselect")

