;;;
;;; glim/presentation.scm - presentation
;;;
;;;  Copyright(C) 2002-2003 by Shiro Kawai (shiro@acm.org)
;;;
;;;  Permission to use, copy, modify, distribute this software and
;;;  accompanying documentation for any purpose is hereby granted,
;;;  provided that existing copyright notices are retained in all
;;;  copies and that this notice is included verbatim in all
;;;  distributions.
;;;  This software is provided as is, without express or implied
;;;  warranty.  In no circumstances the author(s) shall be liable
;;;  for any damages arising out of the use of this software.
;;;
;;;  $Id: presentation.scm,v 1.2 2003/01/01 00:59:50 shirok Exp $
;;;

(define-module glim.presentation
  (use gauche.parameter)
  (export <glim-presentation>
          glim-presentation?
          glim-presentation-object
          glim-presentation-type
          glim-presentation-type-of
          glim-present
          glim-present-to-string
          glim-object-present
          glim-input-context
          glim-accept
          glim-accept-from-string
          glim-object-accept
          glim-ptype-parameterize
          <glim-any> <glim-number> <glim-complex> <glim-real> <glim-rational>
          <glim-integer>
          ))
(select-module glim.presentation)

;;=============================================================
;; Presentation base class
;;
(define-class <glim-presentation> ()
  ((object :init-keyword :object :accessor glim-presentation-object)
   (type   :init-keyword :type :accessor glim-presentation-type)
   ))

(define (glim-presentation? obj)
  (is-a? obj <glim-presentation>))

;;=============================================================
;; Present
;;

(define (glim-present object . args)
  (receive (ptype opts)
      (if (and (not (null? args)) (not (keyword? (car args))))
          (values (ensure-ptype (car args)) (cdr args))
          (values (glim-presentation-type-of object) args))
    (let ((viewport (get-keyword :view opts (current-output-port))))
      (apply glim-object-present object type (ensure-viewport viewport) opts))))

; define glim-present-to-string

(define-method glim-object-present (object type viewport . args)
  (write object (ref viewport 'oport)))

;;=============================================================
;; Accept
;;

(define glim-input-context (make-parameter '()))

; define glim-accept

; define glim-accept-from-string

;;=============================================================
;; Viewport
;;

;; Viewport is a combination of extended-stream and view in CLIM.
;; The view is defined as a class, and the instance of the view
;; contains the necessary information about the stream.

(define-class <glim-viewport> ()
  ())

;; The default viewport

(define-class <glim-textual-viewport> (<glim-viewport>)
  ((iport :init-keyword :iport :init-form (current-input-port))
   (oport :init-keyword :oport :init-form (current-output-port)))
  )

(define-method ensure-viewport ((vp <glim-viewport>)) vp)
  
;;=============================================================
;; Presentation types
;;

;; Unlike CLIM, presentation types are specified by the first
;; class objects.

(define-class <glim-ptype-meta> (<class>)
  ())

(define-class <glim-any> ()
  ()
  :metaclass <glim-ptype-meta>)

(define-class <glim-nil> (<glim-any>)
  ())

;;--------------------------------------------------------------
;; mechanism of parameterization
;;  ensure-ptype allows the programmer to pass either a ptype class
;;  (in case it is not parameterized) or a ptype instance (parameterized).
;;  Unlike CLIM, presentation type arguments and options can be passed
;;  together.  The method glim-ptype-parameterize of the specified
;;  presentation type has to interpret them.

(define-method object-apply ((ptype <glim-ptype-meta>) . args)
  (let1 pt (make ptype)
    (apply glim-ptype-parameterize (make ptype) args)
    pt))

(define-method glim-ptype-parameterize ((ptype <glim-any>) . args)
  (errorf "presentation type ~a can't be parameterized with arguments ~s"
          (class-of ptype) args))

(define-method ensure-ptype ((ptype <glim-any>)) ptype)
(define-method ensure-ptype ((ptype <glim-ptype-meta>)) (make ptype))

;;--------------------------------------------------------------
;; predefined presentation types
;;

;; Numeric types
(define-class <glim-number> (<glim-any>)
  ())

(define-class <glim-complex> (<glim-number>)
  ())

(define-class <glim-real> (<glim-complex>)
  ((low  :init-value #f)
   (high :init-value #f))
  )

(define-method glim-ptype-parameterize ((ptype <glim-real>) low high)
  (slot-set! ptype 'low low)
  (slot-set! ptype 'high high))

(define-class <glim-rational> (<glim-real>)
  ())

(define-class <glim-integer> (<glim-rational>)
  ())

;; Character and string
(define-class <glim-char> (<glim-any>)
  ())

(define-class <glim-string> (<glim-any>)
  ((length :init-value #f)))

(define-method glim-ptype-parameterize ((ptype <glim-string>))
  ptype)

(define-method glim-ptype-parameterize ((ptype <glim-string>) length)
  (slot-set! ptype 'length length))

;; One-of and some-of
;;  <glim-set> implements the common function.  Others are just for
;;  different interfaces.
(define-class <glim-sel> (<glim-any>)
  ((possibilities :init-value '())
   (tester        :init-value eqv?)
   (value-key     :init-value identity)
   (name-key      :init-value identity)
   ))

(define-class <glim-multisel-mixin> ()
  ((separator     :init-value ", ")))

(define-method glim-sel-common-keywords ((ptype <glim-sel>) args)
  (let ((tester    (get-keyword :test args #f))
        (value-key (get-keyword :value-key args #f))
        (name-key  (get-keyword :name-key args #f)))
    (when tester    (slot-set! ptype 'tester tester))
    (when value-key (slot-set! ptype 'value-key value-key))
    (when name-key  (slot-set! ptype 'name-key name-key))
    ))

(define-method glim-sel-common-keywords ((ptype <glim-multisel-mixin>) args)
  (next-method)
  (let ((separator (get-keyword :separator args #f)))
    (when separator (slot-set! ptype 'separator separator))))

(define-class <glim-member-sequence> (<glim-sel>) ())
(define-class <glim-subset-sequence> (<glim-multisel-mixin>
                                      <glim-member-sequence>) ())

(define-method glim-ptype-parameterize ((ptype <glim-member-sequence>)
                                        (coll <collection>) . args)
  (glim-sel-common-keywords ptype args)
  (slot-set! ptype 'possibilities coll))

(define <glim-completion> <glim-member-sequence>) ;just a synonym
(define <glim-subset-completion> <glim-subset-sequence>) ;ditto

(define-class <glim-member> (<glim-sel>) ())
(define-class <glim-subset> (<glim-multisel-mixin> <glim-member>) ())

(define-method glim-ptype-parameterize ((ptype <glim-member>) . args)
  (slot-set! ptype 'possibilities args))

(define-class <glim-member-alist> (<glim-sel>) ())
(define-class <glim-subset-alist> (<glim-multisel-mixin> <glim-member-alist>)
  ())

(define-method initialize ((self <glim-member-alist>) initargs)
  (next-method)
  (slot-set! self 'name-key car)
  (slot-set! self 'value-key cadr)
  )

(define-method glim-ptype-parameterize ((ptype <glim-member-alist>)
                                        alist . args)
  (glim-sel-common-keywords ptype args)
  (slot-set! ptype 'possibilities alist))

;;--------------------------------------------------------------
;; sequence presentation
;;

;; writeme

;;--------------------------------------------------------------
;; meta presentation
;;

(define-class <glim-or> (<glim-any>)
  ((types :init-value '())))

(define-method glim-ptype-parameterize ((ptype <glim-or>) . types)
  (slot-set! ptype 'types types))

;; glim-and

;;--------------------------------------------------------------
;; lisp form presentation
;;

;; writeme



(provide "glim/presentation")
