;;;
;;; x3d/field - X3D & VRML Field definitions
;;;
;;;  Copyright(C) 2003 by Shiro Kawai (shiro@acm.org)
;;;
;;;  Permission to use, copy, modify, distribute this software and
;;;  accompanying documentation for any purpose is hereby granted,
;;;  provided that existing copyright notices are retained in all
;;;  copies and that this notice is included verbatim in all
;;;  distributions.
;;;  This software is provided as is, without express or implied
;;;  warranty.  In no circumstances the author(s) shall be liable
;;;  for any damages arising out of the use of this software.
;;;
;;;  $Id: field.scm,v 1.6 2003/02/21 05:33:41 shirok Exp $
;;;

(define-module x3d.field
  (use srfi-1)
  (use srfi-13)
  (use gauche.uvector)
  (use gauche.sequence)
  (use gauche.parameter)
  (use gauche.charconv)
  (use x3d.node)
  (export-all))
(select-module x3d.field)

;;===========================================================
;; Fields
;;

(define-class <x3d-field-meta> (<class>)
  ((x3d-class-name :init-keyword :x3d-class-name) ; like "SFBool"
   (default-value  :init-keyword :default-value)  ; class default value
   (element-type   :init-keyword :element-type)   ; scheme class of elt type
   ))

(define-method initialize ((self <x3d-field-meta>) initargs)
  (next-method)
  (unless (slot-bound? self 'x3d-class-name)
    (slot-set! self 'x3d-class-name
               (string-trim-both (x->string (class-name self)) #[<>])))
  )

(define-class <x3d-field> ()
  ((x3d-name       :init-keyword :x3d-name)       ;slot name in X3D (string)
   (slot-name      :init-keyword :slot-name)      ;slot name in Scheme
   (container-name :init-keyword :container-name) ;Scheme class name of 
					          ;  container
   (default-value  :init-keyword :default-value)  ;override default value
   )
  :metaclass <x3d-field-meta>)

(define-method initialize ((self <x3d-field>) initargs)
  (next-method)
  (unless (slot-bound? self 'default-value)
    (slot-set! self 'default-value
               (if (slot-bound? (class-of self) 'default-value)
                   (ref (class-of self) 'default-value)
                   #f)))
  )

;; Check given value is the same as the default value of field or not
(define-method x3d-default-value? ((self <x3d-field>) val)
  (equal? (slot-ref self 'default-value) val))

;;-----------------------------------------------------------
;; single value field mixin
;;

(define-class <x3d-SField> (<x3d-field>)
  ())

(define-method x3d-write ((self <x3d-SField>) val port dict)
  (unless (and (x3d-write-compact) (x3d-default-value? self val))
    (display (ref self 'x3d-name) port)
    (display " " port)
    (x3d-write-basic self val port dict)
    (newline port)))

;;-----------------------------------------------------------
;; multi value field class
;;

(define-class <x3d-MField> (<x3d-field>)
  ()
  :default-value '())

(define-method x3d-write ((self <x3d-MField>) val port dict)
  (unless (and (x3d-write-compact) (x3d-default-value? self val))
    (display (ref self 'x3d-name) port)
    (display " [" port)
    (unless (null? val)
      (x3d-write-basic self (car val) port dict)
      (for-each (lambda (v)
                  (display ", " port)
                  (x3d-write-basic self v port dict))
                (cdr val)))
    (display "]\n" port)))

;;============================================================
;; Fields
;;

(define-macro (define-x3d-field field elttype . maybe-default)
  (let-optionals* maybe-default ((default `(make ,elttype)))
    (let* ((base-field (string->symbol #`"<x3d-,|field|>"))
           (sfield     (string->symbol #`"<S,|field|>"))
           (mfield     (string->symbol #`"<M,|field|>")))
      `(begin
         (define-class ,base-field () ())
         (define-class ,sfield (,base-field <x3d-SField>) ()
           :element-type ,elttype :default-value ,default)
         (define-class ,mfield (,base-field <x3d-MField>) ()
           :element-type ,elttype :default-value '()))
      )))

;; SFBool and MFBool ------------------------------------------
(define-x3d-field FBool <boolean> #f)

(define-method x3d-write-basic ((self <x3d-FBool>) val port dict)
  (display (if val "TRUE" "FALSE") port))

;; SFColor and MFColor ----------------------------------------
(define-x3d-field FColor <sequence> (f32vector 0 0 0))

(define-method x3d-write-basic ((self <x3d-FColor>) val port dict)
  (format port "~a ~a ~a" (ref val 0) (ref val 1) (ref val 2)))

;; SFFloat and MFFloat ----------------------------------------
(define-x3d-field FFloat <real> 0.0)

(define-method x3d-write-basic ((self <x3d-FFloat>) val port dict)
  (display val port))

;; SFImage and MFImage ----------------------------------------
;;   pixels may be u32, u16 or u8 vector.

(define-class <x3d-image> ()
  ((width  :init-keyword :width :init-value 0)
   (height :init-keyword :height :init-value 0)
   (num-components :init-keyword :num-components :init-value 0)
   (pixels :init-keyword :pixels :init-value #u8())
   ))

(define-x3d-field FImage <x3d-image>)

(define-method x3d-write-basic ((self <x3d-FImage>) val port dict)
  (format port "~a ~a ~a "
          (ref val 'width) (ref val 'height)
          (ref val 'num-components))
  (for-each (lambda (val)
              (if (zero? val)
                  (display "0" port)
                  (format port "0x~X" val)))
            (ref val 'pixels)))

;; SFInt32 and MFInt32e ----------------------------------------

(define-x3d-field FInt32 <integer> 0)

(define-method x3d-write-basic ((self <x3d-FInt32>) val port dict)
  (display val port))

;; SFNode and MFNode -------------------------------------------

(define-x3d-field FNode <x3d-node> #f)

(define-method x3d-write-basic ((self <x3d-FNode>) (val <x3d-node>) port dict)
  (x3d-write-node val port dict))

(define-method x3d-write-basic ((self <x3d-FNode>) val port dict)
  (display "NULL" port))

;; SFRotation and MFRotation -----------------------------------
(define-x3d-field FRotation <sequence> (f32vector 0 0 1 0))

(define-method x3d-write-basic ((self <x3d-FRotation>) val port dict)
  (format port "~a ~a ~a ~a" (ref val 0) (ref val 1) (ref val 2) (ref val 3))
  )

;; SFString and MFString ---------------------------------------

(define-x3d-field FString <string> "")

(define-method x3d-write-basic ((self <x3d-FString>) val port dict)
  (let1 in (wrap-with-input-conversion (open-input-string val) 'utf-8)
    (with-port-locking in
      (lambda ()
        (write-char #\" port)
        (port-for-each (lambda (b)
                         (cond ((memv b '(#x22 #x5c)) ;; #\" or #\\
                                (write-byte #x5c port)
                                (write-byte b port))
                               (else (write-byte b port))))
                       (cut read-byte in))
        (write-char #\" port)
        ))))

;; SFTime and MFTime -------------------------------------------

(define-x3d-field FTime <real> -1)

(define-method x3d-write-basic ((self <x3d-FTime>) val port dict)
  (display val port))

;; SFVec2f and MFVec2f -----------------------------------------

(define-x3d-field FVec2f <sequence> (f32vector 0 0))

(define-method x3d-write-basic ((self <x3d-FVec2f>) val port dict)
  (format port "~a ~a" (ref val 0) (ref val 1)))

;; SFVec3f and MFVec3f -----------------------------------------
(define-x3d-field FVec3f <sequence> (f32vector 0 0 0))

(define-method x3d-write-basic ((self <x3d-FVec3f>) val port dict)
  (format port "~a ~a ~a" (ref val 0) (ref val 1) (ref val 2)))

(provide "x3d/field")
