;;;
;;; x3d/node - X3D & VRML common node definitions
;;;
;;;  Copyright(C) 2003 by Shiro Kawai (shiro@acm.org)
;;;
;;;  Permission to use, copy, modify, distribute this software and
;;;  accompanying documentation for any purpose is hereby granted,
;;;  provided that existing copyright notices are retained in all
;;;  copies and that this notice is included verbatim in all
;;;  distributions.
;;;  This software is provided as is, without express or implied
;;;  warranty.  In no circumstances the author(s) shall be liable
;;;  for any damages arising out of the use of this software.
;;;
;;;  $Id: node.scm,v 1.6 2003/02/18 04:25:46 shirok Exp $
;;;

(define-module x3d.node
  (use srfi-2)
  (use srfi-13)
  (use gauche.parameter)
  (use gauche.mop.instance-pool)
  (export-all))
(select-module x3d.node)

;;=====================================================================
;; Node metaclass
;;

(define-class <x3d-node-meta> (<instance-pool-meta>)
  ((x3d-fields      :initform '())
   (x3d-eventios    :initform '())
   (x3d-class-name  :init-keyword :x3d-class-name :getter x3d-class-name)
   ))

;;-------------------------------------------------------------------
;; Node base class
;;

(define-class <x3d-node> ()
  ((node-name  :init-keyword :node-name :accessor node-name
               :initform #f)
   (routes     :init-keyword :routes :initform '()) ; route information
   )
  :metaclass <x3d-node-meta>)

;; node dictionary is a temporary hashtable used during 
;; deserialization

(define (make-x3d-node-dict) (make-hash-table 'string=?))

(define (x3d-node-dict-lookup dict name)
  (hash-table-get dict name #f))

(define (x3d-node-dict-intern dict name node)
  (hash-table-put! dict name node))
   
;; Parameter to switch whether the field that has default value shoudl
;; be printed or not
(define x3d-write-compact (make-parameter #f))


(define-generic x3d-write )

;; Generic node writer
(define-method x3d-write-scene ((scene <x3d-node>) . keys)
  (let-keywords* keys ((port (current-output-port))
                       (compact #f))
    (with-port-locking port
      (lambda ()
        (parameterize ((x3d-write-compact compact))
          (display "#VRML V2.0 utf8\n" port)
          (x3d-write-node scene port (make-x3d-node-dict))
          (newline port))))))

(define-method x3d-write-node ((self <x3d-node>) port dict)
  (let ((name (node-name self)))
    (if (and name (x3d-node-dict-lookup dict name))
        (display #`"USE ,name\n" port)
        (begin
          (when name
            (x3d-node-dict-intern dict name self)
            (display #`"DEF ,name " port))
          (display (x3d-class-name (class-of self)) port)
          (display " {\n" port)
          (dolist (slot (class-slots (class-of self)))
            (and-let* ((fclass (slot-definition-option slot :field-class #f))
                       (name   (slot-definition-name slot)))
              (x3d-write fclass (slot-ref self name) port dict)))
          (dolist (route (slot-ref self 'routes))
            (format port "ROUTE ~A.~A TO ~A.~A~%"
                    (from-node route) (from-field route)
                    (to-node route) (to-field route)))
          (display "}\n" port)))
    ))

;;-------------------------------------------------------------------
;; VRML node meta operations
;;

(define-macro (define-vrml-class name supers slots)

  (define x3d-class-name
    (string-trim-both (x->string name) #[<>]))

  (define (get-real-slot slot-def)
    (unless (= (length slot-def) 3)
      (error "bad slot definition" slot-def))
    (list* (car slot-def)
           :init-keyword (make-keyword (car slot-def))
           (cond ((or (get-keyword :field (cdr slot-def) #f)
                      (get-keyword :exposed-field (cdr slot-def) #f))
                  => (cut apply field-slot (car slot-def) <>))
                 ((or (get-keyword :event-in (cdr slot-def) #f)
                      (get-keyword :event-out (cdr slot-def) #f))
                  ;; TODO: for now
                  => (lambda (f) `(:event-io ',f)))
                 (else (error "bad slot definition" slot-def)))))

  (define (field-slot slot-name field-class field-name . maybe-default)
    (list :field-class
          `(make ,field-class
                 :x3d-name ,field-name :slot-name ',slot-name
                 :container-name ,x3d-class-name
                 ,@(if (null? maybe-default)
                       '()
                       `(:default-value ,(car maybe-default))))))

  `(begin
     (define-class ,name (,@supers <x3d-node>)
       ,(map get-real-slot slots)
       :x3d-class-name ,x3d-class-name)
     (define-method initialize ((self ,name) initargs)
       (next-method)
       (initialize-vrml-node self)
       )
     (define-macro (,(string->symbol x3d-class-name) . args)
       `(make ,,name ,@args))
     )
  )

(define-method initialize-vrml-node ((self <x3d-node>))
  (dolist (slot (class-slots (class-of self)))
    (and-let* ((fclass (slot-definition-option slot :field-class #f))
               (slot-name (slot-definition-name slot))
               ((not (slot-bound? self slot-name))))
      (slot-set! self slot-name (ref fclass 'default-value))))
  )

;;-------------------------------------------------------------------
;; Route
;;   A structure used to keep route information

(define-class <x3d-route> ()
  ((from-node  :init-keyword :from-node  :getter from-node)
   (from-field :init-keyword :from-field :getter from-field)
   (to-node    :init-keyword :to-node    :getter to-node)
   (to-field   :init-keyword :to-field   :getter to-field)
   ))

(provide "x3d/node")
