;;;
;;; banyan.server - server-side repository management
;;;
;;;  Copyright(C) 2002-2003 by Shiro Kawai (shiro@acm.org)
;;;
;;;  Permission to use, copy, modify, distribute this software and
;;;  accompanying documentation for any purpose is hereby granted,
;;;  provided that existing copyright notices are retained in all
;;;  copies and that this notice is included verbatim in all
;;;  distributions.
;;;  This software is provided as is, without express or implied
;;;  warranty.  In no circumstances the author(s) shall be liable
;;;  for any damages arising out of the use of this software.
;;;
;;;  $Id: server.scm,v 1.19 2003/02/22 11:49:09 shirok Exp $
;;;

;; The banyan data can be accessed like url 
;;   http://server/.../repository/path/to/data
;; The "path/to/data" part is used to name the resource.  When accessed
;; from normal web client, this script returns an HTML that explains
;; the resouce.  When accessed from Banyan client, which is distinguished
;; by Accept: application/x-banyan request header, this script returns
;; S-expression.

(define-module banyan.server
  (use www.cgi)
  (use srfi-1)
  (use srfi-2)
  (use srfi-13)
  (use file.util)
  (use text.html-lite)
  (use text.tree)
  (use rfc.cookie)
  (use rfc.base64)
  (use gauche.logger)
  (use gauche.parameter)
  (use gauche.sequence)
  (use gauche.mop.singleton)
  (use util.list)
  (use banyan.license)
  (export <repository> banyan-main)
  )
(select-module banyan.server)

(autoload banyan.server.cvs
          cvs-commit cvs-add cvs-add-project cvs-add-path*)
(autoload banyan.server.data
          repository-read)
(autoload banyan.server.export
          server-export-to-vrml)

(define plog
  (let ((remote (sys-getenv "REMOTE_ADDR")))
    (lambda (format . args)
      (apply log-format (string-append "~a:~a:" format)
             remote (current-user) args))))

;;=================================================
;; Repository
;;

(define-class <repository> ()
  (;; Customizable parameters
   (root            :init-keyword :root
                    :init-value   "/home/shiro/data/BANYAN")
   (uri             :init-keyword :uri
                    :init-value   "/schemearts.com/banyan/repository")
   (name            :init-keyword :name
                    :init-value   "Test repository")
   (account-file    :init-keyword :account-file
                    :init-value   "users.bn")
   (style-sheet     :init-keyword :style-sheet
                    :init-value   "banyan.css")
   (charset         :init-keyword :charset
                    :init-value   "iso-8859-1")
   (discussion-uri  :init-keyword :discussion-uri
                    :init-value   "/schemearts.com/banyan/wiki")
   (vrml-exporter   :init-keyword :vrml-exporter
                    :init-value   "/home/shiro/bin/banyan2vrml")
   ))

(define-method repo-path ((self <repository>) . paths)
  (apply build-path (ref self 'root) "checkout"
         (map sanitize-path paths)))

(define-method repo-uri ((self <repository>) . paths)
  (apply build-path (ref self 'uri) paths))

(define-method account-path ((self <repository>))
  (and-let* ((file (ref self 'account-file))
             (path (if (absolute-path? file)
                       file
                       (build-path (ref self 'root) file)))
             ((file-exists? path)))
    path))

(define-method style-sheet-path ((self <repository>))
  (and-let* ((file (ref self 'style-sheet))
             (path (if (absolute-path? file)
                       file
                       (build-path (sys-dirname (ref self 'uri)) file))))
    path))

(define-method open-log ((self <repository>))
  (log-open (build-path (ref self 'root) "log") :lock-policy 'file))

(define (sanitize-path path-component)
  (when (string-index path-component #[^[:alpha:][:digit:]/._-])
    (error "bad path component" path-component))
  path-component)

(define-constant part-file "part.bn")   ;actual file name

(define-constant thumbnail-file "thumb.png") ; thumbnail base name

;;=================================================
;; Metadata retrieval
;;
(define (repository-read-metadata repo path)
  (let ((file (repo-path repo path part-file)))
    (and (file-exists? file)
         (call-with-input-file file read))))

(define (meta-get meta key . default)
  (get-keyword key (cdr meta) (get-optional default #f)))

(define (meta-revision meta)
  (format-revision (meta-get meta :revision)))

(define (meta-description meta)
  (meta-get meta :description))

(define (meta-modification-time meta)
  (format-time (meta-get meta :modification-time 0)))

(define (meta-modified-by meta)
  (meta-get meta :modified-by))

(define (meta-creation-time meta)
  (format-time (meta-get meta :creation-time 0)))

(define (meta-created-by meta)
  (meta-get meta :created-by))

(define (meta-license meta)
  (meta-get meta :license))

;;=================================================
;; Retrieve request parameters
;;
(define-syntax memo-thunk
  (syntax-rules ()
    ((_ . body)
     (let1 r (delay . body) (lambda args (force r))))))

(define request-uri
  (memo-thunk (sys-getenv "REQUEST_URI")))

(define (resource-path repo)
  (or (and-let* ((uri  (request-uri))
                 (path (car (string-split uri #\?)))
                 (z    (member (sys-basename (ref repo 'uri))
                               (string-split path #\/))))
        (canonicalize-resource-path (cdr z)))
      ""))

(define request
  (memo-thunk
   (and-let* ((uri (request-uri)))
     (cond ((string-suffix? "?info" uri) 'info)
           ((string-suffix? "?list" uri) 'list)
           ((string-suffix? "?login" uri) 'login)
           ((string-suffix? "?logout" uri) 'logout)
           ((string-suffix? "?wrl" uri) 'get-vrml)
           ((string-scan uri #\? 'after)
            => (cut error "unknown command" <>))
           (else #f)))))

(define cookies
  (memo-thunk
   (and-let* ((str (sys-getenv "HTTP_COOKIE")))
     (parse-cookie-string str))))

(define client-is-banyan?
  (memo-thunk
   (and-let* ((accept (sys-getenv "HTTP_ACCEPT")))
     (string-contains-ci accept "application/x-banyan"))))

(define request-method
  (memo-thunk
   (if (and-let* ((meth (sys-getenv "REQUEST_METHOD")))
         (string-ci=? meth "POST"))
       'post
       'get)))

(define current-user (make-parameter #f))
(define current-pass (make-parameter #f))

;;=================================================
;; Authentication
;;

;; For the time being, authentication is done via http cookie protocol.
;; We just send username and password over the network; later, probably
;; challenge/response protocol will be implemented.

;; Return values of authenticate routine:
;;   'ok  - user&pass is valid.  current-user and current-pass is set.
;;   'bad-user - no such user.
;;   'bad-pass - bad password
;;
;; Note that it's ok if no auth info is given from the client.
;; current-user and current-pass is set to #f in such a case.

;; if there's no password file, accept any user.
(define (valid-user? repo user pass)
  (or (and-let* ((user)
                 (path (account-path repo))
                 (user-pass (call-with-input-file path port->sexp-list)))
        (let1 record (assoc user user-pass)
          (cond ((not record) 'bad-user)
                ((equal? pass (cadr record)) 'ok)
                (else 'bad-pass))))
      'ok))

(define (decode-auth string)
  (apply values (string-split (base64-decode-string string) #\x00)))

(define (encode-auth user pass)
  (base64-encode-string (string-append user "\x00" pass)))

(define (authenticate repo)
  (receive (user pass)
      (cond ((and-let* ((cook (cookies))
                        (p (assoc "banyan-auth" cook)))
               (cadr p))
             => decode-auth)
            (else (values #f #f)))
    (let1 r (valid-user? repo user pass)
      (when (eq? r 'ok) (current-user user) (current-pass pass))
      r)))

(define (banyan-cookies repo)
  (construct-cookie-string
   (if (and (current-user) (current-pass))
       `(("banyan-auth"
          ,(encode-auth (current-user) (current-pass))
          :max-age 864000 :path ,(ref repo 'uri)
          ))
       `(("banyan-auth" "" :max-age 0 :path ,(ref repo 'uri))))))

;;=================================================
;; Utilities
;;

;; copied from www.cgi - allowing extra headers
(define (cgi-header . args)
  (let ((content-type (get-keyword :content-type args "text/html"))
        (location     (get-keyword :location args #f))
        (cookies      (get-keyword :cookies args '()))
        (cookies2     (get-keyword :cookies2 args '()))
        (headers      (get-keyword :headers args '())))
    (list (if location
              (list "Location: " (x->string location) "\n")
              (list "Content-type: " (x->string content-type) "\n"))
          (map (cut list "Set-cookie: " <> "\n") cookies)
          (map (cut list "Set-cookie2: " <> "\n") cookies2)
          (map (lambda (header-pair)
                 (map (cut list (car header-pair) ":" <> "\n")
                      (cdr header-pair)))
               headers)
          "\n")));;terminating header

(define (http-redirect repo)
  (write-tree
   (cgi-header :headers '(("status" "302 Moved"))
               :location (ref repo 'uri)
               :cookies (banyan-cookies repo))))

(define (banyan-reply expr . more)
  (let ((status (if (pair? expr)
                    (case (car expr)
                      ((error) "error")
                      ((bad-auth) (cadr expr))
                      (else "ok"))
                    "ok")))
    (write-tree (cgi-header :content-type "application/x-banyan"
                            :headers `(("x-banyan-status" ,status))))
    (write expr) (newline)
    (for-each (lambda (x) (write x) (newline)) more)))

(define (html-reply repo head body)
  (write-tree
   (list (cgi-header
          :content-type #`"text/html; charset=,(ref repo 'charset)"
          :content-style-type "text/css"
          :cookies (banyan-cookies repo))
         (html-doctype)
         (html:html
          (html:head head
                     (if repo
                         (html:link :rel "stylesheet"
                                    :href (style-sheet-path repo)
                                    :type "text/css")
                         '()))
          body))
   ))

;;; Standard error method generator
;;;   two error methods are required; one is called before repo is
;;;   instantiated, and the other is called after.
(define (cgi-error-proc repo)
  (lambda (e)
    (plog "error ~a" (slot-ref e 'message))
    (if (client-is-banyan?)
        (banyan-reply `(error ,(ref e 'message)))
        (let ((trace (vm-get-stack-trace-lite)))
          (html-reply
           repo
           (html:title "Error")
           (html:body (html:h1 "Error")
                      (html:p (html-escape-string (slot-ref e 'message)))
                      (map (lambda (s)
                             (html:p (html-escape-string (write-to-string s))))
                           trace)))
          )))
  )

(define (list-parts repo path)
  (let ((dir (repo-path repo path)))
    (reverse!
     (fold
      (lambda (entry knil)
        (cond ((or (string=? entry ".") (string=? entry "..")) knil)
              ((not (file-is-directory? (build-path dir entry))) knil)
              (else
               (let ((subdir? (has-subdirectory? (build-path dir entry)))
                     (part? (file-exists? (build-path dir entry part-file))))
                 (if (or subdir? part?)
                     (cons (list entry :part part? :subdir subdir?) knil)
                     knil)))))
      '()
      (sort (sys-readdir dir))))))

(define (has-subdirectory? path)
  (let loop ((entries (sys-readdir path)))
    (cond
     ((null? entries) #f)
     ((member (car entries) '("." ".." "CVS"))
      (loop (cdr entries)))
     ((file-is-directory? (build-path path (car entries))) #t)
     (else (loop (cdr entries))))))

(define (canonicalize-resource-path components)
  (let loop ((r '())
             (p components))
    (cond ((null? p) (string-join (reverse! r) "/"))
          ((string=? (car p) ".") (loop r (cdr p)))
          ((string=? (car p) "..")
           (when (null? r)
             (error "bad resource path" (string-join components "/")))
           (loop (cdr r) (cdr p)))
          ((#/^[\w_-]+[\w_\.-]*$/ (car p))
           (loop (cons (car p) r) (cdr p)))
          ((string=? (car p) "")
           (loop r (cdr p)))
          (else
           (errorf "bad resource path ~s (component ~s has invalid char)"
                   (string-join components "/") (car p)))
          )))

(define (get-thumbnail-path repo path)
  (let1 tpath (build-path (ref repo 'root) path thumbnail-file)
    (if (file-exists? tpath)
        tpath
        (build-path (ref repo 'uri) ".." thumbnail-file))))

;;=================================================
;; HTML output
;;

(define (format-time time)
  (if (number? time)
      (sys-strftime "%Y/%m/%d %H:%M:%S %Z" (sys-localtime time))
      "-"))

(define (format-revision rev)
  (if rev (string-join (map x->string rev)  ".") "?"))

(define (html-title repo path)
  (if (string-null? path) (ref repo 'name) (sys-basename path)))

(define (html-user-tab repo)
  (let ((user (current-user)))
    (if user
        (html:p :class "userTab" #`"Welcome, ,|user|. "
                (html:a :href "?logout" "[logout]"))
        (html:p :class "userTab" "Welcome, guest. "
                (html:a :href "?login" "[login]")))))

(define (html-path-navi repo path)
  (html:p "Path: "
          (html:a :href (repo-uri repo) (html:b #`"[,(ref repo 'name)]"))
          (pair-fold
           (lambda (cdrs r)
             (list* (html:b "/")
                    (html:a :href (apply repo-uri repo (reverse cdrs))
                            (html:b (car cdrs)))
                    r))
           '()
           (reverse (string-split path #\/)))))

(define (html-part-list repo path)
  (let1 parts
      (directory-list (repo-path repo path)
                      :children? #t
                      :filter
                      (lambda (p)
                        (and
                         (not (equal? p "CVS"))
                         (file-is-directory? (repo-path repo path p)))))
    (and
     (pair? parts)
     (html:table
      (html:tr (html:th :align "left" :class "listActive" "Name")
               (html:th :align "left" :class "listNormal" "Rev")
               (html:th :align "left" :class "listNormal" "Last modified")
               (html:th :align "left" :class "listNormal" "Modified by")
               (html:th :align "left" :class "listNormal" "Description")
               )
      (map (lambda (part class)
             (or
              (and-let* ((subpath (if (string-null? path)
                                      part
                                      (build-path path part)))
                         (meta (repository-read-metadata repo subpath)))
                (html:tr (html:td :class class
                                  (html:a :href (repo-uri repo subpath) part))
                         (html:td :class class (meta-revision meta))
                         (html:td :class class (meta-modification-time meta))
                         (html:td :class class (meta-modified-by meta))
                         (html:td :class class (meta-description meta))
                         ))
              (html:tr (html:td :class class
                                (html:a :href (repo-uri repo path part) part))
                       (html:td :class class "----")
                       (html:td :class class "----")
                       (html:td :class class "----")
                       (html:td :class class "----")
                       )))
           parts
           (circular-list "oddRow" "evenRow"))
      ))
    ))

(define (html-node-info repo path meta)
  (html:table
   (html:tr (html:td :rowspan 4 :valign "top"
                     (html:img :src (get-thumbnail-path repo path)))
            (html:th "Name")
            (html:td (meta-get meta :name))
            (html:th "Revision")
            (html:td (meta-revision meta)))
   (html:tr (html:th "Modified by")
            (html:td (meta-modified-by meta))
            (html:th "Last modified")
            (html:td (meta-modification-time meta)))
   (html:tr (html:th "Created by")
            (html:td (meta-created-by meta))
            (html:th "Creation time")
            (html:td (meta-creation-time meta)))
   (html:tr (html:th :align "left" "Description")
            (html:td :align "left" :colspan 3 (meta-description meta)))
   ;;(html:tr (html:th :align "left" "License")
   ;;        (html:td :align "left" :colspan 3 (meta-license meta)))
   ))

(define (html-node-history repo path meta)
  (html:table
   (html:tr (html:th :align "left" :class "listNormal" "Revision")
            (html:th :align "left" :class "listNormal" "Modified by")
            (html:th :align "left" :class "listNormal" "Time")
            (html:th :align "left" :class "listNormal" "Message"))
   (map-with-index
    (lambda (i e)
      (let ((class (if (odd? i) "oddRow" "evenRow")))
        (html:tr
         (html:td :class class
                  (format-revision (get-keyword :revision e #f)))
         (html:td :class class
                  (get-keyword :modified-by e "?"))
         (html:td :class class
                  (format-time (get-keyword :modification-time e #f)))
         (html:td :class class
                  (get-keyword :log e "*** empty log ***")))))
    (take* (meta-get meta :history '()) 20))
   ))

(define (html-info-page repo path)
  (html:body
   (html:h1 (html-title repo path))
   (html-user-tab repo)
   (html-path-navi repo path)
   (html:hr)
   (let1 meta (repository-read-metadata repo path)
     (cond-list
      (meta (html-node-info repo path meta))
      ((and meta (ref repo 'vrml-exporter))
       (html:div :align "right"
                 (html:a :href #`",(repo-uri repo path)?wrl"
                         "Export to VRML")))
      ((ref repo 'discussion-uri)
       => (lambda (wiki)
            (html:div :align "right"
                      (html:a :href #`",|wiki|?comment/,|path|" "View")
                      " or "
                      (html:a :href #`",|wiki|?c=e&p=comment/,|path|" "Add")
                      #`" comments on [,(ref repo 'name)]/,path ["
                      (html:a :href #`",|wiki|?Help" "?") "]")))
      ((html-part-list repo path)
       => (cut list (html:hr) (html:h3 "Subparts") <>))
      (meta (list (html:hr)
                  (html:h3 "ChangeLog")
                  (html-node-history repo path meta)))))
   (html:hr)
   (html:div :align "right" "Banyan repository manager $Revision: 1.19 $")
   ))

;;=================================================
;; Commands
;;

(define (cmd-error repo path)
  (errorf "bad request for ~s" path))

(define (cmd-change-info repo path)
  (errorf "change-info: not supported yet for ~s" path))

(define (cmd-check-in repo path)
  (unless (client-is-banyan?)
    (error "you can only check-in from Banyan client"))
  (plog "check in ~a" path)
  (let* ((len (string->number (or (sys-getenv "CONTENT_LENGTH") "0")))
         (data (if (positive? len)
                   (string-incomplete->complete (read-block len))
                   (port->string (current-input-port))))
         (fn   (repo-path repo path part-file))
         (fnb  (string-append fn ".orig")))
    (if (file-exists? fn)
        (begin
          (sys-rename fn fnb)
          (with-output-to-file fn (cut display data))
          (cvs-commit repo #f (build-path path part-file))
          (banyan-reply '()))
        (begin
          (cvs-add-path* repo path)
          (with-output-to-file fn (cut display data))
          (cvs-commit repo #t (build-path path part-file))
          (banyan-reply '())))))

(define (cmd-get-info repo path)
  (errorf "get-info: not supported yet for ~s" path))

(define (cmd-check-out repo path)
  (if (client-is-banyan?)
      (begin
        (plog "check out ~s" path)
        (cond
         ((file-exists? (repo-path repo path part-file))
          (apply banyan-reply (repository-read repo path)))
         (else (errorf "check-out: no such path: ~s" path))))
      (html-reply repo
                  (html:title (html-title repo path))
                  (html-info-page repo path))))

(define (cmd-list repo path)
  (if (client-is-banyan?)
      (begin
        (plog "list ~s" path)
        (banyan-reply (list-parts repo path)))
      (html-reply repo
                  (html:title (html-title repo path))
                  (html-info-page repo path))))

(define (cmd-login-screen repo message)
  (if (client-is-banyan?)
      (banyan-reply '())
      (html-reply
       repo
       (html:title "Banyan Login")
       (html:body
        message 
        (html:h1 #`"Login to ,(ref repo 'name)")
        (html:form
         :method "POST" :action #`",(ref repo 'uri)?login"
         (html:table
          :border 1
          (html:tr
           (html:th "username:")
           (html:td (html:input :type "text" :name "user" :size 16)))
          (html:tr
           (html:th "password:")
           (html:td (html:input :type "password" :name "pass" :size 16)))
          )
         (html:input :type "submit" :name "submit" :value "Login"))))
      ))

(define (cmd-get-vrml repo path)
  (server-export-to-vrml repo path))

(define (cmd-login repo)
  (if (client-is-banyan?)
      (banyan-reply '())
      (let* ((q     (cgi-parse-parameters))
             (user  (cgi-get-parameter "user" q))
             (pass  (cgi-get-parameter "pass" q)))
        (if (and user pass (valid-user? repo user pass))
            (begin (current-user user)
                   (current-pass pass)
                   (http-redirect repo))
            (cmd-login-screen repo
                              (html:p :style "color: red;"
                                      "username or password is invalid"))))
      ))

(define (cmd-logout repo)
  (current-user #f)
  (current-pass #f)
  (http-redirect repo))

(define (cmd-bad-auth repo auth)
  (plog "bad auth (~a)" auth)
  (if (client-is-banyan?)
      (banyan-reply `(bad-auth ,auth))
      (cmd-login-screen
       repo
       (html:p :style "color: red; font: large;"
               (if (eq? auth 'bad-user)
                   "Username is not recognized.  Please login."
                   "Wrong password.  Please login again.")))
      ))

;;; NB: we don't use cgi-main, for the "POST" method sends
;;; S-expresion, not the form data.
(define (banyan-main repo)
  (with-error-handler (cgi-error-proc repo)
    (lambda ()
      (open-log repo)
      (let ((path (resource-path repo))
            (auth (authenticate repo)))
        (case auth
          ((ok)
           (if (eq? (request-method) 'post)
               (case (request)
                 ((info) (cmd-change-info repo path))
                 ((list) (cmd-error repo path))
                 ((login) (cmd-login repo))
                 ((logout) (cmd-logout repo))
                 ((get-vrml) (error "invalid command" (request)))
                 (else   (cmd-check-in repo path)))
               (case (request)
                 ((info) (cmd-get-info repo path))
                 ((list) (cmd-list repo path))
                 ((login) (cmd-login-screen repo '()))
                 ((logout) (cmd-logout repo))
                 ((get-vrml) (cmd-get-vrml repo path))
                 (else   (cmd-check-out repo path)))))
          (else (cmd-bad-auth repo auth))
          ))))
  0)

(provide "banyan/server")

;; Local variables:
;; mode: scheme
;; end:
