/*
 * light.c - light handling
 *
 *  Copyright(C) 2002-2003 by Shiro Kawai (shiro@acm.org)
 *
 *  Permission to use, copy, modify, distribute this software and
 *  accompanying documentation for any purpose is hereby granted,
 *  provided that existing copyright notices are retained in all
 *  copies and that this notice is included verbatim in all
 *  distributions.
 *  This software is provided as is, without express or implied
 *  warranty.  In no circumstances the author(s) shall be liable
 *  for any damages arising out of the use of this software.
 *
 *  $Id: light.c,v 1.6 2003/12/10 01:00:50 shirok Exp $
 */

#include "banyan.h"

/*===============================================================
 * Lights
 */

/*----------------------------------------------------------------
 * Class stuff
 */

static ScmObj light_allocate(ScmClass *, ScmObj);
static void light_render(BanyanLight *light, int id);

SCM_DEFINE_BUILTIN_CLASS(Banyan_LightClass, NULL,
                      NULL, NULL, light_allocate,
                      Banyan_NodeCPL);

ScmObj light_allocate(ScmClass *klass, ScmObj initargs)
{
    BanyanLight *light = SCM_ALLOCATE(BanyanLight, klass);
    SCM_SET_CLASS(light, klass);
    BANYAN_HEADER_INIT(light);
    light->renderer  = light_render;
    light->ambient   = SCM_VECTOR4F(Scm_MakeVector4f(0.0,0.0,0.0,1.0));
    light->diffuse   = SCM_VECTOR4F(Scm_MakeVector4f(1.0,1.0,1.0,1.0));
    light->specular  = SCM_VECTOR4F(Scm_MakeVector4f(1.0,1.0,1.0,1.0));
    light->transform = BANYAN_TRANSFORM(Banyan_MakeTransform());
    light->spot_exponent = 0.0;
    light->spot_cutoff = 180.0;
    light->constant_attenuation = 1.0;
    light->linear_attenuation = 0.0;
    light->quadratic_attenuation = 0.0;
    return SCM_OBJ(light);
}

ScmObj Banyan_MakeLight(void)
{
    return light_allocate(BANYAN_LIGHT_CLASS, SCM_NIL);
}

#define LIGHT_COLOR_GET_SET(field)                                        \
  static ScmObj SCM_CPP_CAT3(light_,field,_get)(ScmObj light)             \
  {                                                                       \
    return SCM_OBJ(BANYAN_LIGHT(light)->field);                           \
  }                                                                       \
  static void SCM_CPP_CAT3(light_,field,_set)(ScmObj light, ScmObj value) \
  {                                                                       \
    if (!SCM_VECTOR4FP(value)) {                                          \
        Scm_Error("<vector4f> is required, but got %S", value);           \
    }                                                                     \
    BANYAN_LIGHT(light)->field = SCM_VECTOR4F(value);                     \
  }

LIGHT_COLOR_GET_SET(ambient);
LIGHT_COLOR_GET_SET(diffuse);
LIGHT_COLOR_GET_SET(specular);

static ScmObj light_transform_get(ScmObj light)
{
    return SCM_OBJ(BANYAN_LIGHT(light)->transform);
}

static void light_transform_set(ScmObj light, ScmObj value)
{
    if (!BANYAN_TRANSFORM_P(value)) {
        Scm_Error("<%transform> is required, but got %S", value);
    }
    BANYAN_LIGHT(light)->transform = BANYAN_TRANSFORM(value);
}

#define LIGHT_PARAM_GET_SET(field)                                        \
  static ScmObj SCM_CPP_CAT3(light_,field,_get)(ScmObj light)             \
  {                                                                       \
    return Scm_MakeFlonum((double)BANYAN_LIGHT(light)->field);            \
  }                                                                       \
  static void SCM_CPP_CAT3(light_,field,_set)(ScmObj light, ScmObj value) \
  {                                                                       \
    BANYAN_LIGHT(light)->field = (float)Scm_GetDouble(value);             \
  }

LIGHT_PARAM_GET_SET(spot_exponent);
LIGHT_PARAM_GET_SET(spot_cutoff);
LIGHT_PARAM_GET_SET(constant_attenuation);
LIGHT_PARAM_GET_SET(linear_attenuation);
LIGHT_PARAM_GET_SET(quadratic_attenuation);

/*
 * Default render method
 */

static void light_render(BanyanLight *light, int id)
{
    float position[4];
    float *gmat;
        
    glLightfv(id, GL_AMBIENT, SCM_VECTOR4F_D(light->ambient));
    glLightfv(id, GL_DIFFUSE, SCM_VECTOR4F_D(light->diffuse));
    glLightfv(id, GL_SPECULAR, SCM_VECTOR4F_D(light->specular));

    gmat = SCM_MATRIX4F_D(light->transform->gmatrix);
    position[0] = gmat[12]; position[1] = gmat[13];
    position[2] = gmat[14]; position[3] = gmat[15];

    glLoadIdentity();
    glLightfv(id, GL_POSITION, position);
    /* direction set */
    glLightf(id, GL_SPOT_EXPONENT, light->spot_exponent);
    glLightf(id, GL_SPOT_CUTOFF, light->spot_cutoff);
    glLightf(id, GL_CONSTANT_ATTENUATION, light->constant_attenuation);
    glLightf(id, GL_LINEAR_ATTENUATION, light->linear_attenuation);
    glLightf(id, GL_QUADRATIC_ATTENUATION, light->quadratic_attenuation);
}

/*===============================================================
 * Light Groups
 */

/*----------------------------------------------------------------
 * Class stuff
 */

static ScmObj lightgroup_allocate(ScmClass *, ScmObj);

SCM_DEFINE_BUILTIN_CLASS(Banyan_LightGroupClass, NULL,
                      NULL, NULL, lightgroup_allocate,
                      Banyan_NodeCPL);

ScmObj lightgroup_allocate(ScmClass *klass, ScmObj initargs)
{
    BanyanLightGroup *lg = SCM_ALLOCATE(BanyanLightGroup, klass);
    SCM_SET_CLASS(lg, klass);
    BANYAN_HEADER_INIT(lg);
    lg->lights  = SCM_VECTOR(Scm_MakeVector(BANYAN_MAX_HARDWARE_LIGHTS,
                                            SCM_FALSE));
    return SCM_OBJ(lg);
}

ScmObj Banyan_MakeLightGroup(void)
{
    return lightgroup_allocate(BANYAN_LIGHT_GROUP_CLASS, SCM_NIL);
}

static ScmObj lightgroup_lights_get(ScmObj lg)
{
    return SCM_OBJ(BANYAN_LIGHT_GROUP(lg)->lights);
}

static void lightgroup_lights_set(ScmObj lg, ScmObj value)
{
    int i, len;
    if (!SCM_VECTORP(value)) {
        Scm_Error("<vector> is required, but got %S", value);
    }
    len = SCM_VECTOR_SIZE(value);
    for (i=0; i<len; i++) {
        ScmObj e = SCM_VECTOR_ELEMENT(value, i);
        if (!SCM_FALSEP(e) && !BANYAN_LIGHT_P(e)) {
            Scm_Error("light vector contains bad object: %S", e);
        }
    }
    BANYAN_LIGHT_GROUP(lg)->lights = SCM_VECTOR(value);
}

/*----------------------------------------------------------------
 * Render protocol
 */

void Banyan_LightGroupRender(BanyanLightGroup *lg, ScmObj mode)
{
    int i;
    static GLenum ids[] = { GL_LIGHT0, GL_LIGHT1, GL_LIGHT2, GL_LIGHT3,
                            GL_LIGHT4, GL_LIGHT5, GL_LIGHT6, GL_LIGHT7 };
    
    if (!SCM_EQ(mode, Banyan_RenderShade)) return;
    for (i=0; i<BANYAN_MAX_HARDWARE_LIGHTS; i++) {
        ScmObj light = SCM_VECTOR_ELEMENT(lg->lights, i);
        if (BANYAN_LIGHT_P(light)) {
            BANYAN_LIGHT(light)->renderer(BANYAN_LIGHT(light), ids[i]);
            glEnable(ids[i]);
        } else {
            glDisable(ids[i]);
        }
    }
}

/*===============================================================
 * Initialization stuff
 */
                                        
static ScmClassStaticSlotSpec light_slots[] = {
    BANYAN_NODE_SLOT_SPEC,
    SCM_CLASS_SLOT_SPEC("ambient", light_ambient_get, light_ambient_set),
    SCM_CLASS_SLOT_SPEC("diffuse", light_diffuse_get, light_diffuse_set),
    SCM_CLASS_SLOT_SPEC("specular", light_specular_get, light_specular_set),
    SCM_CLASS_SLOT_SPEC("transform", light_transform_get, light_transform_set),
    SCM_CLASS_SLOT_SPEC("spot-exponent",
                        light_spot_exponent_get, light_spot_exponent_set),
    SCM_CLASS_SLOT_SPEC("spot-cutoff",
                        light_spot_cutoff_get, light_spot_cutoff_set),
    SCM_CLASS_SLOT_SPEC("constant-attenuation",
                        light_constant_attenuation_get, light_constant_attenuation_set),
    SCM_CLASS_SLOT_SPEC("linear-attenuation",
                        light_linear_attenuation_get, light_linear_attenuation_set),
    SCM_CLASS_SLOT_SPEC("quadratic-attenuation",
                        light_quadratic_attenuation_get, light_quadratic_attenuation_set),
    { NULL }
};

static ScmClassStaticSlotSpec lightgroup_slots[] = {
    BANYAN_NODE_SLOT_SPEC,
    SCM_CLASS_SLOT_SPEC("lights", lightgroup_lights_get, lightgroup_lights_set),
    { NULL }
};

void Banyan__InitLight(ScmModule *mod)
{
    Scm_InitBuiltinClass(&Banyan_LightClass, "<%light>", light_slots,
                         sizeof(BanyanLight), mod);
    Scm_InitBuiltinClass(&Banyan_LightGroupClass, "<%light-group>", lightgroup_slots,
                         sizeof(BanyanLightGroup), mod);
}
