/*
 * shape.c - shape handling
 *
 *  Copyright(C) 2002-2003 by Shiro Kawai (shiro@acm.org)
 *
 *  Permission to use, copy, modify, distribute this software and
 *  accompanying documentation for any purpose is hereby granted,
 *  provided that existing copyright notices are retained in all
 *  copies and that this notice is included verbatim in all
 *  distributions.
 *  This software is provided as is, without express or implied
 *  warranty.  In no circumstances the author(s) shall be liable
 *  for any damages arising out of the use of this software.
 *
 *  $Id: shape.c,v 1.11 2003/12/10 01:00:50 shirok Exp $
 */

#include "banyan.h"

/* Check if an element of shape->facep is sane */
#define VALID_FACEP(fp)                                 \
    (SCM_PAIRP(fp)&&SCM_INTP(SCM_CAR(fp))               \
     &&SCM_PAIRP(SCM_CDR(fp))&&SCM_UVECTORP(SCM_CADR(fp)))

static int num_faces_in_fragment(ScmObj fp);

/* Render normal scale.  if 0, normals are not rendered. */
static float renderNormalScale = 0.0;

/*----------------------------------------------------------------
 * Class stuff
 */

static ScmObj shape_allocate(ScmClass *, ScmObj);

SCM_DEFINE_BUILTIN_CLASS(Banyan_ShapeClass, NULL,
                      NULL, NULL, shape_allocate,
                      Banyan_NodeCPL);

ScmObj shape_allocate(ScmClass *klass, ScmObj initargs)
{
    BanyanShape *shape = SCM_ALLOCATE(BanyanShape, klass);
    SCM_SET_CLASS(shape, klass);
    BANYAN_HEADER_INIT(shape);
    shape->vertices = SCM_FALSE;
    shape->normals = SCM_FALSE;
    shape->faces = SCM_NIL;
    shape->numFaces = 0;
    shape->facetNormals = SCM_FALSE;
    return SCM_OBJ(shape);
}

ScmObj Banyan_MakeShape(void)
{
    return shape_allocate(BANYAN_SHAPE_CLASS, SCM_NIL);
}

static ScmObj shape_vertices_get(ScmObj shape)
{
    return BANYAN_SHAPE(shape)->vertices;
}

static void shape_vertices_set(ScmObj shape, ScmObj value)
{
    if (!SCM_FALSEP(value) && !SCM_POINT4F_ARRAY_P(value)) {
        Scm_Error("point4f-array or #f is required, but got %S", value);
    }
    BANYAN_SHAPE(shape)->vertices = value;
}

static ScmObj shape_normals_get(ScmObj shape)
{
    return BANYAN_SHAPE(shape)->normals;
}

static void shape_normals_set(ScmObj shape, ScmObj value)
{
    if (!SCM_FALSEP(value) && !SCM_VECTOR4F_ARRAY_P(value)) {
        Scm_Error("vector4f-array or #f is required, but got %S", value);
    }
    BANYAN_SHAPE(shape)->normals = value;
    if (SCM_FALSEP(value)) {
        /* TODO: revise semantics */
        /* Usually, setting #f to normals is the way to tell the shape
           to recalculate normals.  It is likely that facetNormals should
           be recalculated, so we invalidate it. */
        BANYAN_SHAPE(shape)->facetNormals = SCM_FALSE;
    }
}

static ScmObj shape_faces_get(ScmObj shape)
{
    return BANYAN_SHAPE(shape)->faces;
}

static void shape_faces_set(ScmObj shape, ScmObj value)
{
    ScmObj faces;
    int nfaces = 0;
    
    if (!SCM_LISTP(value)) {
        Scm_Error("list required, but got %S", value);
    }
    SCM_FOR_EACH(faces, value) {
        ScmObj fp = SCM_CAR(faces);
        if (!VALID_FACEP(fp)) {
            Scm_Error("bad data for shape faces: %S", value);
        }
        nfaces += num_faces_in_fragment(fp);
    }
    BANYAN_SHAPE(shape)->faces = value;
    BANYAN_SHAPE(shape)->numFaces = nfaces;
    /* invalidate cached facenormals */
    BANYAN_SHAPE(shape)->facetNormals = SCM_FALSE;
}

static ScmObj shape_numfaces_get(ScmObj shape)
{
    return Scm_MakeInteger(BANYAN_SHAPE(shape)->numFaces);
}

static ScmObj shape_facetnormals_get(ScmObj shape)
{
    return BANYAN_SHAPE(shape)->facetNormals;
}

/*----------------------------------------------------------------
 * Render protocol
 */

static void render_normals(BanyanShape *shape);

void Banyan_ShapeRender(BanyanShape *shape, ScmObj mode)
{
    ScmObj faces, fp;
    if (SCM_POINT4F_ARRAY_P(shape->vertices)) {
        glVertexPointer(4, GL_FLOAT, 0, SCM_POINT4F_ARRAY_D(shape->vertices));
        if (SCM_EQ(mode, BANYAN_RENDER_SHADE)) {
            glEnableClientState(GL_NORMAL_ARRAY);
            if (!SCM_VECTOR4F_ARRAY_P(shape->normals)) {
                Banyan_ShapeSetVertexNormals(shape);
            }
            glNormalPointer(GL_FLOAT, 4*sizeof(float),
                            SCM_VECTOR4F_ARRAY_D(shape->normals));
        }
        SCM_FOR_EACH(faces, shape->faces) {
            ScmUVector *v;
            GLenum mode;
            fp = SCM_CAR(faces);
            SCM_ASSERT(VALID_FACEP(fp));
            mode = SCM_INT_VALUE(SCM_CAR(fp));
            v = SCM_UVECTOR(SCM_CADR(fp));
            if (SCM_U32VECTORP(v)) {
                glDrawElements(mode, SCM_U32VECTOR_SIZE(v),
                               GL_UNSIGNED_INT,
                               SCM_U32VECTOR_ELEMENTS(v));
            } else if (SCM_U16VECTORP(v)) {
                glDrawElements(mode, SCM_U16VECTOR_SIZE(v),
                               GL_UNSIGNED_SHORT,
                               SCM_U16VECTOR_ELEMENTS(v));
            } else if (SCM_U8VECTORP(v)) {
                glDrawElements(mode, SCM_U8VECTOR_SIZE(v),
                               GL_UNSIGNED_BYTE,
                               SCM_U8VECTOR_ELEMENTS(v));
            }
        }
        if (renderNormalScale > 0.0) {
            render_normals(shape);
        }
    }
}

/* special function to render normals */
static void render_normals(BanyanShape *shape)
{
    BanyanFaceIter iter;
    
    if (!SCM_VECTOR4F_ARRAY_P(shape->facetNormals)) {
        Banyan_ShapeSetFacetNormals(shape);
        if (!SCM_VECTOR4F_ARRAY_P(shape->facetNormals)) return;
    }
    if (!Banyan_FaceIterInit(shape, &iter)) return;
    do {
        float from[4];
        float to[4];
        float *normal = SCM_VECTOR4F_ARRAY_REFV(shape->facetNormals,
                                                BANYAN_FACE_ITER_COUNT(&iter));
        Banyan_FaceIterGetCenter(&iter, from);
        SCM_VECTOR4F_OP(_, to[_] = from[_] + normal[_] * renderNormalScale);
        glBegin(GL_LINES);
        glVertex4fv(from);
        glVertex4fv(to);
        glEnd();
    } while (Banyan_FaceIterNext(&iter));
}

void Banyan_ShapeRenderNormals(double scale)
{
    renderNormalScale = scale;
}

/*----------------------------------------------------------------
 * Face iterators
 */

/* utility */
int num_faces_in_fragment(ScmObj fp)
{
    int mode = SCM_INT_VALUE(SCM_CAR(fp));
    ScmUVector *v = SCM_UVECTOR(SCM_CADR(fp));
    switch (mode) {
    case GL_TRIANGLES:      return SCM_UVECTOR_SIZE(v)/3;
    case GL_TRIANGLE_STRIP: return SCM_UVECTOR_SIZE(v)-2;
    case GL_TRIANGLE_FAN:   return SCM_UVECTOR_SIZE(v)-2;
    case GL_QUADS:          return SCM_UVECTOR_SIZE(v)/4;
    case GL_QUAD_STRIP:     return (SCM_UVECTOR_SIZE(v)-2)/2;
    case GL_POLYGON:        return 1;
    default:
        Scm_Warn("unsupported GL draw mode: %d", mode);
        return 0;
    }
}

/* initialize iterator */
int Banyan_FaceIterInit(BanyanShape *shape, BanyanFaceIter *iter)
{
    iter->shape = shape;
    iter->facep = shape->faces;
    if (!SCM_NULLP(iter->facep)) {
        ScmObj fp = SCM_CAR(iter->facep);
        SCM_ASSERT(VALID_FACEP(fp));
        iter->mode = SCM_INT_VALUE(SCM_CAR(fp));
        iter->ivec = SCM_UVECTOR(SCM_CADR(fp));
        iter->index = 0;
        iter->size = SCM_UVECTOR_SIZE(iter->ivec);
        iter->count = 0;
        return TRUE;
    } else {
        return FALSE;
    }
}

/* proceed iterator.  returns TRUE if there's a data, or FALSE if it's
   already end. */
int Banyan_FaceIterNext(BanyanFaceIter *iter)
{
    ScmObj fp;
    if (BANYAN_FACE_ITER_END(iter)) return FALSE;
    iter->count++;
    switch(iter->mode) {
    case GL_TRIANGLES:
        iter->index += 3;
        if (iter->index < iter->size-2) return TRUE;
        break;
    case GL_TRIANGLE_STRIP:
        iter->index++;
        if (iter->index < iter->size-2) return TRUE;
        break;
    case GL_TRIANGLE_FAN:
        iter->index++;
        if (iter->index < iter->size-1) return TRUE;
        break;
    case GL_QUADS:
        iter->index += 4;
        if (iter->index < iter->size-3) return TRUE;
        break;
    case GL_QUAD_STRIP:
        iter->index += 2;
        if (iter->index < iter->size-3) return TRUE;
        break;
    case GL_POLYGON:
        break;
    default:
        Scm_Warn("unsupported GL draw mode: %d", iter->mode);
        return FALSE;
    }
    /* next faceset */
    iter->facep = SCM_CDR(iter->facep);
    if (!SCM_NULLP(iter->facep)) {
        fp = SCM_CAR(iter->facep);
        SCM_ASSERT(VALID_FACEP(fp));
        iter->mode = SCM_INT_VALUE(SCM_CAR(fp));
        iter->ivec = SCM_UVECTOR(SCM_CADR(fp));
        iter->index = 0;
        iter->size = SCM_UVECTOR_SIZE(iter->ivec);
        return TRUE;
    }
    return FALSE;
}

/* returns # of vertices of the face current iterator points */
int Banyan_FaceIterGetNumVertices(BanyanFaceIter *iter)
{
    if (BANYAN_FACE_ITER_END(iter)) return 0;
    switch(iter->mode) {
    case GL_TRIANGLES:;      /*FALLTHROUGH*/
    case GL_TRIANGLE_STRIP:; /*FALLTHROUGH*/
    case GL_TRIANGLE_FAN:
        return 3;
    case GL_QUADS:;          /*FALLTHROUGH*/
    case GL_QUAD_STRIP:
        return 4;
    case GL_POLYGON:
        return iter->size;
    default:
        Scm_Warn("unsupported GL draw mode: %d", iter->mode);
        return 0;
    }
    /*NOTREACHED*/
}

/* generic accessor for uvector */
static int index_vector_ref(ScmUVector *ivec, int index)
{
    if (SCM_U32VECTORP(ivec)) return (int)SCM_U32VECTOR_ELEMENTS(ivec)[index];
    if (SCM_U16VECTORP(ivec)) return (int)SCM_U16VECTOR_ELEMENTS(ivec)[index];
    if (SCM_U8VECTORP(ivec))  return (int)SCM_U8VECTOR_ELEMENTS(ivec)[index];
    Scm_Warn("bad uniform vector in face data: %S", ivec);
    return 0;
}

/* fills vertex indices of current face */
void Banyan_FaceIterGetIndices(BanyanFaceIter *iter, GLuint indices[])
{
    if (BANYAN_FACE_ITER_END(iter)) return;
    switch(iter->mode) {
    case GL_TRIANGLES:
        indices[0] = index_vector_ref(iter->ivec, iter->index);
        indices[1] = index_vector_ref(iter->ivec, iter->index+1);
        indices[2] = index_vector_ref(iter->ivec, iter->index+2);
        break;
    case GL_TRIANGLE_STRIP:
        indices[0] = index_vector_ref(iter->ivec, iter->index);
        if (iter->index % 2 == 0) {
            indices[1] = index_vector_ref(iter->ivec, iter->index+1);
            indices[2] = index_vector_ref(iter->ivec, iter->index+2);
        } else {
            indices[2] = index_vector_ref(iter->ivec, iter->index+1);
            indices[1] = index_vector_ref(iter->ivec, iter->index+2);
        }
        break;
    case GL_TRIANGLE_FAN:
        indices[0] = index_vector_ref(iter->ivec, 0);
        indices[1] = index_vector_ref(iter->ivec, iter->index);
        indices[2] = index_vector_ref(iter->ivec, iter->index+1);
        break;
    case GL_QUADS:
        indices[0] = index_vector_ref(iter->ivec, iter->index);
        indices[1] = index_vector_ref(iter->ivec, iter->index+1);
        indices[2] = index_vector_ref(iter->ivec, iter->index+2);
        indices[3] = index_vector_ref(iter->ivec, iter->index+3);
        break;
    case GL_QUAD_STRIP:
        indices[0] = index_vector_ref(iter->ivec, iter->index);
        indices[1] = index_vector_ref(iter->ivec, iter->index+1);
        indices[3] = index_vector_ref(iter->ivec, iter->index+2);
        indices[2] = index_vector_ref(iter->ivec, iter->index+3);
        break;
    case GL_POLYGON: {
        int i;
        for (i=0; i<iter->size; i++) {
            indices[i] = index_vector_ref(iter->ivec, i);
        }
        break;
    }
    default:
        Scm_Warn("unsupported GL draw mode: %d", iter->mode);
    }
}

/* returns vertex index of vindex-th vertex of the face. */
int Banyan_FaceIterGetIndex(BanyanFaceIter *iter, int index)
{
    if (BANYAN_FACE_ITER_END(iter)) return 0;
    
    switch(iter->mode) {
    case GL_TRIANGLES:
        index += iter->index;
        break;
    case GL_TRIANGLE_STRIP:
        if (index == 0) {
            index = iter->index;
        } else if (iter->index % 2 == 0) {
            index += iter->index;
        } else {
            if (index == 1) index = iter->index+2;
            else            index = iter->index+1;
        }
        break;
    case GL_TRIANGLE_FAN:
        if (index == 0) break;
        else index = iter->index + (index-1);
        break;
    case GL_QUADS:
        index += iter->index;
        break;
    case GL_QUAD_STRIP:
        switch (index) {
        case 0:; case 1: index += iter->index; break;
        case 2:; case 3: index = iter->index + (5-index); break;
        }
        break;
    case GL_POLYGON: {
        break;
    }
    default:
        Scm_Warn("unsupported GL draw mode: %d", iter->mode);
    }
    return index_vector_ref(iter->ivec, index);
}

float *Banyan_FaceIterGetVertexPosition(BanyanFaceIter *iter, int vindex)
{
    ScmPoint4fArray *varr;
    int index;
    if (BANYAN_FACE_ITER_END(iter)) return NULL;
    if (!SCM_POINT4F_ARRAY_P(iter->shape->vertices)) return NULL;
    varr = SCM_POINT4F_ARRAY(iter->shape->vertices);
    index = Banyan_FaceIterGetIndex(iter, vindex);
    return SCM_POINT4F_ARRAY_REFV(varr, index);
}

float *Banyan_FaceIterGetVertexNormal(BanyanFaceIter *iter, int vindex)
{
    ScmVector4fArray *vnor;
    int index;
    if (BANYAN_FACE_ITER_END(iter)) return NULL;
    if (!SCM_VECTOR4F_ARRAY_P(iter->shape->normals)) return NULL;
    vnor = SCM_VECTOR4F_ARRAY(iter->shape->normals);
    index = Banyan_FaceIterGetIndex(iter, vindex);
    return SCM_VECTOR4F_ARRAY_REFV(vnor, index);
}

void Banyan_FaceIterGetCenter(BanyanFaceIter *iter, float pt[])
{
    ScmPoint4fArray *varr;
    int nv, i;
    float f[4];
    
    if (BANYAN_FACE_ITER_END(iter)) return;
    nv = Banyan_FaceIterGetNumVertices(iter);
    if (nv == 0) return;
    SCM_VECTOR4F_OP(_, f[_] = 0.0);
    for (i=0; i<nv; i++) {
        float *p = Banyan_FaceIterGetVertexPosition(iter, i);
        SCM_VECTOR4F_OP(_, f[_] += p[_]);
    }
    SCM_VECTOR4F_OP(_, pt[_] = f[_]/nv);
}

/*
 * Scheme-level iterators
 */
typedef ScmObj (*FaceFoldProc)(BanyanFaceIter *iter, ScmObj proc, ScmObj knil);

/* common C continuation */
static ScmObj face_fold_cc(ScmObj value, void *data[])
{
    BanyanFaceIter *iter = (BanyanFaceIter*)data[0];
    ScmObj proc; 
    FaceFoldProc ffcb;

    if (!Banyan_FaceIterNext(iter)) return value;
    
    proc = SCM_OBJ(data[1]);
    ffcb = (FaceFoldProc)(data[2]);
    Scm_VMPushCC(face_fold_cc, data, 3);
    return ffcb(iter, proc, value);
}

static ScmObj face_fold_index(BanyanFaceIter *iter, ScmObj proc, ScmObj knil)
{
    int n = Banyan_FaceIterGetNumVertices(iter);
    ScmObj v = Scm_MakeU32Vector(n, 0);
    Banyan_FaceIterGetIndices(iter, (GLuint*)SCM_U32VECTOR_ELEMENTS(v));
    return Scm_VMApply2(proc, v, knil);
}

ScmObj Banyan_VMFaceFoldIndices(BanyanShape *shape, ScmObj proc, ScmObj knil)
{
    BanyanFaceIter *iter = SCM_NEW(BanyanFaceIter);
    void *data[3];

    if (!Banyan_FaceIterInit(shape, iter)) return knil;
    data[0] = (void*)iter;
    data[1] = (void*)proc;
    data[2] = (void*)face_fold_index;
    Scm_VMPushCC(face_fold_cc, data, 3);
    return face_fold_index(iter, proc, knil);
}

static ScmObj face_fold_count_index(BanyanFaceIter *iter, ScmObj proc,
                                    ScmObj knil)
{
    int n = Banyan_FaceIterGetNumVertices(iter);
    int cnt = BANYAN_FACE_ITER_COUNT(iter);
    ScmObj v = Scm_MakeU32Vector(n, 0);
    Banyan_FaceIterGetIndices(iter, (GLuint*)SCM_U32VECTOR_ELEMENTS(v));
    return Scm_VMApply3(proc, SCM_MAKE_INT(cnt), v, knil);
}

ScmObj Banyan_VMFaceFoldCountIndices(BanyanShape *shape, ScmObj proc,
                                     ScmObj knil)
{
    BanyanFaceIter *iter = SCM_NEW(BanyanFaceIter);
    void *data[3];

    if (!Banyan_FaceIterInit(shape, iter)) return knil;
    data[0] = (void*)iter;
    data[1] = (void*)proc;
    data[2] = (void*)face_fold_count_index;
    Scm_VMPushCC(face_fold_cc, data, 3);
    return face_fold_count_index(iter, proc, knil);
}

static ScmObj face_fold_vertex(BanyanFaceIter *iter, ScmObj proc, ScmObj knil)
{
    int n = Banyan_FaceIterGetNumVertices(iter), i;
    ScmObj v = Scm_MakePoint4fArrayv(n, NULL);
    for (i=0; i<n; i++) {
        float *f = Banyan_FaceIterGetVertexPosition(iter, i);
        SCM_POINT4F_ARRAY_SET(v, i, f[0], f[1], f[2], f[3]);
    }
    return Scm_VMApply2(proc, v, knil);
}

ScmObj Banyan_VMFaceFoldVertices(BanyanShape *shape, ScmObj proc, ScmObj knil)
{
    BanyanFaceIter *iter = SCM_NEW(BanyanFaceIter);
    void *data[3];

    if (!Banyan_FaceIterInit(shape, iter)) return knil;
    data[0] = (void*)iter;
    data[1] = (void*)proc;
    data[2] = (void*)face_fold_vertex;
    Scm_VMPushCC(face_fold_cc, data, 3);
    return face_fold_vertex(iter, proc, knil);
}

static ScmObj face_fold_count_vertex(BanyanFaceIter *iter,
                                     ScmObj proc, ScmObj knil)
{
    int n = Banyan_FaceIterGetNumVertices(iter), i;
    int cnt = BANYAN_FACE_ITER_COUNT(iter);
    ScmObj v = Scm_MakePoint4fArrayv(n, NULL);
    for (i=0; i<n; i++) {
        float *f = Banyan_FaceIterGetVertexPosition(iter, i);
        SCM_POINT4F_ARRAY_SET(v, i, f[0], f[1], f[2], f[3]);
    }
    return Scm_VMApply3(proc, SCM_MAKE_INT(cnt), v, knil);
}

ScmObj Banyan_VMFaceFoldCountVertices(BanyanShape *shape,
                                      ScmObj proc, ScmObj knil)
{
    BanyanFaceIter *iter = SCM_NEW(BanyanFaceIter);
    void *data[3];

    if (!Banyan_FaceIterInit(shape, iter)) return knil;
    data[0] = (void*)iter;
    data[1] = (void*)proc;
    data[2] = (void*)face_fold_count_vertex;
    Scm_VMPushCC(face_fold_cc, data, 3);
    return face_fold_count_vertex(iter, proc, knil);
}

static ScmObj face_fold_center(BanyanFaceIter *iter, ScmObj proc, ScmObj knil)
{
    ScmObj v = Scm_MakePoint4fv(NULL);
    Banyan_FaceIterGetCenter(iter, SCM_POINT4F_D(v));
    return Scm_VMApply2(proc, v, knil);
}

ScmObj Banyan_VMFaceFoldCenter(BanyanShape *shape, ScmObj proc, ScmObj knil)
{
    BanyanFaceIter *iter = SCM_NEW(BanyanFaceIter);
    void *data[3];

    if (!Banyan_FaceIterInit(shape, iter)) return knil;
    data[0] = (void*)iter;
    data[1] = (void*)proc;
    data[2] = (void*)face_fold_center;
    Scm_VMPushCC(face_fold_cc, data, 3);
    return face_fold_center(iter, proc, knil);
}


/*----------------------------------------------------------------
 * Normal calculation
 */
static inline void normal_from_3_vertex(float n[],
                                        float v0[],
                                        float v1[],
                                        float v2[])
{
    float x01[4], x12[4];
    SCM_VECTOR4F_SUBV(x01, v0, v1);
    SCM_VECTOR4F_SUBV(x12, v1, v2);
    SCM_VECTOR4F_CROSSV(n, x01, x12);
    SCM_VECTOR4F_NORMALIZEV(n);
}

static inline void normal_from_N_vertex(float n[], float **v, int nv)
{
    float nn[4], *v0, *v1;
    int i;
    SCM_VECTOR4F_OP(_, nn[_] = 0.0);
    for (i=0; i<nv; i++) {
        v0 = v[i]; v1 = (i==nv-1)? v[0] : v[i+1];
        nn[0] -= (v1[1] - v0[1]) * (v1[2] + v0[2]);
        nn[1] -= (v1[2] - v0[2]) * (v1[0] + v0[0]);
        nn[2] -= (v1[0] - v0[0]) * (v1[1] + v0[1]);
    }
    SCM_VECTOR4F_NORMALIZEV(nn);
    SCM_VECTOR4F_OP(_, n[_] = nn[_]);
}

#define SMALL_POLYGON  20

void Banyan_ShapeSetFacetNormals(BanyanShape *shape)
{
    ScmObj normals;
    ScmPoint4fArray *vtx;
    BanyanFaceIter iter;
    GLuint ind[SMALL_POLYGON];
    float *pnt[SMALL_POLYGON];
    GLuint *pind = ind, pind_alloc = 0, i;
    float **ppnt = pnt;
    
    if (!SCM_PAIRP(shape->faces)) return;
    if (!SCM_POINT4F_ARRAY_P(shape->vertices)) return;
    vtx = SCM_POINT4F_ARRAY(shape->vertices);
    normals = Scm_MakeVector4fArrayv(shape->numFaces, NULL);

    if (!Banyan_FaceIterInit(shape, &iter)) {
        shape->facetNormals = SCM_FALSE;
        return;
    }
    do {
        int nv = Banyan_FaceIterGetNumVertices(&iter);
        int cnt = BANYAN_FACE_ITER_COUNT(&iter);
        float *norm = SCM_VECTOR4F_ARRAY_REFV(normals, cnt);
        if (nv == 3) {
            Banyan_FaceIterGetIndices(&iter, ind);
            normal_from_3_vertex(norm,
                                 SCM_POINT4F_ARRAY_REFV(vtx, ind[0]),
                                 SCM_POINT4F_ARRAY_REFV(vtx, ind[1]),
                                 SCM_POINT4F_ARRAY_REFV(vtx, ind[2]));
        } else {
            if (nv >= SMALL_POLYGON && pind_alloc < nv) {
                pind_alloc = nv;
                pind = SCM_NEW_ATOMIC2(GLuint*, nv * sizeof(GLuint));
                ppnt = SCM_NEW_ATOMIC2(float**, nv * sizeof(float*));
            }
            Banyan_FaceIterGetIndices(&iter, pind);
            for (i=0; i<nv; i++) {
                ppnt[i] = SCM_POINT4F_ARRAY_REFV(vtx, pind[i]);
            }
            normal_from_N_vertex(norm, ppnt, nv);
        }
    } while (Banyan_FaceIterNext(&iter));
    shape->facetNormals = normals;
}

void Banyan_ShapeSetVertexNormals(BanyanShape *shape)
{
    ScmObj normals;
    ScmVector4fArray *facetNormals;
    BanyanFaceIter iter;
    int numVertices;
    GLuint ind[SMALL_POLYGON];
    GLuint *pind = ind, pind_alloc = 0, i;
    
    if (!SCM_PAIRP(shape->faces)) return;
    if (!SCM_POINT4F_ARRAY_P(shape->vertices)) return;
    numVertices = SCM_POINT4F_ARRAY_SIZE(shape->vertices);
    if (!SCM_VECTOR4F_ARRAY_P(shape->facetNormals)) {
        Banyan_ShapeSetFacetNormals(shape);
        SCM_ASSERT(SCM_VECTOR4F_ARRAY_P(shape->facetNormals));
        facetNormals = SCM_VECTOR4F_ARRAY(shape->facetNormals);
    }
    normals = Scm_MakeVector4fArrayv(numVertices, NULL);

    if (!Banyan_FaceIterInit(shape, &iter)) {
        shape->normals = SCM_FALSE;
        return;
    }
    do {
        int nv = Banyan_FaceIterGetNumVertices(&iter);
        int cnt = BANYAN_FACE_ITER_COUNT(&iter);
        float *fnorm, *vnorm;
        
        if (nv >= SMALL_POLYGON && pind_alloc < nv) {
            pind_alloc = nv;
            pind = SCM_NEW_ATOMIC2(GLuint*, nv * sizeof(GLuint));
        }
        Banyan_FaceIterGetIndices(&iter, pind);
        fnorm = SCM_VECTOR4F_ARRAY_REFV(facetNormals, cnt);
        for (i=0; i<nv; i++) {
            vnorm = SCM_VECTOR4F_ARRAY_REFV(normals, pind[i]);
            SCM_VECTOR4F_OP(_, vnorm[_] += fnorm[_]);
        }
    } while (Banyan_FaceIterNext(&iter));
    for (i=0; i<numVertices; i++) {
        float *vnorm = SCM_VECTOR4F_ARRAY_REFV(normals, i);
        SCM_VECTOR4F_NORMALIZEV(vnorm);
    }
    shape->normals = normals;
}

/*----------------------------------------------------------------
 * Initialization stuff
 */
                                        
static ScmClassStaticSlotSpec shape_slots[] = {
    BANYAN_NODE_SLOT_SPEC,
    SCM_CLASS_SLOT_SPEC("vertices", shape_vertices_get, shape_vertices_set),
    SCM_CLASS_SLOT_SPEC("normals", shape_normals_get, shape_normals_set),
    SCM_CLASS_SLOT_SPEC("faces", shape_faces_get, shape_faces_set),
    SCM_CLASS_SLOT_SPEC("num-faces", shape_numfaces_get, NULL),
    SCM_CLASS_SLOT_SPEC("facet-normals", shape_facetnormals_get, NULL),
    { NULL }
};

void Banyan__InitShape(ScmModule *mod)
{
    Scm_InitBuiltinClass(&Banyan_ShapeClass, "<%shape>", shape_slots,
                         sizeof(BanyanShape), mod);
}
