/*
 * transform.c - transform handling
 *
 *  Copyright(C) 2002-2003 by Shiro Kawai (shiro@acm.org)
 *
 *  Permission to use, copy, modify, distribute this software and
 *  accompanying documentation for any purpose is hereby granted,
 *  provided that existing copyright notices are retained in all
 *  copies and that this notice is included verbatim in all
 *  distributions.
 *  This software is provided as is, without express or implied
 *  warranty.  In no circumstances the author(s) shall be liable
 *  for any damages arising out of the use of this software.
 *
 *  $Id: transform.c,v 1.7 2003/12/10 01:00:50 shirok Exp $
 */

#include "banyan.h"

/*----------------------------------------------------------------
 * Class stuff
 */

static ScmObj transform_allocate(ScmClass *, ScmObj);

SCM_DEFINE_BUILTIN_CLASS(Banyan_TransformClass, NULL,
                      NULL, NULL, transform_allocate,
                      Banyan_NodeCPL);

ScmObj transform_allocate(ScmClass *klass, ScmObj initargs)
{
    BanyanTransform *transform = SCM_ALLOCATE(BanyanTransform, klass);
    SCM_SET_CLASS(transform, klass);
    BANYAN_HEADER_INIT(transform);
    transform->gmatrix = SCM_MATRIX4F(Scm_MakeMatrix4fv(NULL));
    transform->lmatrix  = SCM_MATRIX4F(Scm_MakeMatrix4fv(NULL));
    transform->parent = NULL;
    return SCM_OBJ(transform);
}

ScmObj Banyan_MakeTransform(void)
{
    return transform_allocate(BANYAN_TRANSFORM_CLASS, SCM_NIL);
}

static ScmObj transform_gmatrix_get(ScmObj transform)
{
    return SCM_OBJ(BANYAN_TRANSFORM(transform)->gmatrix);
}

static void transform_gmatrix_set(ScmObj transform, ScmObj value)
{
    if (!SCM_MATRIX4FP(value)) {
        Scm_Error("<matrix4f> is required, but got %S", value);
    }
    BANYAN_TRANSFORM(transform)->gmatrix = SCM_MATRIX4F(value);
}

static ScmObj transform_lmatrix_get(ScmObj transform)
{
    return SCM_OBJ(BANYAN_TRANSFORM(transform)->lmatrix);
}

static void transform_lmatrix_set(ScmObj transform, ScmObj value)
{
    if (!SCM_MATRIX4FP(value)) {
        Scm_Error("<matrix4f> is required, but got %S", value);
    }
    BANYAN_TRANSFORM(transform)->lmatrix = SCM_MATRIX4F(value);
}

static ScmObj transform_cmatrix_get(ScmObj transform)
{
    ScmMatrix4f *cm = BANYAN_TRANSFORM(transform)->cmatrix;
    return cm == NULL? SCM_FALSE : SCM_OBJ(cm);
}

static void transform_cmatrix_set(ScmObj transform, ScmObj value)
{
    if (SCM_FALSEP(value)) {
        BANYAN_TRANSFORM(transform)->cmatrix = NULL;
    } else if (SCM_MATRIX4FP(value)) {
        BANYAN_TRANSFORM(transform)->cmatrix = SCM_MATRIX4F(value);
    } else {
        Scm_Error("<matrix4f> or #f is required, but got %S", value);
    }
}

static ScmObj transform_imatrix_get(ScmObj transform)
{
    ScmMatrix4f *cm = BANYAN_TRANSFORM(transform)->imatrix;
    return cm == NULL? SCM_FALSE : SCM_OBJ(cm);
}

static void transform_imatrix_set(ScmObj transform, ScmObj value)
{
    if (SCM_FALSEP(value)) {
        BANYAN_TRANSFORM(transform)->imatrix = NULL;
    } else if (SCM_MATRIX4FP(value)) {
        BANYAN_TRANSFORM(transform)->imatrix = SCM_MATRIX4F(value);
    } else {
        Scm_Error("<matrix4f> or #f is required, but got %S", value);
    }
}

static ScmObj transform_parent_get(ScmObj transform)
{
    return BANYAN_TRANSFORM(transform)->parent ? SCM_OBJ(BANYAN_TRANSFORM(transform)->parent) : SCM_FALSE;
}

static void transform_parent_set(ScmObj transform, ScmObj value)
{
    if (SCM_ISA(value, BANYAN_TRANSFORM_CLASS)) {
        BANYAN_TRANSFORM(transform)->parent = BANYAN_TRANSFORM(value);
    } else if (SCM_FALSEP(value)) {
        BANYAN_TRANSFORM(transform)->parent = NULL;
    } else {
        Scm_Error("<%transform> or #f is required, but got %S", value);
    }
}

/*----------------------------------------------------------------
 * Update protocol
 */

void Banyan_TransformUpdate(BanyanTransform *transform)
{
    BanyanTransform *parent;
    if (transform->parent) {
        if (transform->cmatrix) {
            float tmp[16];
            Scm_Matrix4fMulMatrix4fv(tmp,
                                     SCM_MATRIX4F_D(transform->parent->gmatrix),
                                     SCM_MATRIX4F_D(transform->cmatrix));
            Scm_Matrix4fMulMatrix4fv(SCM_MATRIX4F_D(transform->gmatrix),
                                     tmp,
                                     SCM_MATRIX4F_D(transform->lmatrix));
        } else {
            Scm_Matrix4fMulMatrix4fv(SCM_MATRIX4F_D(transform->gmatrix),
                                     SCM_MATRIX4F_D(transform->parent->gmatrix),
                                     SCM_MATRIX4F_D(transform->lmatrix));
        }
    } else {
        memcpy(SCM_MATRIX4F_D(transform->gmatrix),
               SCM_MATRIX4F_D(transform->lmatrix),
               sizeof(float)*16);
    }
}

/*----------------------------------------------------------------
 * Initialization stuff
 */
                                        
static ScmClassStaticSlotSpec transform_slots[] = {
    BANYAN_NODE_SLOT_SPEC,
    SCM_CLASS_SLOT_SPEC("gmatrix", transform_gmatrix_get, transform_gmatrix_set),
    SCM_CLASS_SLOT_SPEC("lmatrix", transform_lmatrix_get, transform_lmatrix_set),
    SCM_CLASS_SLOT_SPEC("cmatrix", transform_cmatrix_get, transform_cmatrix_set),
    SCM_CLASS_SLOT_SPEC("imatrix", transform_imatrix_get, transform_imatrix_set),
    SCM_CLASS_SLOT_SPEC("parent", transform_parent_get, transform_parent_set),
    { NULL }
};

void Banyan__InitTransform(ScmModule *mod)
{
    Scm_InitBuiltinClass(&Banyan_TransformClass, "<%transform>", transform_slots,
                         sizeof(BanyanTransform), mod);
}
